/*
 * Copyright (c) 2015, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>

#include <ucg_uart.h>
#include <ucg_stm32_uart.h>

static void disable_tx_irq(struct ucg_uart *uart)
{
	struct ucg_stm32_uart *stm32_uart = uart->driver_data;

	stm32_uart->uart->SR &= ~(USART_FLAG_TXE);
	stm32_uart->uart->CR1 &= ~(USART_CR1_TXEIE);
}

static void enable_tx_irq(struct ucg_uart *uart)
{
	struct ucg_stm32_uart *stm32_uart = uart->driver_data;

	stm32_uart->uart->CR1 |= USART_CR1_TXEIE;
}

static uint8_t tx_ready(struct ucg_uart *uart)
{
	struct ucg_stm32_uart *stm32_uart = uart->driver_data;

	return !!(stm32_uart->uart->SR & USART_FLAG_TXE);
}

static uint8_t rx_ready(struct ucg_uart *uart)
{
	struct ucg_stm32_uart *stm32_uart = uart->driver_data;

	return !!(stm32_uart->uart->SR & USART_FLAG_RXNE);
}

static void set_udr(struct ucg_uart *uart, char c)
{
	struct ucg_stm32_uart *stm32_uart = uart->driver_data;

	stm32_uart->uart->DR = c;
}

static char get_udr(struct ucg_uart *uart)
{
	struct ucg_stm32_uart *stm32_uart = uart->driver_data;

	return stm32_uart->uart->DR;
}

static int set_conf(struct ucg_uart *uart, const struct ucg_uart_config *conf)
{
	USART_InitTypeDef u;
	GPIO_InitTypeDef gpio;
	NVIC_InitTypeDef nvic;
	struct ucg_stm32_uart *stm32_uart = uart->driver_data;
	int i;

	/* even if it is asked to disable, keep the RCC enabled */

	/* Enable the peripheral clock. */
	RCC_APB1PeriphClockCmd(stm32_uart->rcc_uart, ENABLE);
	__asm("dsb");

	/* Enable the peripheral clock for GPIO. */
	RCC_AHB1PeriphClockCmd(stm32_uart->rcc_gpio, ENABLE);
	__asm("dsb");

	if (conf->enable == 0) {
		USART_ITConfig(stm32_uart->uart, USART_IT_TXE, DISABLE);
		USART_ITConfig(stm32_uart->uart, USART_IT_RXNE, DISABLE);
		USART_Cmd(stm32_uart->uart, ENABLE);

		nvic.NVIC_IRQChannel = stm32_uart->irq;
		nvic.NVIC_IRQChannelCmd = DISABLE;
		NVIC_Init(&nvic);
		return 0;
	}

	if (conf->nbits != 8)
		return -EINVAL;

	/* connect to alternate function */
	for (i = 0; i < 16; i++) {
		if ((1 << i) & stm32_uart->gpio_pins) {
			GPIO_PinAFConfig(stm32_uart->gpio, i,
				stm32_uart->gpio_af);
		}
	}

	GPIO_StructInit(&gpio);
	gpio.GPIO_Speed = stm32_uart->gpio_speed;

	/* configure rx and tx */
	gpio.GPIO_Pin = stm32_uart->gpio_pins;
	gpio.GPIO_Mode = GPIO_Mode_AF;
	GPIO_Init(stm32_uart->gpio, &gpio);

	/* USART configuration */
	u.USART_BaudRate = conf->baudrate;
	u.USART_WordLength = USART_WordLength_8b;
	u.USART_StopBits = USART_StopBits_1;
	u.USART_Parity = USART_Parity_No;
	u.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
	u.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;
	USART_Init(stm32_uart->uart, &u);

	/* Enable and set EXTI Interrupt to the defined priority */
	nvic.NVIC_IRQChannel = stm32_uart->irq;
	/* nvic.NVIC_IRQChannelPreemptionPriority = stm32_uart->irq_preempt_prio; */
	/* nvic.NVIC_IRQChannelSubPriority = stm32_uart->irq_sub_prio; */
	nvic.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&nvic);

	/* Enable RX interruption */
	USART_ITConfig(stm32_uart->uart, USART_IT_RXNE, ENABLE);

	/* Enable USART */
	USART_Cmd(stm32_uart->uart, ENABLE);

	return 0;
}

const struct ucg_uart_driver_ops stm32_uart_ops = {
	.disable_tx_irq = disable_tx_irq,
	.enable_tx_irq = enable_tx_irq,
	.tx_ready = tx_ready,
	.rx_ready = rx_ready,
	.set_udr = set_udr,
	.get_udr = get_udr,
	.set_conf = set_conf,
};
