/*
 * Copyright (c) 2015, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if defined(__ARM_EABI__)
#include "stm32f4xx.h"
#elif defined(__AVR__)
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <ctype.h>

#include <ucg_irq.h>
#include <ucg_cirbuf.h>
#include <ucg_uart.h>

#include "uart.h"

/* rx & tx buffers */
static char rx_buf[64];
static struct ucg_cirbuf rx_cirbuf;
static char tx_buf[64];
static struct ucg_cirbuf tx_cirbuf;
/* generic uart struct */
struct ucg_uart main_uart;

#if defined(__ARM_EABI__)

#include <ucg_gloss_chardev.h>
#include <ucg_stm32_uart.h>

static struct ucg_stm32_uart stm32_uart_data = {
	.uart = USART2,
	.rcc_uart = RCC_APB1Periph_USART2,
	.rcc_gpio = RCC_AHB1Periph_GPIOA,
	.gpio = GPIOA,
	.gpio_af = GPIO_AF_USART2,
	.gpio_pins = GPIO_Pin_2 | GPIO_Pin_3,
	.gpio_speed = GPIO_Speed_25MHz,
	.irq = USART2_IRQn,
	.irq_preempt_prio = 2,
	.irq_sub_prio = 0,
};

static _ssize_t
uart_write_r(__attribute__((unused)) struct _reent *r,
	__attribute__((unused)) int fd,
	const void *ptr, size_t len)
{
	size_t i;
	const char *buf = ptr;

	for (i = 0; i < len; i++)
		ucg_uart_send(&main_uart, buf[i], WAIT);

	return len;
}

static _ssize_t
uart_read_r(__attribute__((unused)) struct _reent *r,
	__attribute__((unused)) int fd,
	void *ptr, size_t len)
{
	int c;
	char *buf = ptr;

	c = ucg_uart_recv(&main_uart, NOWAIT);
	if (c < 0) {
		r->_errno = EAGAIN;
		return 0;
	}
	buf[0] = c;
	return 1;
}

void USART2_IRQHandler(void)
{
	if ((USART2->SR & USART_FLAG_TXE))
		ucg_uart_tx_intr(&main_uart);
	if ((USART2->SR & USART_FLAG_RXNE))
		ucg_uart_rx_intr(&main_uart);
}

struct ucg_chardev uart_stdin_dev = {
	.name = "stdin",
	.open_r = NULL,
	.close_r = NULL,
	.read_r = uart_read_r,
	.write_r = NULL,
};

struct ucg_chardev uart_stdout_dev = {
	.name = "stdout",
	.open_r = NULL,
	.close_r = NULL,
	.read_r = NULL,
	.write_r = uart_write_r,
};

struct ucg_chardev uart_stderr_dev = {
	.name = "stderr",
	.open_r = NULL,
	.close_r = NULL,
	.read_r = NULL,
	.write_r = uart_write_r,
};

static void uart_register_stdio(void)
{
	ucg_chardev_register(&uart_stdin_dev);
	ucg_chardev_register(&uart_stdout_dev);
	ucg_chardev_register(&uart_stderr_dev);

	/* Disable buffering */
	setbuf(stdin, NULL);
	setbuf(stdout, NULL);
	setbuf(stderr, NULL);
}

#elif defined(__AVR__)

#include <ucg_avr_uart.h>
/* avr-specific uart struct */
static struct ucg_avr_uart avr_uart_data = {
	.reg_udr = &UDR0,
	.reg_ucsra = &UCSR0A,
	.reg_ucsrb = &UCSR0B,
	.reg_ucsrc = &UCSR0C,
	.reg_ubrrl = &UBRR0L,
	.reg_ubrrh = &UBRR0H,
	.bit_udre = UDRE0,
	.bit_rxc = RXC0,
	.bit_udrie = UDRIE0,
	.bit_rxen = RXEN0,
	.bit_txen = TXEN0,
	.bit_rxcie = RXCIE0,
	.bit_u2x = U2X0,
};

/* send on stdout */
static int std_send(char c, FILE *f)
{
	(void)f;
	ucg_uart_send(&main_uart, c, WAIT);
	return 0;
}

/* recv on stdin */
static int std_recv(FILE *f)
{
	int16_t c;

	(void)f;
	c = ucg_uart_recv(&main_uart, NOWAIT);
	if (c < 0)
		return _FDEV_EOF;

	return c;
}

SIGNAL(USART_RX_vect)
{
	ucg_uart_rx_intr(&main_uart);
}

SIGNAL(USART_UDRE_vect)
{
	ucg_uart_tx_intr(&main_uart);
}
#endif

int uart_init(void)
{
	int ret;
	struct ucg_uart_config conf;
	const void *uart_ops;
	void *uart_data;

#if defined(__ARM_EABI__)
	uart_ops = &stm32_uart_ops;
	uart_data = &stm32_uart_data;
#else
	uart_ops = &avr_uart_ops;
	uart_data = &avr_uart_data;
#endif

	ret = ucg_uart_init(&main_uart, uart_ops, uart_data,
		&rx_cirbuf, rx_buf, sizeof(rx_buf),
		&tx_cirbuf, tx_buf, sizeof(tx_buf));
	if (ret < 0)
		return ret;

	ucg_uart_getconf(&main_uart, &conf);
	conf.baudrate = 57600;
	ret = ucg_uart_setconf(&main_uart, &conf);
	if (ret < 0)
		return ret;

#if defined(__ARM_EABI__)
	uart_register_stdio();
#elif defined(__AVR__)
	fdevopen(std_send, std_recv);
#endif
	return ret;
}
