/*
 * Copyright 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UCG_CIRBUF_H_
#define UCG_CIRBUF_H_

#include <stdio.h>

/**
 * A circular buffer.
 */
struct ucg_cirbuf {
	unsigned maxlen; /**< Total length of the fifo (number of elements). */
	unsigned start;  /**< Index of the first element. */
	unsigned len;    /**< Current len of fifo. */
	char *buf;       /**< Pointer to the data buffer. */
};

/**
 * Initialize a circular buffer.
 *
 * @param cbuf
 *   A pointer to an uninitialized circular buffer structure.
 * @param buf
 *   The buffer used to store the data.
 * @param start
 *   The index of head at initialization.
 * @param maxlen
 *   The size of the buffer.
 */
void ucg_cirbuf_init(struct ucg_cirbuf *cbuf, char *buf, unsigned start,
	unsigned maxlen);

/**
 * Check if the circular buffer is full.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @return
 *   1 if the circular buffer is full, else 0.
 */
static inline int ucg_cirbuf_is_full(const struct ucg_cirbuf *cbuf)
{
	return cbuf->len == cbuf->maxlen;
}

/**
 * Check if the circular buffer is empty.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @return
 *   1 if the circular buffer is empty, else 0.
 */
static inline int ucg_cirbuf_is_empty(const struct ucg_cirbuf *cbuf)
{
	return cbuf->len == 0;
}

/**
 * Get the length of data in the circular buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @return
 *   The current length of data in the circular buffer.
 */
static inline unsigned ucg_cirbuf_get_len(const struct ucg_cirbuf *cbuf)
{
	return cbuf->len;
}

/**
 * Get the size of the circular buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @return
 *   Return the maximum size of the circular buffer (used + free elements)
 */
static inline unsigned ucg_cirbuf_get_maxlen(const struct ucg_cirbuf *cbuf)
{
	return cbuf->maxlen;
}

/**
 * Get the lenght of free space in the circular buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @return
 *   Return the length of free space.
 */
static inline unsigned ucg_cirbuf_get_freelen(const struct ucg_cirbuf *cbuf)
{
	return cbuf->maxlen - cbuf->len;
}

/**
 * Iterator for a circular buffer
 *
 *   cirbuf: struct cirbuf pointer
 *   i: an integer internally used in the macro
 *   elt: char that takes the value for each iteration
 */
#define UCG_CIRBUF_FOREACH(cirbuf, i, elt)				\
	for (i = 0, elt = (cirbuf)->buf[(cirbuf)->start];		\
	     i < ((cirbuf)->len);					\
	     i ++,  elt = (cirbuf)->buf[((cirbuf)->start + i) %		\
		     ((cirbuf)->maxlen)])

/**
 * Add a character at the head of the circular buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @param c
 *   The character to add.
 * @return
 *   Return 0 on success, or a negative value on error.
 */
int ucg_cirbuf_add_head_safe(struct ucg_cirbuf *cbuf, char c);

/**
 * Add a character at the head of the circular buffer.
 *
 * The function does not check that there is enough free space
 * in the buffer, so it has to be done by the caller. If it's
 * not the case, undefined behavior will occur.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @param c
 *   The character to add.
 */
void ucg_cirbuf_add_head(struct ucg_cirbuf *cbuf, char c);

/**
 * Add a character at the tail of the circular buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @param c
 *   The character to add.
 * @return
 *   Return 0 on success, or a negative value on error.
 */
int ucg_cirbuf_add_tail_safe(struct ucg_cirbuf *cbuf, char c);

/**
 * Add a character at the tail of the circular buffer.
 *
 * The function does not check that there is enough free space
 * in the buffer, so it has to be done by the caller. If it's
 * not the case, undefined behavior will occur.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @param c
 *   The character to add.
 */
void ucg_cirbuf_add_tail(struct ucg_cirbuf *cbuf, char c);

/**
 * Remove a char at the head of the circular buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @return
 *   Return 0 on success, or a negative value on error.
 */
int ucg_cirbuf_del_head_safe(struct ucg_cirbuf *cbuf);

/**
 * Remove a char at the head of the circular buffer.
 *
 * The function does not check that there is enough elements
 * in the buffer, so it has to be done by the caller. If it's
 * not the case, undefined behavior will occur.
 *
 * @param cbuf
 *   The circular buffer pointer.
 */
void ucg_cirbuf_del_head(struct ucg_cirbuf *cbuf);

/**
 * Remove a char at the tail of the circular buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @return
 *   Return 0 on success, or a negative value on error.
 */
int ucg_cirbuf_del_tail_safe(struct ucg_cirbuf *cbuf);

/**
 * Remove a char at the tail of the circular buffer.
 *
 * The function does not check that there is enough elements
 * in the buffer, so it has to be done by the caller. If it's
 * not the case, undefined behavior will occur.
 *
 * @param cbuf
 *   The circular buffer pointer.
 */
void ucg_cirbuf_del_tail(struct ucg_cirbuf *cbuf);

/**
 * Return the element at the tail of the circular buffer.
 *
 * The circular buffer must not be empty or an undefined character
 * will be returned.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @return
 *   The character at the tail of the circular buffer.
 */
char ucg_cirbuf_get_head(const struct ucg_cirbuf *cbuf);

/**
 * Return the element at the tail of the circular buffer.
 *
 * The circular buffer must not be empty or an undefined character
 * will be returned.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @return
 *   The character at the tail of the circular buffer.
 */
char ucg_cirbuf_get_tail(const struct ucg_cirbuf *cbuf);

/**
 * Add a buffer at the head of the circular buffer.
 *
 * Add 'n' bytes of buffer pointed by 'buf' ad the head of th
 * circular buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @param buf
 *   The pointer to the buffer.
 * @param n
 *   Number of bytes to add.
 * @return
 *   Return 0 on success, or a negative value on error.
 */
int ucg_cirbuf_add_buf_head(struct ucg_cirbuf *cbuf, const char *buf,
	unsigned n);

/**
 * Add a buffer at the tail of the circular buffer.
 *
 * Add 'n' bytes of buffer pointed by 'buf' ad the tail of th
 * circular buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @param buf
 *   The pointer to the buffer.
 * @param n
 *   Number of bytes to add.
 * @return
 *   Return 0 on success, or a negative value on error.
 */
int ucg_cirbuf_add_buf_tail(struct ucg_cirbuf *cbuf, const char *buf,
	unsigned n);

/**
 * Remove chars at the head of the circular buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @param n
 *   Number of bytes to remove.
 * @return
 *   Return 0 on success, or a negative value on error.
 */
int ucg_cirbuf_del_buf_head(struct ucg_cirbuf *cbuf, unsigned n);

/**
 * Remove chars at the tail of the circular buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @param n
 *   Number of bytes to remove.
 * @return
 *   Return 0 on success, or a negative value on error.
 */
int ucg_cirbuf_del_buf_tail(struct ucg_cirbuf *cbuf, unsigned n);

/**
 * Copy multiple bytes from the head of the circular buffer.
 *
 * Copy a maximum of 'n' characters from the head of the circular buffer
 * into a flat buffer pointed by 'buf'. If the circular buffer is
 * smaller than n, less bytes are copied.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @param buf
 *   The pointer to the buffer.
 * @param n
 *   Maximum number of bytes to copy.
 * @return
 *   Return the number of copied chars.
 */
int ucg_cirbuf_get_buf_head(const struct ucg_cirbuf *cbuf, char *buf,
	unsigned n);

/**
 * Copy multiple bytes from the tail of the circular buffer.
 *
 * Copy a maximum of 'n' characters from the tail of the circular buffer
 * into a flat buffer pointed by 'buf'. If the circular buffer is
 * smaller than n, less bytes are copied.
 *
 * @param cbuf
 *   The circular buffer pointer.
 * @param buf
 *   The pointer to the buffer.
 * @param n
 *   Maximum number of bytes to copy.
 * @return
 *   Return the number of copied chars.
 */
int ucg_cirbuf_get_buf_tail(const struct ucg_cirbuf *cbuf, char *buf,
	unsigned n);

/**
 * Set the start of the data to the index 0 of the internal buffer.
 *
 * After a call to this function, it is possible for the caller to
 * use cbuf->buf as a linear (read-only) buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 */
void ucg_cirbuf_align_left(struct ucg_cirbuf *cbuf);

/**
 * Set the end of the data to the last index of the internal buffer.
 *
 * After a call to this function, it is possible for the caller to
 * use cbuf->buf as a linear (read-only) buffer.
 *
 * @param cbuf
 *   The circular buffer pointer.
 */
void ucg_cirbuf_align_right(struct ucg_cirbuf *cbuf);

#endif /* CIRBUF_H_ */
