/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>
#include <errno.h>

#include <ucg_cirbuf.h>

/* init a circular buffer */
void
ucg_cirbuf_init(struct ucg_cirbuf *cbuf, char *buf, unsigned start,
	unsigned maxlen)
{
	cbuf->maxlen = maxlen;
	cbuf->len = 0;
	cbuf->start = start;
	cbuf->buf = buf;
}

/* return the index of the tail (contains an empty element) */
static unsigned
ucg_cirbuf_get_end(const struct ucg_cirbuf *cbuf)
{
	unsigned end;

	end = cbuf->start + cbuf->len;
	if (end >= cbuf->maxlen)
		end -= cbuf->maxlen;

	return end;
}

/* multiple add at head */
int
ucg_cirbuf_add_buf_head(struct ucg_cirbuf *cbuf, const char *buf, unsigned n)
{
	unsigned remain = n;
	int copy_start;
	unsigned copy_len;

	if (remain == 0 || remain > ucg_cirbuf_get_freelen(cbuf))
		return -EINVAL;

	copy_start = cbuf->start - remain;
	if (copy_start < 0)
		copy_start += cbuf->maxlen;
	cbuf->start = copy_start;
	cbuf->len += remain;
	copy_len = remain;
	if ((unsigned)copy_start + copy_len >= cbuf->maxlen)
		copy_len = cbuf->maxlen - copy_start;
	if (copy_len > 0) {
		memcpy(cbuf->buf + copy_start, buf, copy_len);
		buf += copy_len;
		remain -= copy_len;
	}
	if (remain == 0)
		return n;

	copy_len = remain;
	memcpy(cbuf->buf, buf, copy_len);

	return n;
}

/* multiple add at tail */
int
ucg_cirbuf_add_buf_tail(struct ucg_cirbuf *cbuf, const char *buf, unsigned n)
{
	unsigned remain = n;
	unsigned copy_start, copy_len;

	if (remain == 0 || remain > ucg_cirbuf_get_freelen(cbuf))
		return -EINVAL;

	copy_start = ucg_cirbuf_get_end(cbuf);
	cbuf->len += remain;
	copy_len = remain;
	if (copy_start + copy_len >= cbuf->maxlen)
		copy_len = cbuf->maxlen - copy_start;
	if (copy_len > 0) {
		memcpy(cbuf->buf + copy_start, buf, copy_len);
		buf += copy_len;
		remain -= copy_len;
	}
	if (remain == 0)
		return n;

	copy_len = remain;
	memcpy(cbuf->buf, buf, copy_len);

	return n;
}

/* single add at head */
static inline void
__cirbuf_add_head(struct ucg_cirbuf *cbuf, char c)
{
	unsigned start = cbuf->start;

	if (start == 0)
		start = cbuf->maxlen - 1;
	else
		start = start - 1;
	cbuf->buf[start] = c;
	cbuf->start = start;
	cbuf->len++;
}

/* single add at head, checking if full first */
int
ucg_cirbuf_add_head_safe(struct ucg_cirbuf *cbuf, char c)
{
	if (!ucg_cirbuf_is_full(cbuf)) {
		__cirbuf_add_head(cbuf, c);
		return 0;
	}
	return -EINVAL;
}

/* single add at head */
void
ucg_cirbuf_add_head(struct ucg_cirbuf *cbuf, char c)
{
	__cirbuf_add_head(cbuf, c);
}

/* single add at tail */
static inline void
__cirbuf_add_tail(struct ucg_cirbuf *cbuf, char c)
{
	unsigned end = ucg_cirbuf_get_end(cbuf);

	cbuf->buf[end] = c;
	cbuf->len++;
}

/* single add at tail, checking if full first */
int
ucg_cirbuf_add_tail_safe(struct ucg_cirbuf *cbuf, char c)
{
	if (!ucg_cirbuf_is_full(cbuf)) {
		__cirbuf_add_tail(cbuf, c);
		return 0;
	}
	return -EINVAL;
}

/* single add at tail */
void
ucg_cirbuf_add_tail(struct ucg_cirbuf *cbuf, char c)
{
	__cirbuf_add_tail(cbuf, c);
}

/* multiple delete at head */
int
ucg_cirbuf_del_buf_head(struct ucg_cirbuf *cbuf, unsigned size)
{
	if (size == 0 || size > ucg_cirbuf_get_len(cbuf))
		return -EINVAL;

	cbuf->len -= size;
	if (ucg_cirbuf_is_empty(cbuf)) {
		cbuf->start += size - 1;
		cbuf->start %= cbuf->maxlen;
	}
	else {
		cbuf->start += size;
		cbuf->start %= cbuf->maxlen;
	}
	return 0;
}

/* multiple delete at tail */
int
ucg_cirbuf_del_buf_tail(struct ucg_cirbuf *cbuf, unsigned size)
{
	if (size == 0 || size > ucg_cirbuf_get_len(cbuf))
		return -EINVAL;

	cbuf->len -= size;
	return 0;
}

/* single del at head */
static inline void
__cirbuf_del_head(struct ucg_cirbuf *cbuf)
{
	cbuf->len --;
	if (!ucg_cirbuf_is_empty(cbuf)) {
		cbuf->start ++;
		cbuf->start %= cbuf->maxlen;
	}
}

/* single del at head, checking if empty first */
int
ucg_cirbuf_del_head_safe(struct ucg_cirbuf *cbuf)
{
	if (cbuf && !ucg_cirbuf_is_empty(cbuf)) {
		__cirbuf_del_head(cbuf);
		return 0;
	}
	return -EINVAL;
}

/* single del at head */
void
ucg_cirbuf_del_head(struct ucg_cirbuf *cbuf)
{
	__cirbuf_del_head(cbuf);
}

/* single del at tail */
static inline void
__cirbuf_del_tail(struct ucg_cirbuf *cbuf)
{
	cbuf->len--;
}

/* single del at tail, checking if empty first */
int
ucg_cirbuf_del_tail_safe(struct ucg_cirbuf *cbuf)
{
	if (cbuf && !ucg_cirbuf_is_empty(cbuf)) {
		__cirbuf_del_tail(cbuf);
		return 0;
	}
	return -EINVAL;
}

/* single del at tail */
void
ucg_cirbuf_del_tail(struct ucg_cirbuf *cbuf)
{
	__cirbuf_del_tail(cbuf);
}

/* convert to buffer */
int
ucg_cirbuf_get_buf_head(const struct ucg_cirbuf *cbuf, char *buf, unsigned n)
{
	unsigned remain = n;
	unsigned cirbuf_len, copy_start, copy_len;

	cirbuf_len = ucg_cirbuf_get_len(cbuf);
	if (remain >= cirbuf_len)
		remain = cirbuf_len;

	if (remain == 0)
		return 0;

	copy_start = cbuf->start;
	copy_len = remain;
	if (copy_start + copy_len >= cbuf->maxlen)
		copy_len = cbuf->maxlen - copy_start;
	if (copy_len > 0) {
		memcpy(buf, cbuf->buf + copy_start, copy_len);
		buf += copy_len;
		remain -= copy_len;
	}
	if (remain == 0)
		return n;

	copy_len = remain;
	memcpy(buf, cbuf->buf, copy_len);

	return n;
}

/* convert to buffer */
int
ucg_cirbuf_get_buf_tail(const struct ucg_cirbuf *cbuf, char *buf, unsigned n)
{
	unsigned remain = n;
	int copy_start;
	unsigned cirbuf_len, copy_len;

	cirbuf_len = ucg_cirbuf_get_len(cbuf);
	if (remain >= cirbuf_len)
		remain = cirbuf_len;

	if (remain == 0)
		return 0;

	copy_start = ucg_cirbuf_get_end(cbuf) - remain;
	if (copy_start < 0)
		copy_start += cbuf->maxlen;
	copy_len = remain;
	if ((unsigned)copy_start + copy_len >= cbuf->maxlen)
		copy_len = cbuf->maxlen - copy_start;
	if (copy_len > 0) {
		memcpy(buf, cbuf->buf + copy_start, copy_len);
		buf += copy_len;
		remain -= copy_len;
	}
	if (remain == 0)
		return n;

	copy_len = remain;
	memcpy(buf, cbuf->buf, copy_len);

	return n;
}

/* get head */
char
ucg_cirbuf_get_head(const struct ucg_cirbuf *cbuf)
{
	return cbuf->buf[cbuf->start];
}

/* get tail */
char
ucg_cirbuf_get_tail(const struct ucg_cirbuf *cbuf)
{
	unsigned end;

	/* should not happen */
	if (cbuf->len == 0)
		return 0;

	end = cbuf->start + cbuf->len - 1;
	if (end >= cbuf->maxlen)
		end -= cbuf->maxlen;
	return cbuf->buf[end];
}

static void
__ucg_cirbuf_shift(struct ucg_cirbuf *cbuf, unsigned n)
{
	char tmp, tmp2;
	unsigned start, cur, min;

	start = 0;
	cur = 0;
	tmp = cbuf->buf[0];
	min = cbuf->maxlen;

	while (1) {
		cur = cur + n;
		if (cur >= cbuf->maxlen)
			cur -= cbuf->maxlen;
		tmp2 = cbuf->buf[cur];
		cbuf->buf[cur] = tmp;
		tmp = tmp2;
		if (cur == start) {
			if ((cur + 1) == min)
				break;
			cur++;
			tmp = cbuf->buf[cur];
			start = cur;
		} else if (cur < min) {
			min = cur;
		}
	}

	cbuf->start += n;
	if (cbuf->start >= cbuf->maxlen)
		cbuf->start -= cbuf->maxlen;
}

void ucg_cirbuf_align_left(struct ucg_cirbuf *cbuf)
{
	__ucg_cirbuf_shift(cbuf, cbuf->maxlen - cbuf->start);
}

void ucg_cirbuf_align_right(struct ucg_cirbuf *cbuf)
{
	unsigned end = ucg_cirbuf_get_end(cbuf);
	__ucg_cirbuf_shift(cbuf, cbuf->maxlen - end);
}

#ifdef TEST_CIRBUF
#include <stdint.h>
#include <assert.h>

void dump_it(struct ucg_cirbuf * cbuf)
{
	unsigned i;
	int idx;
	char e;

	printf("sta=%2.2d len=%2.2d/%2.2d { ",
		cbuf->start,
		ucg_cirbuf_get_len(cbuf),
		ucg_cirbuf_get_maxlen(cbuf));

	for (i = 0; i < ucg_cirbuf_get_maxlen(cbuf); i++) {
		idx = i - cbuf->start;
		if (idx < 0)
			idx += cbuf->maxlen;
		if (idx < (int)ucg_cirbuf_get_len(cbuf))
			printf("%2.2x, ", cbuf->buf[i] & 0xFF);
		else
			printf("XX, ");
	}
	printf("}  -> ");

	printf("[ ");
	UCG_CIRBUF_FOREACH(cbuf, i, e) {
		printf("%2.2x, ", e & 0xFF);
	}
	printf("]\n");
}

int main(void)
{
	unsigned i;
	struct ucg_cirbuf my_fifo;
	char fifo_buf[16];

	char buf1[] = { 0x10, 0x11, 0x12 };
	char buf2[] = { 0x20, 0x21, 0x22, 0x23 };

	char tmp_buf[16];
	char ref_buf[] = { 0x20, 0x21, 0x22, 0x23, 0x01, 0x10, 0x11, 0x12 };

	/* Test 1 */

	printf("Test 1\n");

	ucg_cirbuf_init(&my_fifo, fifo_buf, 0, 4);
	assert(ucg_cirbuf_is_empty(&my_fifo));
	assert(!ucg_cirbuf_is_full(&my_fifo));
	dump_it(&my_fifo);

	ucg_cirbuf_add_tail(&my_fifo, 1);
	assert(ucg_cirbuf_get_head(&my_fifo) == 1);
	assert(ucg_cirbuf_get_tail(&my_fifo) == 1);
	dump_it(&my_fifo);


	ucg_cirbuf_add_tail(&my_fifo, 2);
	assert(!ucg_cirbuf_is_empty(&my_fifo));
	assert(!ucg_cirbuf_is_full(&my_fifo));
	dump_it(&my_fifo);

	ucg_cirbuf_add_tail(&my_fifo, 3);
	assert(ucg_cirbuf_get_head(&my_fifo) == 1);
	assert(ucg_cirbuf_get_tail(&my_fifo) == 3);
	dump_it(&my_fifo);

	ucg_cirbuf_add_tail(&my_fifo, 4);
	assert(!ucg_cirbuf_is_empty(&my_fifo));
	assert(ucg_cirbuf_is_full(&my_fifo));
	dump_it(&my_fifo);

	ucg_cirbuf_del_tail(&my_fifo);
	dump_it(&my_fifo);
	assert(ucg_cirbuf_get_tail(&my_fifo) == 3);
	assert(ucg_cirbuf_get_head(&my_fifo) == 1);

	ucg_cirbuf_del_head(&my_fifo);
	assert(ucg_cirbuf_get_tail(&my_fifo) == 3);
	assert(ucg_cirbuf_get_head(&my_fifo) == 2);
	dump_it(&my_fifo);

	ucg_cirbuf_del_head(&my_fifo);
	assert(ucg_cirbuf_get_tail(&my_fifo) == 3);
	assert(ucg_cirbuf_get_head(&my_fifo) == 3);
	dump_it(&my_fifo);

	ucg_cirbuf_del_head(&my_fifo);
	assert(ucg_cirbuf_is_empty(&my_fifo));
	dump_it(&my_fifo);


	/* Test 2 */

	printf("Test 2\n");

	ucg_cirbuf_init(&my_fifo, fifo_buf, 2, 4);
	dump_it(&my_fifo);

	ucg_cirbuf_add_head(&my_fifo, 4);
	assert(ucg_cirbuf_get_head(&my_fifo) == 4);
	assert(ucg_cirbuf_get_tail(&my_fifo) == 4);
	dump_it(&my_fifo);


	ucg_cirbuf_add_head(&my_fifo, 3);
	assert(!ucg_cirbuf_is_empty(&my_fifo));
	assert(!ucg_cirbuf_is_full(&my_fifo));
	dump_it(&my_fifo);

	ucg_cirbuf_add_head(&my_fifo, 2);
	assert(ucg_cirbuf_get_head(&my_fifo) == 2);
	assert(ucg_cirbuf_get_tail(&my_fifo) == 4);
	dump_it(&my_fifo);

	ucg_cirbuf_add_head(&my_fifo, 1);
	assert(!ucg_cirbuf_is_empty(&my_fifo));
	assert(ucg_cirbuf_is_full(&my_fifo));
	dump_it(&my_fifo);


	/* Test 3 */

	printf("Test 3\n");

	for (i = 0; i < 16; i++) {
		ucg_cirbuf_init(&my_fifo, fifo_buf, i, 16);
		dump_it(&my_fifo);
		ucg_cirbuf_add_buf_head(&my_fifo, buf1, sizeof(buf1));
		dump_it(&my_fifo);
		ucg_cirbuf_add_head(&my_fifo, 1);
		dump_it(&my_fifo);
		ucg_cirbuf_add_buf_head(&my_fifo, buf2, sizeof(buf2));
		dump_it(&my_fifo);
		ucg_cirbuf_get_buf_head(&my_fifo, tmp_buf, sizeof(tmp_buf));
		assert(memcmp(tmp_buf, ref_buf, sizeof(ref_buf)) == 0);
	}

	/* Test 4 */

	printf("Test 4\n");

	for (i = 0; i < 16; i++) {
		ucg_cirbuf_init(&my_fifo, fifo_buf, i, 16);
		dump_it(&my_fifo);
		ucg_cirbuf_add_buf_tail(&my_fifo, buf2, sizeof(buf2));
		dump_it(&my_fifo);
		ucg_cirbuf_add_tail(&my_fifo, 1);
		dump_it(&my_fifo);
		ucg_cirbuf_add_buf_tail(&my_fifo, buf1, sizeof(buf1));
		dump_it(&my_fifo);
		ucg_cirbuf_get_buf_tail(&my_fifo, tmp_buf, sizeof(tmp_buf));
		assert(memcmp(tmp_buf, ref_buf, sizeof(ref_buf)) == 0);

		printf("align left\n");
		ucg_cirbuf_align_left(&my_fifo);
		dump_it(&my_fifo);
		ucg_cirbuf_get_buf_tail(&my_fifo, tmp_buf, sizeof(tmp_buf));
		assert(memcmp(tmp_buf, ref_buf, sizeof(ref_buf)) == 0);
		assert(my_fifo.start == 0);

		printf("align right\n");
		ucg_cirbuf_align_right(&my_fifo);
		dump_it(&my_fifo);
		ucg_cirbuf_get_buf_tail(&my_fifo, tmp_buf, sizeof(tmp_buf));
		assert(memcmp(tmp_buf, ref_buf, sizeof(ref_buf)) == 0);
		assert(my_fifo.start + my_fifo.len == my_fifo.maxlen);
	}

	/* Test 5 */

	printf("Test 5\n");

	ucg_cirbuf_init(&my_fifo, fifo_buf, 10, 16);
	dump_it(&my_fifo);
	i = 0;
	while (ucg_cirbuf_add_tail_safe(&my_fifo, i) == 0)
		i++;
	dump_it(&my_fifo);
	ucg_cirbuf_del_buf_tail(&my_fifo, 10);
	dump_it(&my_fifo);
	assert(ucg_cirbuf_get_len(&my_fifo) == 6);
	assert(ucg_cirbuf_del_buf_tail(&my_fifo, 10) != 0);
	assert(ucg_cirbuf_get_tail(&my_fifo) == 5);
	assert(ucg_cirbuf_get_head(&my_fifo) == 0);

	return 0;
}
#endif
