/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UCG_CMD_H_
#define UCG_CMD_H_

#include <stdarg.h>
#include <stdio.h>
#ifdef UCG_CMD_HAVE_TERMIOS
#include <termios.h>
#endif

#include "ucg_cmd_parse.h"
#include "ucg_cmd_rdline.h"

/**
 * A command line structure.
 */
struct ucg_cmd {
	ucg_cmd_ctx_t *ctx;      /**< The list of commands for this context. */
	struct ucg_rdline rdl;   /**< The associated rdline structure. */
	char prompt[UCG_RDLINE_PROMPT_SIZE];  /**< The command line prompt. */
#ifdef UCG_CMD_HAVE_TERMIOS
	struct termios oldterm;  /**< The old termios info */
#endif
	void *opaque;            /**< A user opaque pointer. */
};

/**
 * Allocate and initialize a new command line structure
 *
 * Allocate and initialize a new command line structure, using the
 * specified context, prompt and input/output streams.
 *
 * Once unused, the command line structure should be freed using
 * ucg_cmd_free().
 *
 * @param ctx
 *   The command line context.
 * @param prompt
 *   The command line prompt. The string is copied in the ucg_cmd
 *   structure. The prompt must be smaller than UCG_RDLINE_PROMPT_SIZE.
 * @param f_in
 *   The input stream.
 * @param f_out
 *   The output stream.
 * @return
 *   The newly allocated command line structure.
 */
struct ucg_cmd *ucg_cmd_new(ucg_cmd_ctx_t *ctx,
	const char *prompt, FILE *f_in, FILE *f_out);

/**
 * Initialize a new command line structure
 *
 * Initialize a command line structure, using the specified prompt and
 * specified input/output streams.
 *
 * @param cl
 *   The uninitialize command line structure
 * @param ctx
 *   The command line context
 * @param prompt
 *   The command line prompt. The string is copied in the ucg_cmd
 *   structure. The prompt must be smaller than UCG_RDLINE_PROMPT_SIZE.
 * @param f_in
 *   The input stream
 * @param f_out
 *   The output stream
 */
void ucg_cmd_init(struct ucg_cmd *cl, ucg_cmd_ctx_t *ctx,
	const char *prompt, FILE *f_in, FILE *f_out);

/**
 * Set the prompt of the given command line.
 */
void ucg_cmd_set_prompt(struct ucg_cmd *cl, const char *prompt);

/**
 * Free a previously allocated command line
 *
 * The structure pointed by cl is freed. The streams f_in and f_out are
 * closed, except if it's stdin, stdout, or stderr. Note: the user
 * should call ucg_cmd_quit() before calling this function.
 *
 * @param cl
 *   The command line pointer.
 */
void ucg_cmd_free(struct ucg_cmd *cl);

/**
 * Parse a file use the given cmd context
 *
 * The output file descriptor (used for instance by cmd_printf) is
 * given as a parameter. It can be -1 if no output is required.
 */
struct ucg_cmd *ucg_cmd_file_new(ucg_cmd_ctx_t *ctx,
	const char *prompt, const char *path, FILE *f_out);

/**
 * Print data on the output of the command line
 *
 * This function is a wrapper to rdline_printf().
 *
 * @param cl
 *   The command line pointer.
 * @param fmt
 *   The format string, followed by variable arguments.
 * @return
 *   On success, return the number of characters printed (not including
 *   the trailing '\0'). On error, a negative value is returned.
 */
int ucg_cmd_printf(struct ucg_cmd *cl, const char *fmt, ...);

/**
 * Push an input buffer to the command line.
 *
 * Typically, this function is called by ucg_cmd_interact() to send the
 * input characters to the command line process. It can also be called
 * by a user, for instance when new data is received from an input
 * socket.
 *
 * @param cl
 *   The command line pointer.
 * @param buf
 *   The address of the input buffer.
 * @param size
 *   The len of the input buffer.
 * @return
 *   The function returns the number of processed characters, or a
 *   negative value on error (ex: EOF reached or command line exited).
 */
int ucg_cmd_in(struct ucg_cmd *cl, const char *buf, int size);

/* flags for ucg_cmd_interact */
#define UCG_CMD_F_IGNORE_EOF 0x0001   /**< ignore eof when polling */

/**
 * Start command line on configured file descriptor. This function
 * loops until the user explicitelly call ucg_cmd_quit(), or if the
 * input fd reaches EOF.
 *
 * @param cl
 *   The command line pointer
 * @param flags
 *   Any flags from UCG_CMD_F_*
 *   - UCG_CMD_F_IGNORE_EOF: on eof, clear error and continue polling
 */
void ucg_cmd_interact(struct ucg_cmd *cl, unsigned flags);

/**
 * Stop a running command line.
 *
 * Actually it will call ucg_rdline_quit() on the associated rdline.
 *
 * @param cl
 *   The command line pointer.
 */
void ucg_cmd_quit(struct ucg_cmd *cl);

#endif /* UCG_CMD_H_ */
