/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UCG_CMD_PARSE_H_
#define UCG_CMD_PARSE_H_

#include <sys/types.h>

#ifndef offsetof
#define offsetof(type, field)  ((size_t) &( ((type *)0)->field) )
#endif

// XXX config
#define UCG_CMD_MAX_TOKEN_SIZE 32 /* including '\0' */
#define UCG_CMD_MAX_DSTBUF_SIZE 128

/**
 * A token header.
 *
 * Stores a pointer to the ops struct, and the offset: the place to
 * write the parsed result in the destination structure.
 */
struct ucg_cmd_tk_hdr {
	struct ucg_cmd_tk_ops *ops;
	unsigned int offset;
};
typedef struct ucg_cmd_tk_hdr ucg_cmd_tk_hdr_t;

/**
 * Operations on token.
 */
struct ucg_cmd_tk_ops {
	/**
	 * parse() converts a buffer containing a token into its parsed
	 * value. Ex: an integer string is converted into its integer
	 * value. The result is stored in "result" whose size is
	 * "res_size". It returns 0 on success and a negative value on
	 * error.
	 */
	int (*parse)(ucg_cmd_tk_hdr_t *tk, const char *line,
		void *result, unsigned int res_size);

	/**
	 * complete_start() prepares a completion operation. The
	 * "opaque" arg is an opaque pointer that will be given to
	 * complete_iterate() function. It can be used to store private
	 * data for this completion. For each complete_start() call, the
	 * user must call complete_end() at the end of iterations (if
	 * defined) in case some data have to be freed.
	 *
	 * Return a negative value if completion is not possible, or 0
	 * on success.
	 */
	int (*complete_start)(ucg_cmd_tk_hdr_t *tk, const char *line,
		void **opaque);

	/**
	 * complete_iterate() copy in "dst" buffer the next possible
	 * completion for this token. Return 0 on success (final
	 * completion = completion until the end of the token), 1 if
	 * it's an intermediate completion (token not fully completed),
	 * or a negative value on error (or when there is no more
	 * completion). Refer to ucg_cmd_complete_string_iterate() for
	 * an example.
	 */
	int (*complete_iterate)(ucg_cmd_tk_hdr_t *tk, void **opaque,
		char *dst, unsigned int dst_size);

	/**
	 * complete_end() is called when the iteration on this token is
	 * finished, this function should free all things allocated
	 * during complete_start().
	 */
	void (*complete_end)(ucg_cmd_tk_hdr_t *tk, void **opaque);

	/**
	 * help() fills the dstbuf with the help for the token. It returns
	 * -1 on error and 0 on success.
	 */
	int (*help)(ucg_cmd_tk_hdr_t *tk, char *dst, unsigned int dst_size);
};

struct ucg_cmd;

/**
 * Store a command, defined by a list of tokens and a callback function.
 */
struct ucg_cmd_inst {
	/** callback function when the commend is parsed */
	void (*f)(void *parsed_result, struct ucg_cmd *cl, void *opaque);
	void *data;     /**< opaque pointer given as is to the callback */
	char *help_str; /**< help for this command */
	ucg_cmd_tk_hdr_t *tokens[]; /**< list of tokens */
};
typedef struct ucg_cmd_inst ucg_cmd_inst_t;

/**
 * A context is a list of commands.
 */
struct ucg_cmd_ctx {
	const char *name;                /**< The name of the context */
	const ucg_cmd_inst_t *insts[];   /**< List of commands */
};
typedef struct ucg_cmd_ctx ucg_cmd_ctx_t;

/* return status for parsing */
#define UCG_CMD_PARSE_SUCCESS        0
#define UCG_CMD_PARSE_EMPTY         -1
#define UCG_CMD_PARSE_NOMATCH       -2
#define UCG_CMD_PARSE_AMBIGUOUS     -3
#define UCG_CMD_PARSE_UNTERMINATED_QUOTE -4

/**
 * Try to parse a buffer according to the specified context. The
 * argument linebuf must end with "\n\0".
 *
 * The function returns:
 * - UCG_CMD_PARSE_SUCCESS (0) on success
 * - UCG_CMD_PARSE_EMPTY if there is nothing to parse
 * - UCG_CMD_PARSE_NOMATCH if line does not match any command
 * - UCG_CMD_PARSE_AMBIGUOUS if several commands match
 * - UCG_CMD_PARSE_UNTERMINATED_QUOTE if a quote is used incorrectly
 */
int ucg_cmd_parse(struct ucg_cmd *cl, const char *linebuf, void *opaque);

/* return status for completion */
#define UCG_CMD_COMPLETE_APPEND    0
#define UCG_CMD_COMPLETE_NONE     -1
#define UCG_CMD_COMPLETE_MANY     -2

/**
 * ucg_cmd_complete() tries to complete the buffer given as a parameter.
 *
 * It returns:
 *   - UCG_CMD_COMPLETE_APPEND (0) on success, when a completion is
 *     done (one possible choice). In this case, the chars are
 *     appended in dst buffer.
 *   - UCG_CMD_COMPLETE_NONE: error, no possible completion
 *   - UCG_CMD_COMPLETE_MANY: error, many possble completions, need to call
 *     ucg_cmd_help() function to see all the possibilities.
 */
int ucg_cmd_complete(struct ucg_cmd *cl, const char *buf, char *dst,
	unsigned int size);

/**
 * Display a contextual help.
 *
 * @param cl
 *   The command line pointer
 * @param line
 *   The current line buffer.
 */
void ucg_cmd_help(struct ucg_cmd *cl, const char *line);

/**
 * Check if the character ends a token
 *
 * @param c
 *   The character.
 * @return
 *   True if (c == '\0' || iscomment(c) || isblank(c) ||
 *   isendofline(c))
 */
int ucg_cmd_isendoftoken(char c);

/**
 * Quote a string and escape original quotes
 *
 * @param dst
 *   The destination buffer.
 * @param dstlen
 *   The length of the destination buffer.
 * @param src
 *   The source buffer.
 * @return
 *   Return 0 on success, a negative value on error.
 */
int ucg_cmd_quote_token(char *dst, unsigned dstlen, const char *src);

/**
 * Get one token
 *
 * The function removes quotes (if any) and stop copying when the end of
 * token is reached. The destination buffer is '\0'-terminated.
 *
 * @param dst
 *   The destination buffer.
 * @param dstlen
 *   The length of the destination buffer.
 * @param src
 *   The source buffer.
 * @return
 *   Return the number of "consumed" bytes from the source buffer, or
 *   a negative value on error
 */
int ucg_cmd_get_token(char *dst, unsigned dstlen, const char *src);

#endif /* UCG_CMD_PARSE_H_ */
