/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef CMD_RDLINE_H_
#define CMD_RDLINE_H_

/**
 * This file is a small equivalent to the GNU readline library, it was
 * originally designed for very small systems (an 8-bits Atmel AVR
 * microcontroller), but the library can now run on many emmbedded
 * systems with or without OS.
 *
 * Obviously, it does not support as many things as the GNU readline,
 * but at least it supports some interresting features like a kill
 * buffer and a command history.
 *
 * It also have a feature that does not have the GNU readline (as far
 * as I know): it is possible to have several instances of readline
 * running at the same time, even on a monothread program, since it
 * works with callbacks.
 */

#include <sys/types.h>
#include <stdio.h>

#include <ucg_cirbuf.h>
#include <ucg_cmd_vt100.h>

/* configuration */
#define UCG_RDLINE_BUF_SIZE 32
#define UCG_RDLINE_PROMPT_SIZE  16
#define UCG_RDLINE_VT100_BUF_SIZE  8
#define UCG_RDLINE_HISTORY_BUF_SIZE 64
#define UCG_RDLINE_MAX_LINES 23  /* pager */

enum ucg_rdline_status {
	UCG_RDLINE_STOPPED,
	UCG_RDLINE_RUNNING,
	UCG_RDLINE_EXITED
};

struct ucg_rdline;

/**
 * type of callback given to ucg_rdline_help() to display the content of
 * the help. The first argument is the rdline pointer. The other args
 * are buffer and size.
 */
typedef int (ucg_rdline_printf_t)(struct ucg_rdline *rdl,
	const char *fmt, ...);

/**
 * type of callback invoked when a command is parsed. "rdl" is a pointer
 * to the ucg_rdline structure, "line" is a pointer to the current line
 * string ('\0' terminated).
 */
typedef void (ucg_rdline_validate_t)(struct ucg_rdline *rdl,
	const char *line);

/**
 * type of callback invoked when a completion is requested. "rdl" is a
 * pointer to the ucg_rdline structure, "line" is a pointer to the
 * current line string ('\0' terminated). The characters to append
 * are written to dstbuf.
 * Return 0 on success: dstbuf contains the characters to append to
 *   the current line ('\0' terminated)
 * Else return a negative value if no completion is performed.
 */
typedef int (ucg_rdline_complete_t)(struct ucg_rdline *rdl, const char *line,
	char *dstbuf, unsigned int dstsize);

/**
 * callback invoked when a the help is requested. "rdl" is a pointer to
 * the ucg_rdline structure, "line" is a pointer to the current line string
 * ('\0' terminated).
 */
typedef void (ucg_rdline_help_t)(struct ucg_rdline *rdl, const char *line);

typedef void (ucg_rdline_pager_cb_t)(struct ucg_rdline *, void *);

struct ucg_rdline {
	enum ucg_rdline_status status;
	FILE *f_in;
	FILE *f_out;

	/* rdline bufs */
	struct ucg_cirbuf left;
	struct ucg_cirbuf right;
	char left_buf[UCG_RDLINE_BUF_SIZE+1]; /* reserve 1 char for the \0 */
	char right_buf[UCG_RDLINE_BUF_SIZE];

	char prompt[UCG_RDLINE_PROMPT_SIZE];

#ifndef UCG_CMD_NO_RDLINE_KILL_BUF
	char kill_buf[UCG_RDLINE_BUF_SIZE];
	unsigned int kill_size;
#endif

#ifndef UCG_CMD_NO_RDLINE_HISTORY
	/* history */
	struct ucg_cirbuf history;
	char history_buf[UCG_RDLINE_HISTORY_BUF_SIZE];
	int history_cur_line;
#endif

	/* callbacks and func pointers */
	ucg_rdline_validate_t *validate;
	ucg_rdline_complete_t *complete;
	ucg_rdline_help_t *help;

	/* vt100 parser */
	struct ucg_cmd_vt100 vt100;

	/* opaque pointer */
	void *opaque;

#ifndef UCG_CMD_NO_PAGER
	char *pager_buf; /* buffer used to store paged data */
	int pager_len; /* total len of buffer */
	int pager_off; /* offset of next data */
	int pager_lines; /* number of lines displayed */
	ucg_rdline_pager_cb_t *pager_cb; /* callback once paging is finished */
	void *pager_arg; /* argument of callback */
	int pager_ret; /* saved return value */
#endif
};

/**
 * Init fields for a struct ucg_rdline.
 *
 * @param rdl A pointer to an uninitialized struct ucg_rdline
 * @param fd_in
 *   Input file descriptor
 * @param fd_out
 *   Output file descriptor
 * @param validate
 *   A pointer to the function to execute when the user validates the
 *   buffer.
 * @param complete
 *   A pointer to the function to execute when the user completes the
 *   buffer.
 * @param help
 *   A pointer to the function to execute when the user ask for
 *   contextual help.
 */
void ucg_rdline_init(struct ucg_rdline *rdl,
		 FILE *f_in, FILE *f_out,
		 ucg_rdline_validate_t *validate,
		 ucg_rdline_complete_t *complete,
		 ucg_rdline_help_t *help);


/**
 * Init the current buffer, and display a prompt.
 *
 * Also set the rdline status to "running", overriding previous
 * ucg_rdline_stop() or ucg_rdline_quit().
 *
 * @param rdl
 *   A pointer to an initialized struct ucg_rdline
 * @param prompt
 *   A string containing the prompt, or NULL to keep existing one
 */
void ucg_rdline_newline(struct ucg_rdline *rdl, const char *prompt);

/**
 * Ignore all subsequent received chars.
 *
 * @param rdl
 *   A pointer to a struct ucg_rdline
 */
void ucg_rdline_stop(struct ucg_rdline *rdl);

/**
 * Exit from running rdline loop
 *
 * Same than ucg_rdline_stop() except that next calls to ucg_rdline_char_in()
 * will return UCG_RDLINE_RES_EXITED. Hence, any running rdline() function is
 * interrupted.
 *
 * @param rdl
 *   A pointer to a struct ucg_rdline
 */
void ucg_rdline_quit(struct ucg_rdline *rdl);

/**
 * Restart after a call to ucg_rdline_stop() or ucg_rdline_quit()
 *
 * @param rdl
 *   A pointer to a struct ucg_rdline
 */
void ucg_rdline_restart(struct ucg_rdline *rdl);

/**
 * Redisplay the current buffer
 *
 * @param rdl
 *   A pointer to a struct ucg_rdline
 */
void ucg_rdline_redisplay(struct ucg_rdline *rdl);

/* return status for ucg_rdline_char_in() */
#define UCG_RDLINE_RES_SUCCESS          0
#define UCG_RDLINE_RES_VALIDATED        1
#define UCG_RDLINE_RES_COMPLETED        2
#define UCG_RDLINE_RES_NOT_RUNNING     -1
#define UCG_RDLINE_RES_EOF             -2
#define UCG_RDLINE_RES_EXITED          -3
#define UCG_RDLINE_RES_CANNOT_COMPLETE -3

/**
 * Append a char to the readline buffer.
 *
 * @param rdl
 *   A pointer to a struct ucg_rdline
 * @param c
 *   The character to append
 * @return
 *   - UCG_RDLINE_RES_VALIDATED when the line has been validated.
 *   - UCG_RDLINE_RES_NOT_RUNNING if it is not running.
 *   - UCG_RDLINE_RES_EOF if EOF (ctrl-d on an empty line).
 *   - UCG_RDLINE_RES_EXITED if user called ucg_rdline_quit()
 *   - Else return UCG_RDLINE_RES_SUCCESS.
 */
int ucg_rdline_char_in(struct ucg_rdline *rdl, char c);

#define UCG_RDLINE_F_IGNORE_EOF 0x0001   /**< ignore eof when polling */

/**
 * Read (and edit) a line
 *
 * @param rdl
 *   A pointer to a struct ucg_rdline
 * @param prompt
 *   The prompt string
 * @param flags
 *   Any flags from UCG_RDLINE_F_*
 *   - UCG_RDLINE_F_IGNORE_EOF: on eof, clear error and continue polling
 * @return
 *   - UCG_RDLINE_RES_VALIDATED when the line has been validated.
 *   - UCG_RDLINE_RES_NOT_RUNNING if it is not running.
 *   - UCG_RDLINE_RES_EOF if EOF (ctrl-d on an empty line).
 *   - UCG_RDLINE_RES_EXITED if user called ucg_rdline_quit()
 */
int ucg_rdline(struct ucg_rdline *rdl, const char *prompt, unsigned flags);

/**
 * write a buffer on rdline file descriptor
 *
 * @param rdl
 *   The rdline descriptor
 * @param buf
 *   Pointer to the buffer
 * @param count
 *   Number of bytes to write
 * @return
 *   On success, the number of bytes written is returned (zero
 *   indicates nothing was written). On error, -1 is returned, and
 *   errno is set appropriately
 */
ssize_t ucg_rdline_write(struct ucg_rdline *rdl, void *buf, size_t count);

/**
 * write on rdline file descriptor according to a format string
 *
 * @param rdl
 *   The rdline descriptor
 * @param fmt
 *   The format strings
 * @return
 *   On success, return the number of characters printed (not including
 *   the trailing '\0'). On error, a negative value is returned.
 */
int ucg_rdline_printf(struct ucg_rdline *rdl, const char *fmt, ...);

/**
 * write on rdline file descriptor according to a format string
 *
 * @param rdl
 *   The rdline descriptor
 * @param fmt
 *   The format strings
 * @param ap
 *   Variable argument list
 * @return
 *   Upon successful return, these functions return the number of
 *   characters printed (not including the trailing '\0' used to end
 *   output to strings). On error, a negative value is returned.
 */
int ucg_rdline_vprintf(struct ucg_rdline *rdl, const char *fmt, va_list ap);

/**
 * Return the current buffer, terminated by '\0'.
 *
 * @param rdl
 *   A pointer to a struct ucg_rdline
 * @return
 *   The rdline buffer
 */
const char *ucg_rdline_get_buffer(struct ucg_rdline *rdl);

/**
 * Add the buffer to history.
 *
 * @param rdl
 *   A pointer to a struct ucg_rdline
 * @param buf
 *   A buffer that is terminated by '\0'
 * @return
 *   - 0 on success
 *   - negative on error
 */
int ucg_rdline_add_history(struct ucg_rdline *rdl, const char *buf);

/**
 * Clear current history
 *
 * @param rdl
 *   A pointer to a struct ucg_rdline
 */
void ucg_rdline_clear_history(struct ucg_rdline *rdl);

/**
 * Get the i-th history item
 *
 * @param rdl
 *   A pointer to a struct ucg_rdline
 * @param i
 *   The index of the history item
 * @return
 *   The i-th string of history, or NULL on error.
 */
const char *ucg_rdline_get_history_item(struct ucg_rdline *rdl, unsigned int i);

#ifndef UCG_CMD_NO_PAGER
/**
 * Write data asynchronously (using pager if needed)
 *
 * If there is enough place to print data on the current page, it is
 * printed synchronously. Else, a temporary buffer is allocated and
 * the data is stored in it. When the main rdline is called again, the
 * pager is flushed before parsing any other commands.
 *
 * @param rdl
 *   The rdline descriptor
 * @param buf
 *   Buffer to be sent
 * @param len
 *   Length of buffer to be sent
 * @return
 *   On success, the number of bytes written is returned (zero
 *   indicates nothing was written). On error, -1 is returned, and
 *   errno is set appropriately
 */
ssize_t ucg_rdline_pager_write(struct ucg_rdline *rdl, void *buf, size_t len);

/**
 * Print data asynchronously (using pager if needed)
 *
 * If there is enough place to print data on the current page, it is
 * printed synchronously. Else, a temporary buffer is allocated and
 * the data is stored in it. When the main rdline is called again, the
 * pager is flushed before parsing any other commands.
 *
 * @param rdl
 *   The rdline descriptor
 * @param fmt
 *   The format strings
 * @return
 *   Upon successful return, these functions return the number of
 *   characters printed (not including the trailing '\0' used to end
 *   output to strings). On error, a negative value is returned.
 */
int ucg_rdline_pager_printf(struct ucg_rdline *rdl, const char *fmt, ...);

/**
 * Set the callback for the pager
 *
 * If there is some data in the pager to be printed, set a callback
 * function that will be called when all the data will be printed. If
 * the pager is empty, don't do anything and return -1.
 * @param rdl
 *   The rdline descriptor
 * @return
 *   - 0 if there is some data in the pager buffer and the callback
 *     is loaded
 *   - -1 if there is no data in pager buffer (in this case the callback
 *     is not called)
 */
int ucg_rdline_pager_set_cb(struct ucg_rdline *rdl,
	ucg_rdline_pager_cb_t *cb, void *arg);
#endif

#endif /* UCG_CMD_RDLINE_H_ */
