/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <inttypes.h>
#include <unistd.h>
#include <fcntl.h>

#include "ucg_cmd_parse.h"
#include "ucg_cmd_rdline.h"
#include "ucg_cmd.h"

static void
default_valid_buffer(struct ucg_rdline *rdl, const char *line)
{
	struct ucg_cmd *cl = rdl->opaque;
	int ret;

	ret = ucg_cmd_parse(cl, line, cl);
	if (ret == UCG_CMD_PARSE_AMBIGUOUS)
		ucg_cmd_printf(cl, "Ambiguous command\n");
	else if (ret == UCG_CMD_PARSE_NOMATCH)
		ucg_cmd_printf(cl, "Bad arguments\n");
	else if (ret == UCG_CMD_PARSE_UNTERMINATED_QUOTE)
		ucg_cmd_printf(cl, "Unterminated quote\n");
}

static int
default_complete_buffer(struct ucg_rdline *rdl, const char *line,
	char *dstbuf, unsigned int dstsize)
{
	int ret;
	struct ucg_cmd *cl = rdl->opaque;
	ret = ucg_cmd_complete(cl, line, dstbuf, dstsize);
	if (ret == UCG_CMD_COMPLETE_APPEND)
		return 0;
	return -1;
}


static void
default_help(struct ucg_rdline *rdl, const char *line)
{
	struct ucg_cmd *cl = rdl->opaque;

	ucg_cmd_help(cl, line);
}

/* ---- Some rdline wrappers ---- */

void
ucg_cmd_set_prompt(struct ucg_cmd *cl, const char *prompt)
{
	snprintf(cl->prompt, sizeof(cl->prompt), "%s", prompt);
}

void
ucg_cmd_init(struct ucg_cmd *cl, ucg_cmd_ctx_t *ctx,
	const char *prompt, FILE *f_in, FILE *f_out)
{
	/* init cmd structure */
	memset(cl, 0, sizeof(struct ucg_cmd));
	cl->ctx = ctx;

	/* init embedded rdline */
	ucg_rdline_init(&cl->rdl, f_in, f_out,
		default_valid_buffer,
		default_complete_buffer,
		default_help);

	cl->rdl.opaque = cl;
	ucg_cmd_set_prompt(cl, prompt);
	ucg_rdline_newline(&cl->rdl, cl->prompt);
}

struct ucg_cmd *
ucg_cmd_new(ucg_cmd_ctx_t *ctx, const char *prompt,
	FILE *f_in, FILE *f_out)
{
	struct ucg_cmd *cl;

	cl = malloc(sizeof(struct ucg_cmd));
	if (cl == NULL)
		return NULL;

	ucg_cmd_init(cl, ctx, prompt, f_in, f_out);
	return cl;
}

struct ucg_cmd *
ucg_cmd_file_new(ucg_cmd_ctx_t *ctx, const char *prompt,
	const char *path, FILE *f_out)
{
#if UCG_CMD_HAVE_FILE
	FILE *f_in;

	f_in = fopen(path, "r");
	if (f_in == NULL)
		return NULL;
	return (ucg_cmd_new(ctx, prompt, f_in, f_out));
#else
	(void)ctx;
	(void)prompt;
	(void)path;
	(void)f_out;
	return NULL;
#endif
}

void
ucg_cmd_free(struct ucg_cmd *cl)
{
	struct ucg_rdline *rdl = &cl->rdl;

	if (rdl->f_in != stdin)
		fclose(rdl->f_in);
	if (rdl->f_out != rdl->f_in && rdl->f_out != stdout &&
		rdl->f_out != stderr)
		fclose(rdl->f_out);
	free(cl);
}

int
ucg_cmd_printf(struct ucg_cmd *cl, const char *fmt, ...)
{
	va_list ap;
	int ret;

	va_start(ap, fmt);
	ret = ucg_rdline_vprintf(&cl->rdl, fmt, ap);
	va_end(ap);

	return ret;
}

/* Push an input buffer in the command line. Typically, this function
 * is called by ucg_cmd_interact() to send the input characters to the
 * cmd process. It can also be called by a user callback function,
 * when a buffer is received from the input socket.
 *
 * The function returns the number of processed characters, or a
 * negative value on error (EOF reached or command line exited. */
int
ucg_cmd_in(struct ucg_cmd *cl, const char *buf, int size)
{
	int ret = 0;
	int i;

	for (i = 0; i < size; i++) {
		ret = ucg_rdline_char_in(&cl->rdl, buf[i]);

		if (ret == UCG_RDLINE_RES_VALIDATED &&
			cl->rdl.status == UCG_RDLINE_STOPPED)
			break;

		if (ret == UCG_RDLINE_RES_VALIDATED)
			ucg_rdline_newline(&cl->rdl, cl->prompt);
		else if (ret == UCG_RDLINE_RES_EOF)
			return -1;
		else if (ret == UCG_RDLINE_RES_EXITED)
			return -1;
	}
	return i;
}

/* Interrupt a running command line (exits from ucg_cmd_interact) */
void
ucg_cmd_quit(struct ucg_cmd *cl)
{
	ucg_rdline_quit(&cl->rdl);
}

/* loop until the user explicitelly call ucg_cmd_quit(), or if the input
 * fd reaches EOF. */
void
ucg_cmd_interact(struct ucg_cmd *cl, unsigned flags)
{
	int ret;
	char c;

	c = -1;
	while (1) {
		ret = fread(&c, 1, 1, cl->rdl.f_in);

		if (ret == 0) {
			if (flags & UCG_CMD_F_IGNORE_EOF) {
				clearerr(cl->rdl.f_in);
				continue;
			}
			break;
		}

		if (ucg_cmd_in(cl, &c, 1) < 0)
			break;
	}
}
