/*
 * Copyright (c) 2009-2015, Olivier MATZ <zer0@droids-corp.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*-
 * Copyright (c) <2010>, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * - Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in
 *   the documentation and/or other materials provided with the
 *   distribution.
 *
 * - Neither the name of Intel Corporation nor the names of its
 *   contributors may be used to endorse or promote products derived
 *   from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UCG_CMD_NO_PAGER
#define _GNU_SOURCE /* for vasprintf */
#endif

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <unistd.h>
#include <alloca.h>

#include <ucg_cirbuf.h>

#include "ucg_cmd_rdline.h"
#include "ucg_cmd_parse.h"

#ifndef UCG_CMD_NO_RDLINE_HISTORY
static void rdline_remove_old_history_item(struct ucg_rdline *rdl);
static void rdline_remove_first_history_item(struct ucg_rdline *rdl);
static unsigned int rdline_get_history_size(struct ucg_rdline *rdl);
#endif /* !UCG_CMD_NO_RDLINE_HISTORY */

#ifndef UCG_CMD_NO_PAGER
static int rdline_pager_next_page(struct ucg_rdline *rdl);
static void rdline_pager_reset(struct ucg_rdline *rdl);
#endif /* !UCG_CMD_NO_PAGER */


/* isblank() needs _XOPEN_SOURCE >= 600 || _ISOC99_SOURCE, so use our
 * own. */
static int
isblank2(char c)
{
	if (c == ' ' ||
		c == '\t' )
		return 1;
	return 0;
}

void
ucg_rdline_init(struct ucg_rdline *rdl,
	FILE *f_in, FILE *f_out,
	ucg_rdline_validate_t *validate,
	ucg_rdline_complete_t *complete,
	ucg_rdline_help_t *help)
{
	memset(rdl, 0, sizeof(*rdl));
	rdl->f_in = f_in;
	rdl->f_out = f_out;
	rdl->validate = validate;
	rdl->complete = complete;
	rdl->help = help;
	rdl->status = UCG_RDLINE_STOPPED;

	/* Disable buffering */
	setbuf(f_in, NULL);
	setbuf(f_out, NULL);

#ifndef UCG_CMD_NO_RDLINE_HISTORY
	ucg_cirbuf_init(&rdl->history, rdl->history_buf, 0,
		UCG_RDLINE_HISTORY_BUF_SIZE);
#endif /* !UCG_CMD_NO_RDLINE_HISTORY */
}

void
ucg_rdline_newline(struct ucg_rdline *rdl, const char *prompt)
{
	ucg_vt100_init(&rdl->vt100);
	ucg_cirbuf_init(&rdl->left, rdl->left_buf, 0, UCG_RDLINE_BUF_SIZE);
	ucg_cirbuf_init(&rdl->right, rdl->right_buf, 0, UCG_RDLINE_BUF_SIZE);

	/* if pointer is the same or NULL, don't copy it */
	if (prompt != NULL && prompt != rdl->prompt)
		snprintf(rdl->prompt, sizeof(rdl->prompt), "%s", prompt);

	ucg_rdline_printf(rdl, "%s", rdl->prompt);
	rdl->status = UCG_RDLINE_RUNNING;

#ifndef UCG_CMD_NO_RDLINE_HISTORY
	rdl->history_cur_line = -1;
#endif /* !UCG_CMD_NO_RDLINE_HISTORY */
}

void
ucg_rdline_stop(struct ucg_rdline *rdl)
{
	rdl->status = UCG_RDLINE_STOPPED;
}

void
ucg_rdline_quit(struct ucg_rdline *rdl)
{
	rdl->status = UCG_RDLINE_EXITED;
}

void
ucg_rdline_restart(struct ucg_rdline *rdl)
{
	rdl->status = UCG_RDLINE_RUNNING;
}

const char *
ucg_rdline_get_buffer(struct ucg_rdline *rdl)
{
	unsigned int len_l, len_r;
	ucg_cirbuf_align_left(&rdl->left);
	ucg_cirbuf_align_left(&rdl->right);

	len_l = ucg_cirbuf_get_len(&rdl->left);
	len_r = ucg_cirbuf_get_len(&rdl->right);
	memcpy(rdl->left_buf+len_l, rdl->right_buf, len_r);

	rdl->left_buf[len_l + len_r] = '\0';
	return rdl->left_buf;
}

static void
display_right_buffer(struct ucg_rdline *rdl, int force)
{
	unsigned int i;
	char tmp;

	if (!force && ucg_cirbuf_is_empty(&rdl->right))
		return;

	ucg_rdline_printf(rdl, ucg_vt100_clear_right);
	UCG_CIRBUF_FOREACH(&rdl->right, i, tmp) {
		ucg_rdline_printf(rdl, "%c", tmp);
	}
	if (!ucg_cirbuf_is_empty(&rdl->right))
		ucg_rdline_printf(rdl, ucg_vt100_multi_left,
			ucg_cirbuf_get_len(&rdl->right));
}

void
ucg_rdline_redisplay(struct ucg_rdline *rdl)
{
	unsigned int i;
	char tmp;

	ucg_rdline_printf(rdl, ucg_vt100_home);
	ucg_rdline_printf(rdl, "%s", rdl->prompt);
	UCG_CIRBUF_FOREACH(&rdl->left, i, tmp) {
		ucg_rdline_printf(rdl, "%c", tmp);
	}
	display_right_buffer(rdl, 1);
}

static int
rdline_parse_char(struct ucg_rdline *rdl, char c)
{
	unsigned int i;
	int cmd;
	char tmp;
#ifndef UCG_CMD_NO_RDLINE_HISTORY
	const char *history;
#endif

	cmd = ucg_vt100_parser(&rdl->vt100, c);
	if (cmd == UCG_VT100_NOT_COMPLETE)
		return UCG_RDLINE_RES_SUCCESS;

#ifndef UCG_CMD_NO_PAGER
	/* display asynchrounous printf if any */
	if (rdl->pager_buf != NULL) {

		/* user ask to exit pager, or last page is displayed*/
		if ((cmd == UCG_VT100_STD_CHAR && c == 'q') ||
			rdline_pager_next_page(rdl) == 0) {
			int ret;

			ret = rdl->pager_ret;
			rdline_pager_reset(rdl);
			if (rdl->pager_cb != NULL) {
				rdl->pager_cb(rdl, rdl->pager_arg);
				rdl->pager_cb = NULL;
			}
			/* maybe the pager was reloaded in the
			 * callback */
			if (rdl->pager_buf != NULL)
				return UCG_RDLINE_RES_SUCCESS;

			/* else, redisplay prompt and return the saved status */
			ucg_rdline_redisplay(rdl);
			return ret;
		}

		/* Some pages remain, lines were displayed in
		 * rdline_pager_next_page() */
		return UCG_RDLINE_RES_SUCCESS;
	}
#endif

	/* process control chars */
	if (cmd != UCG_VT100_STD_CHAR) {
		switch (cmd) {
		case UCG_CMD_KEY_CTRL_B:
		case UCG_CMD_KEY_LEFT_ARR:
			if (ucg_cirbuf_is_empty(&rdl->left))
				break;
			tmp = ucg_cirbuf_get_tail(&rdl->left);
			ucg_cirbuf_del_tail(&rdl->left);
			ucg_cirbuf_add_head(&rdl->right, tmp);
			ucg_rdline_printf(rdl, ucg_vt100_left_arr);
			break;

		case UCG_CMD_KEY_CTRL_F:
		case UCG_CMD_KEY_RIGHT_ARR:
			if (ucg_cirbuf_is_empty(&rdl->right))
				break;
			tmp = ucg_cirbuf_get_head(&rdl->right);
			ucg_cirbuf_del_head(&rdl->right);
			ucg_cirbuf_add_tail(&rdl->left, tmp);
			ucg_rdline_printf(rdl, ucg_vt100_right_arr);
			break;

		case UCG_CMD_KEY_WLEFT:
			while (! ucg_cirbuf_is_empty(&rdl->left) &&
				(tmp = ucg_cirbuf_get_tail(&rdl->left)) &&
				isblank2(tmp)) {
				ucg_rdline_printf(rdl, ucg_vt100_left_arr);
				ucg_cirbuf_del_tail(&rdl->left);
				ucg_cirbuf_add_head(&rdl->right, tmp);
			}
			while (! ucg_cirbuf_is_empty(&rdl->left) &&
				(tmp = ucg_cirbuf_get_tail(&rdl->left)) &&
				!isblank2(tmp)) {
				ucg_rdline_printf(rdl, ucg_vt100_left_arr);
				ucg_cirbuf_del_tail(&rdl->left);
				ucg_cirbuf_add_head(&rdl->right, tmp);
			}
			break;

		case UCG_CMD_KEY_WRIGHT:
			while (! ucg_cirbuf_is_empty(&rdl->right) &&
				(tmp = ucg_cirbuf_get_head(&rdl->right)) &&
				isblank2(tmp)) {
				ucg_rdline_printf(rdl, ucg_vt100_right_arr);
				ucg_cirbuf_del_head(&rdl->right);
				ucg_cirbuf_add_tail(&rdl->left, tmp);
			}
			while (! ucg_cirbuf_is_empty(&rdl->right) &&
				(tmp = ucg_cirbuf_get_head(&rdl->right)) &&
				!isblank2(tmp)) {
				ucg_rdline_printf(rdl, ucg_vt100_right_arr);
				ucg_cirbuf_del_head(&rdl->right);
				ucg_cirbuf_add_tail(&rdl->left, tmp);
			}
			break;

		case UCG_CMD_KEY_BKSPACE:
			if(!ucg_cirbuf_del_tail_safe(&rdl->left)) {
				ucg_rdline_printf(rdl, ucg_vt100_bs);
				display_right_buffer(rdl, 1);
			}
			break;

		case UCG_CMD_KEY_META_BKSPACE:
		case UCG_CMD_KEY_CTRL_W:
			while (! ucg_cirbuf_is_empty(&rdl->left) &&
				isblank2(ucg_cirbuf_get_tail(&rdl->left))) {
				ucg_rdline_printf(rdl, ucg_vt100_bs);
				ucg_cirbuf_del_tail(&rdl->left);
			}
			while (! ucg_cirbuf_is_empty(&rdl->left) &&
				!isblank2(ucg_cirbuf_get_tail(&rdl->left))) {
				ucg_rdline_printf(rdl, ucg_vt100_bs);
				ucg_cirbuf_del_tail(&rdl->left);
			}
			display_right_buffer(rdl, 1);
			break;

		case UCG_CMD_KEY_META_D:
			while (! ucg_cirbuf_is_empty(&rdl->right) &&
				isblank2(ucg_cirbuf_get_head(&rdl->right)))
				ucg_cirbuf_del_head(&rdl->right);
			while (! ucg_cirbuf_is_empty(&rdl->right) &&
				!isblank2(ucg_cirbuf_get_head(&rdl->right)))
				ucg_cirbuf_del_head(&rdl->right);
			display_right_buffer(rdl, 1);
			break;

		case UCG_CMD_KEY_SUPPR:
		case UCG_CMD_KEY_CTRL_D:
			if (cmd == UCG_CMD_KEY_CTRL_D &&
				ucg_cirbuf_is_empty(&rdl->left) &&
				ucg_cirbuf_is_empty(&rdl->right)) {
				return UCG_RDLINE_RES_EOF;
			}
			if (!ucg_cirbuf_del_head_safe(&rdl->right)) {
				display_right_buffer(rdl, 1);
			}
			break;

		case UCG_CMD_KEY_CTRL_A:
			if (ucg_cirbuf_is_empty(&rdl->left))
				break;
			ucg_rdline_printf(rdl, ucg_vt100_multi_left,
				ucg_cirbuf_get_len(&rdl->left));
			while (! ucg_cirbuf_is_empty(&rdl->left)) {
				tmp = ucg_cirbuf_get_tail(&rdl->left);
				ucg_cirbuf_del_tail(&rdl->left);
				ucg_cirbuf_add_head(&rdl->right, tmp);
			}
			break;

		case UCG_CMD_KEY_CTRL_E:
			if (ucg_cirbuf_is_empty(&rdl->right))
				break;
			ucg_rdline_printf(rdl, ucg_vt100_multi_right,
				ucg_cirbuf_get_len(&rdl->right));
			while (! ucg_cirbuf_is_empty(&rdl->right)) {
				tmp = ucg_cirbuf_get_head(&rdl->right);
				ucg_cirbuf_del_head(&rdl->right);
				ucg_cirbuf_add_tail(&rdl->left, tmp);
			}
			break;

#ifndef UCG_CMD_NO_RDLINE_KILL_BUF
		case UCG_CMD_KEY_CTRL_K:
			ucg_cirbuf_get_buf_head(&rdl->right, rdl->kill_buf,
				UCG_RDLINE_BUF_SIZE);
			rdl->kill_size = ucg_cirbuf_get_len(&rdl->right);
			ucg_cirbuf_del_buf_head(&rdl->right, rdl->kill_size);
			ucg_rdline_printf(rdl, ucg_vt100_clear_right);
			break;

		case UCG_CMD_KEY_CTRL_Y:
			i=0;
			while (ucg_cirbuf_get_len(&rdl->right) +
				ucg_cirbuf_get_len(&rdl->left) <
				UCG_RDLINE_BUF_SIZE &&
				i < rdl->kill_size) {
				ucg_cirbuf_add_tail(&rdl->left, rdl->kill_buf[i]);
				ucg_rdline_printf(rdl, "%c", rdl->kill_buf[i]);
				i++;
			}
			display_right_buffer(rdl, 0);
			break;
#endif /* !UCG_CMD_NO_RDLINE_KILL_BUF */

		case UCG_CMD_KEY_CTRL_C:
			ucg_rdline_printf(rdl, "\r\n");
			ucg_rdline_newline(rdl, rdl->prompt);
			break;

		case UCG_CMD_KEY_CTRL_L:
			ucg_rdline_redisplay(rdl);
			break;

		case UCG_CMD_KEY_HELP: {
			if (rdl->help == NULL)
				break;

			ucg_cirbuf_align_left(&rdl->left);
			rdl->left_buf[ucg_cirbuf_get_len(&rdl->left)] = '\0';
			ucg_rdline_printf(rdl, "\r\n");
			rdl->help(rdl, rdl->left_buf);
#ifndef UCG_CMD_NO_PAGER
			if (rdl->pager_buf != NULL)
				return UCG_RDLINE_RES_SUCCESS;
			else
				rdline_pager_reset(rdl);
#endif
			ucg_rdline_redisplay(rdl);
			break;
		}

		case UCG_CMD_KEY_TAB: {
			char tmp_buf[UCG_CMD_MAX_TOKEN_SIZE];
			int ret;
			unsigned int tmp_size;

			if (rdl->complete == NULL)
				break;

			ucg_cirbuf_align_left(&rdl->left);
			rdl->left_buf[ucg_cirbuf_get_len(&rdl->left)] = '\0';

			/* try to complete the  complete() */
			ret = rdl->complete(rdl, rdl->left_buf,
				tmp_buf, sizeof(tmp_buf));

			/* no completion or error */
			if (ret < 0) {
				ucg_cirbuf_align_left(&rdl->left);
				rdl->left_buf[ucg_cirbuf_get_len(&rdl->left)] = '\0';
				ucg_rdline_printf(rdl, "\r\n");
				rdl->help(rdl, rdl->left_buf);
#ifndef UCG_CMD_NO_PAGER
				if (rdl->pager_buf != NULL)
					return UCG_RDLINE_RES_SUCCESS;
				else
					rdline_pager_reset(rdl);
#endif
				ucg_rdline_redisplay(rdl);
				break;
			}

			tmp_size = strlen(tmp_buf);
			/* add chars */
			i = 0;
			while(ucg_cirbuf_get_len(&rdl->right) +
				ucg_cirbuf_get_len(&rdl->left) <
				UCG_RDLINE_BUF_SIZE &&
				i < tmp_size) {
				ucg_cirbuf_add_tail(&rdl->left, tmp_buf[i]);
				ucg_rdline_printf(rdl, "%c", tmp_buf[i]);
				i++;
			}
			display_right_buffer(rdl, 1);
			break;
		}

		case UCG_CMD_KEY_RETURN:
		case UCG_CMD_KEY_RETURN2: {
			char tmp;
			while (!ucg_cirbuf_is_empty(&rdl->right) &&
				(tmp = ucg_cirbuf_get_head(&rdl->right))) {
				ucg_cirbuf_del_head(&rdl->right);
				ucg_cirbuf_add_tail(&rdl->left, tmp);
			}
			ucg_cirbuf_align_left(&rdl->left);
			rdl->left_buf[ucg_cirbuf_get_len(&rdl->left)] = '\0';
			ucg_rdline_printf(rdl, "\r\n");
#ifndef UCG_CMD_NO_RDLINE_HISTORY
			if (rdl->history_cur_line != -1)
				rdline_remove_first_history_item(rdl);
#endif

			if (rdl->validate)
				rdl->validate(rdl, rdl->left_buf);
#ifndef UCG_CMD_NO_PAGER
			/* user may have stopped rdline */
			if (rdl->status == UCG_RDLINE_EXITED) {
				rdline_pager_reset(rdl);
				return UCG_RDLINE_RES_EXITED;
			}
			/* there is something in pager buffer, save
			 * return value that will be return once
			 * paging is finished */
			if (rdl->pager_buf != NULL) {
				rdl->pager_ret = UCG_RDLINE_RES_VALIDATED;
				return UCG_RDLINE_RES_SUCCESS;
			}

			rdline_pager_reset(rdl);
#else
			if (rdl->status == UCG_RDLINE_EXITED)
				return UCG_RDLINE_RES_EXITED;
#endif
			return UCG_RDLINE_RES_VALIDATED;
		}
#ifndef UCG_CMD_NO_RDLINE_HISTORY
		case UCG_CMD_KEY_UP_ARR:
		case UCG_CMD_KEY_CTRL_P:
			if (rdl->history_cur_line == 0) {
				rdline_remove_first_history_item(rdl);
			}
			if (rdl->history_cur_line <= 0) {
				ucg_rdline_add_history(rdl,
					ucg_rdline_get_buffer(rdl));
				rdl->history_cur_line = 0;
			}

			history = ucg_rdline_get_history_item(rdl,
				rdl->history_cur_line + 1);
			if (history == NULL)
				break;

			rdl->history_cur_line++;
			ucg_vt100_init(&rdl->vt100);
			ucg_cirbuf_init(&rdl->left, rdl->left_buf, 0,
				UCG_RDLINE_BUF_SIZE);
			ucg_cirbuf_init(&rdl->right, rdl->right_buf, 0,
				UCG_RDLINE_BUF_SIZE);
			ucg_cirbuf_add_buf_tail(&rdl->left, history,
				strlen(history));
			ucg_rdline_redisplay(rdl);
			break;

		case UCG_CMD_KEY_DOWN_ARR:
		case UCG_CMD_KEY_CTRL_N:
			if (rdl->history_cur_line - 1 < 0)
				break;

			rdl->history_cur_line--;
			history = ucg_rdline_get_history_item(rdl,
				rdl->history_cur_line);
			if (history == NULL)
				break;

			ucg_vt100_init(&rdl->vt100);
			ucg_cirbuf_init(&rdl->left, rdl->left_buf, 0,
				UCG_RDLINE_BUF_SIZE);
			ucg_cirbuf_init(&rdl->right, rdl->right_buf, 0,
				UCG_RDLINE_BUF_SIZE);
			ucg_cirbuf_add_buf_tail(&rdl->left, history,
				strlen(history));
			ucg_rdline_redisplay(rdl);

			break;
#endif /* !UCG_CMD_NO_RDLINE_HISTORY */

		default:
			break;
		}

		return UCG_RDLINE_RES_SUCCESS;
	}

	if (!isprint((int)c))
		return UCG_RDLINE_RES_SUCCESS;

	/* standard chars */
	if (ucg_cirbuf_get_len(&rdl->left) +
		ucg_cirbuf_get_len(&rdl->right) >= UCG_RDLINE_BUF_SIZE)
		return UCG_RDLINE_RES_SUCCESS;

	if (ucg_cirbuf_add_tail_safe(&rdl->left, c))
		return UCG_RDLINE_RES_SUCCESS;

	ucg_rdline_printf(rdl, "%c", c);
	display_right_buffer(rdl, 0);

	return UCG_RDLINE_RES_SUCCESS;
}

int
ucg_rdline_char_in(struct ucg_rdline *rdl, char c)
{
	int ret, same = 0;
	const char *history, *buffer;

	if (rdl->status == UCG_RDLINE_EXITED)
		return UCG_RDLINE_RES_EXITED;
	if (rdl->status != UCG_RDLINE_RUNNING)
		return UCG_RDLINE_RES_NOT_RUNNING;

	ret = rdline_parse_char(rdl, c);

	/* add line to history */
	if (ret == UCG_RDLINE_RES_VALIDATED) {
		buffer = ucg_rdline_get_buffer(rdl);
		history = ucg_rdline_get_history_item(rdl, 0);
		if (history)
			same = !strcmp(buffer, history);

		if (strlen(buffer) >= 1 && same == 0)
			ucg_rdline_add_history(rdl, buffer);
	}

	return ret;
}

int
ucg_rdline(struct ucg_rdline *rdl, const char *prompt, unsigned flags)
{
	char c;
	int ret = UCG_RDLINE_RES_NOT_RUNNING;

	ucg_rdline_newline(rdl, prompt);
	while (1) {
		if (fread(&c, 1, 1, rdl->f_in) == 0) {
			if (flags & UCG_RDLINE_F_IGNORE_EOF) {
				clearerr(rdl->f_in);
				continue;
			}
			break;
		}
		ret = ucg_rdline_char_in(rdl, c);
		if (ret != UCG_RDLINE_RES_SUCCESS)
			break;
	}

	return ret;
}

/* HISTORY */

#ifndef UCG_CMD_NO_RDLINE_HISTORY
static void
rdline_remove_old_history_item(struct ucg_rdline *rdl)
{
	char tmp;

	while (! ucg_cirbuf_is_empty(&rdl->history) ) {
		tmp = ucg_cirbuf_get_head(&rdl->history);
		ucg_cirbuf_del_head(&rdl->history);
		if (!tmp)
			break;
	}
}

static void
rdline_remove_first_history_item(struct ucg_rdline *rdl)
{
	char tmp;

	if ( ucg_cirbuf_is_empty(&rdl->history) ) {
		return;
	}
	else {
		ucg_cirbuf_del_tail(&rdl->history);
	}

	while (! ucg_cirbuf_is_empty(&rdl->history) ) {
		tmp = ucg_cirbuf_get_tail(&rdl->history);
		if (!tmp)
			break;
		ucg_cirbuf_del_tail(&rdl->history);
	}
}

static unsigned int
rdline_get_history_size(struct ucg_rdline *rdl)
{
	unsigned int i, tmp, ret=0;

	UCG_CIRBUF_FOREACH(&rdl->history, i, tmp) {
		if (tmp == 0)
			ret ++;
	}

	return ret;
}

const char *
ucg_rdline_get_history_item(struct ucg_rdline *rdl, unsigned int idx)
{
	unsigned int len, i, tmp;

	len = rdline_get_history_size(rdl);
	if (idx >= len)
		return NULL;

	ucg_cirbuf_align_left(&rdl->history);

	UCG_CIRBUF_FOREACH(&rdl->history, i, tmp) {
		if (idx == len - 1) {
			return rdl->history_buf + i;
		}
		if (tmp == 0)
			len --;
	}

	return NULL;
}

int
ucg_rdline_add_history(struct ucg_rdline *rdl, const char *buf)
{
	unsigned int len;

	len = strlen(buf);
	if (len >= UCG_RDLINE_HISTORY_BUF_SIZE)
		return -1;

	while (len >= ucg_cirbuf_get_freelen(&rdl->history)) {
		rdline_remove_old_history_item(rdl);
	}

	ucg_cirbuf_add_buf_tail(&rdl->history, buf, len);
	ucg_cirbuf_add_tail(&rdl->history, 0);

	return 0;
}

void
ucg_rdline_clear_history(struct ucg_rdline *rdl)
{
	ucg_cirbuf_init(&rdl->history, rdl->history_buf, 0,
		UCG_RDLINE_HISTORY_BUF_SIZE);
}

#else /* !UCG_CMD_NO_RDLINE_HISTORY */

int ucg_rdline_add_history(struct ucg_rdline *rdl, const char *buf)
{
	return -1;
}

void ucg_rdline_clear_history(struct ucg_rdline *rdl)
{
	return;
}

char *ucg_rdline_get_history_item(struct ucg_rdline *rdl, unsigned int i)
{
	return NULL;
}


#endif /* !UCG_CMD_NO_RDLINE_HISTORY */


ssize_t
ucg_rdline_write(struct ucg_rdline *rdl, void *buf, size_t count)
{
	return fwrite(buf, 1, count, rdl->f_out);
}

int
ucg_rdline_vprintf(struct ucg_rdline *rdl, const char *fmt, va_list ap)
{
	if (rdl->f_out == NULL)
		return -1;

	return vfprintf(rdl->f_out, fmt, ap);
}

int
ucg_rdline_printf(struct ucg_rdline *rdl, const char *fmt, ...)
{
	va_list ap;
	int ret;

	va_start(ap, fmt);
	ret = ucg_rdline_vprintf(rdl, fmt, ap);
	va_end(ap);

	return ret;
}

#ifndef UCG_CMD_NO_PAGER
/* reset pager state */
void
rdline_pager_reset(struct ucg_rdline *rdl)
{
	if (rdl->pager_buf) {
		free(rdl->pager_buf);
		rdl->pager_buf = NULL;
	}
	rdl->pager_lines = 0;
	rdl->pager_len = 0;
	rdl->pager_off = 0;
	rdl->pager_cb = NULL;
	rdl->pager_arg = NULL;
	rdl->pager_ret = UCG_RDLINE_RES_SUCCESS;
}

/* Return the offset of the i-th occurence of char c in string s. If
 * there is less than i occurences, return -1 and fill i with the
 * count. */
static int
strnchr(const char *s, char c, int *i)
{
	int n = 0;
	const char *orig = s;

	while (*s) {
		if (*s == c)
			n++;
		if (*i == n)
			return s - orig;
		s++;
	}
	*i = n;
	return -1;
}

/* display a page of data from pager, return 0 if all is displayed */
static int
rdline_pager_next_page(struct ucg_rdline *rdl)
{
	int lines = UCG_RDLINE_MAX_LINES;
	int displen;
	char *s;

	s = rdl->pager_buf;
	if (s == NULL)
		return 0;

	ucg_rdline_printf(rdl, ucg_vt100_home);
	ucg_rdline_printf(rdl, ucg_vt100_clear_right);

	s += rdl->pager_off;

	/* we know that s is 0-terminated */
	displen = strnchr(s, '\n', &lines);
	rdl->pager_lines = lines;

	/* we can display all the data */
	if (displen == -1) {
		fwrite(s, 1, rdl->pager_len, rdl->f_out);
		free(rdl->pager_buf);
		rdl->pager_buf = NULL;
		return 0;
	}

	displen = displen + 1; /* include \n */
	fwrite(s, 1, displen, rdl->f_out);
	rdl->pager_off += displen;
	rdl->pager_len -= displen;

	ucg_rdline_printf(rdl, "--- press a key to continue ---");
	return -1;
}

/* push data in pager */
ssize_t
ucg_rdline_pager_write(struct ucg_rdline *rdl, void *buf, size_t len)
{
	char *s = buf;

	/* display as many lines as we can */
	if (rdl->pager_lines < UCG_RDLINE_MAX_LINES) {
		int lines = UCG_RDLINE_MAX_LINES - rdl->pager_lines;
		int displen;

		/* we know that s is 0-terminated */
		displen = strnchr(s, '\n', &lines);
		rdl->pager_lines += lines;

		/* we can display all the data */
		if (displen == -1) {
			fwrite(s, 1, len, rdl->f_out);
			return 0;
		}
		displen = displen + 1; /* include \n */
		fwrite(s, 1, displen, rdl->f_out);
		s += displen;
		len -= displen;
	}

	if (rdl->pager_buf == NULL) {
		ucg_rdline_printf(rdl, "--- press a key to continue ---");
	}
	rdl->pager_buf = realloc(rdl->pager_buf, rdl->pager_len + len);
	if (rdl->pager_buf == NULL) {
		rdline_pager_reset(rdl);
		return -1;
	}

	memcpy(rdl->pager_buf + rdl->pager_len, s, len);
	rdl->pager_len += len;
	return 0;
}

/* Print data asynchronously (using pager if needed) */
int
ucg_rdline_pager_printf(struct ucg_rdline *rdl, const char *fmt, ...)
{
	int n;
	char *buf = NULL;
	va_list ap;

	if (rdl->f_out == NULL)
		return -1;

	va_start(ap, fmt);
	n = vasprintf(&buf, fmt, ap);
	va_end(ap);

	if (n > 0)
		ucg_rdline_pager_write(rdl, buf, n);
	free(buf);
	return n;
}

int ucg_rdline_pager_set_cb(struct ucg_rdline *rdl,
	ucg_rdline_pager_cb_t *cb, void *arg)
{
	if (rdl->pager_buf == NULL)
		return -1;

	rdl->pager_cb = cb;
	rdl->pager_arg = arg;
	return 0;
}
#endif /* !UCG_CMD_NO_PAGER */
