/*
 * Copyright (c) 2005-2015, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UCG_UART_H_
#define UCG_UART_H_

#include <ucg_cirbuf.h>

/**
 * UART interface, using interrupts. Supports 5 to 8 bits characters
 * (not 9 bits).
 */

/**
 * Callback type of tx/rx event.
 */
typedef void (ucg_uart_cb_t)(char);

struct ucg_uart;

/**
 * Configuration of the uart
 */
struct ucg_uart_config {
	uint8_t enable     : 1, /**< enable or disable the uart */
		parity     : 2, /**< none, odd or even */
		stop_bits  : 1, /**< 1 or 2 bits at the end of the frame */
		reserved   : 4; /**< nothing for now */
	uint8_t nbits;          /**< number of bits in frame, 5, 6, 7 or 8 */
	uint32_t baudrate;      /**< speed of uart */
};

/**
 * Per-arch uart driver operations
 */
struct ucg_uart_driver_ops {
	void (*disable_tx_irq)(struct ucg_uart *uart);  /**< Disable tx intrpt */
	void (*enable_tx_irq)(struct ucg_uart *uart);   /**< Enable tx intrpt */
	uint8_t (*tx_ready)(struct ucg_uart *uart);     /**< Test if tx register empty */
	uint8_t (*rx_ready)(struct ucg_uart *uart);     /**< Test if rx register full */
	void (*set_udr)(struct ucg_uart *uart, char c); /**< Set uart data register */
	char (*get_udr)(struct ucg_uart *uart);         /**< Get uart data register */
	int (*set_conf)(struct ucg_uart *uart,
		const struct ucg_uart_config *conf);    /**< Set configuration */
};

/**
 * Generic Uart structure
 */
struct ucg_uart {
	struct ucg_uart_config conf;            /**< Current configuration */
	ucg_uart_cb_t *rx_cb;                   /**< RX callback */
	ucg_uart_cb_t *tx_cb;                   /**< TX callback */
	struct ucg_cirbuf *rx_fifo;             /**< RX fifo */
	struct ucg_cirbuf *tx_fifo;             /**< TX fifo */
	const struct ucg_uart_driver_ops *ops;  /**< Pointer to arch uart ops */
	void *driver_data;                      /**< Opaque arch uart data */
};

/**
 * Uart wait directive, used in rx/tx functions.
 */
enum ucg_uart_wait {
	WAIT,     /**< Loop until operation can be done */
	NOWAIT,   /**< Return if operation can't be done now */
};

/**
 * Handle TX register empty interrupt
 *
 * @param uart
 *   The uart structure.
 */
void ucg_uart_tx_intr(struct ucg_uart *uart);

/**
 * Handle RX register empty interrupt
 *
 * @param uart
 *   The uart structure.
 */
void ucg_uart_rx_intr(struct ucg_uart *uart);

/**
 * Initialize uart
 *
 * Initialize the per-arch and generic uart structure, the rx and tx
 * fifos, and configure the uart with the default configuration.
 *
 * @param uart
 *   The uart structure.
 * @param ops
 *   A pointer to the per-arch uart operations.
 * @param driver_data
 *   A pointer to a per-arch uart data structure (ex: struct ucg_stm32_uart *)
 * @param rx_cbuf
 *   A pointer to an uninitialized cirbuf, used for rx fifo
 * @param rx_buf
 *   The buffer used by the rx fifo
 * @param rx_bufsize
 *   The length of rx_buf in bytes
 * @param tx_cbuf
 *   A pointer to an uninitialized cirbuf, used for tx fifo
 * @param tx_buf
 *   The buffer used by the tx fifo
 * @param tx_bufsize
 *   The length of tx_buf in bytes
 * @return
 *   0 on success, negative on error.
 */
int ucg_uart_init(struct ucg_uart *uart,
	const struct ucg_uart_driver_ops *ops, void *driver_data,
	struct ucg_cirbuf *rx_cbuf, char *rx_buf, unsigned rx_bufsize,
	struct ucg_cirbuf *tx_cbuf, char *tx_buf, unsigned tx_bufsize);

/**
 * Configure the uart with the given configuration.
 *
 * The function ucg_uart_init() must have been called first.
 *
 * @param uart
 *   The uart structure.
 * @param conf
 *   The configuration to be applied.
 * @return
 *   0 on success, negative on error.
 */
int ucg_uart_setconf(struct ucg_uart *uart, const struct ucg_uart_config *conf);

/**
 * Get the current configuration of the uart.
 *
 * @param uart
 *   The uart structure.
 * @param conf
 *   The pointer to be filled with the current running configuration.
 */
void ucg_uart_getconf(struct ucg_uart *uart, struct ucg_uart_config *conf);

/**
 * Receive a character.
 *
 * Receive the next character, taken from the fifo if any. If NOWAIT is
 * given, the function returns an error if the fifo is empty. If WAIT is
 * given, the function loops until a new character can be returned.
 *
 * The function can be called from an interrupt, but it should be used
 * with care if WAIT is given, as it will loop until a character is
 * received.
 *
 * @param uart
 *   The uart structure.
 * @param wait
 *   The wait directive (WAIT or NOWAIT)
 * @return
 *   Return the character on success (>= 0), or a negative value on error.
 */
int ucg_uart_recv(struct ucg_uart *uart, enum ucg_uart_wait wait);

/**
 * Send a character.
 *
 * Put a new character in the TX fifo. If NOWAIT is given, the function
 * returns an error if the fifo is full. If WAIT is given, the function
 * loops until some room is available in the tx fifo.
 *
 * The function can be called from an interrupt, but it should be used
 * with care if WAIT is given, as it will wait that the current
 * transmission is finished before returning.
 *
 * @param uart
 *   The uart structure.
 * @param wait
 *   The wait directive (WAIT or NOWAIT)
 * @return
 *   Return the transmitted character on success (>= 0), or a negative
 *   value on error.
 */
int ucg_uart_send(struct ucg_uart *uart, char c, enum ucg_uart_wait wait);

/**
 * Flush the TX fifo
 *
 * Loop until the uart TX fifo is empty. The function can be called from
 * an interrupt, but it should be used with care as it will wait that the
 * transmission of all bytes of the fifo is finished before returning.
 *
 * @param uart
 *   The uart structure.
 */
void ucg_uart_flush(struct ucg_uart *uart);

/**
 * Register a TX callback function
 *
 * Register a function that will be called after a character is written
 * in the UART dta register. This function is called with interrupts
 * locked, and the user is free to access or modify the tx fifo if
 * required. Note that the character that has just been transmitted is
 * not in the tx fifo when the callback is invoked.
 *
 * @param uart
 *   The uart structure.
 * @param f
 *   The function to call on tx event, can be NULL to disable.
 * @param arg
 *   The opaque argument given to the function.
 */
void ucg_uart_register_tx_cb(struct ucg_uart *uart, ucg_uart_cb_t *f);

/**
 * Register a TX callback function
 *
 * Register a function that will be called after a character is read
 * from the UART dta register. This function is called with interrupts
 * locked, and the user is free to access or modify the rx fifo if
 * required. Note that the character that has just been received is
 * present in the rx fifo when the callback is invoked.
 *
 * @param uart
 *   The uart structure.
 * @param f
 *   The function to call on tx event, can be NULL to disable.
 * @param arg
 *   The opaque argument given to the function.
 */
void ucg_uart_register_rx_cb(struct ucg_uart *uart, ucg_uart_cb_t *f);

#endif /* UCG_UART_H_ */
