/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>

#include <cfzy_log.h>
#include <cfzy_list.h>
#include <cfzy_conftree.h>
#include <cfzy_dotconfig.h>
#include <cfzy_c_hdr.h>

#define LOG(level, fmt, args...)				\
	CFZY_LOG("cfzy-basic", level, fmt, ##args)

/* Parsed arguments */
struct cfzy_arguments {
	char *input_conftree;
	char *input_dotconfig;
	char *output_dotconfig;
	char *output_c_hdr;
	char *output_multi_c_hdr;
};

static void
usage(const char *prog, int code)
{
	fprintf(stderr,
		"\nUsage : \n");
	fprintf(stderr,
		"  %s [options]\n", prog);
	fprintf(stderr,
		"-h, --help\n"
		"            show help\n");
	fprintf(stderr,
		"-d, --debug-level\n"
		"            set debug level (0: no log, 4 noisy)\n");
	fprintf(stderr,
		"-c, --input-conftree\n"
		"            path to input configuration tree file "
		"(mandatory)\n");
	fprintf(stderr,
		"-i, --input-dotconfig\n"
		"            path to input dotconfig file\n");
	fprintf(stderr,
		"-o, --output-dotconfig\n"
		"            path to output dotconfig file\n");
	fprintf(stderr,
		"-O, --output-c-hdr\n"
		"            path to output C header file\n");
	fprintf(stderr,
		"-m, --output-multi-c-hdr\n"
		"            path to output multiple C header directory\n");
	exit(code);
}

static void
free_args(struct cfzy_arguments *cfzy_args)
{
	if (cfzy_args->input_conftree != NULL)
		free(cfzy_args->input_conftree);
	if (cfzy_args->input_dotconfig != NULL)
		free(cfzy_args->input_dotconfig);
	if (cfzy_args->output_dotconfig != NULL)
		free(cfzy_args->output_dotconfig);
	if (cfzy_args->output_c_hdr != NULL)
		free(cfzy_args->output_c_hdr);
	if (cfzy_args->output_multi_c_hdr != NULL)
		free(cfzy_args->output_multi_c_hdr);
}

static void
parse_args(struct cfzy_arguments *cfzy_args, int argc, char **argv)
{
	int ch;
	const char * prog = argv[0];
	int option_index;
	int level;
	static struct option lgopts[] = {
		{"help", 0, 0, 'h'},
		{"debug-level", 0, 0, 'd'},
		{"input-conftree", 0, 0, 'c'},
		{"input-dotconfig", 0, 0, 'i'},
		{"output-dotconfig", 0, 0, 'o'},
		{"output-c-hdr", 0, 0, 'O'},
		{"output-multi-c-hdr", 0, 0, 'm'},
		{NULL, 0, 0, 0}
	};

	while ((ch = getopt_long(argc, argv,
				 "h"  /* help */
				 "d:"  /* debug-level */
				 "c:"  /* input-conftree */
				 "i:"  /* input-dotconfig */
				 "o:"  /* output-dotconfig */
				 "O:"  /* output-c-hdr */
				 "m:"  /* output-multi-c-hdr */
				 ,
			    lgopts, &option_index)) != -1) {
		switch (ch) {
		case 'h': /* help */
			free_args(cfzy_args);
			usage(prog, 0); /* will exit */
			break;

		case 'd': /* debug-level */
			level = atoi(optarg);
			if (level < 0 || level > CFZY_LOG_DEBUG) {
				LOG(ERR, "invalid log level\n");
				free_args(cfzy_args);
				cfzy_log_exit();
				exit(1);
			}
			cfzy_log_set_default_level(level);
			break;

		case 'c': /* input-conftree */
			cfzy_args->input_conftree = strdup(optarg);
			if (cfzy_args->input_conftree == NULL) {
				LOG(ERR, "Cannot alloc input conftree str\n");
				free_args(cfzy_args);
				cfzy_log_exit();
				exit(1);
			}
			break;

		case 'i': /* input-dotconfig */
			cfzy_args->input_dotconfig = strdup(optarg);
			if (cfzy_args->input_dotconfig == NULL) {
				LOG(ERR, "Cannot alloc input dotconfig str\n");
				free_args(cfzy_args);
				cfzy_log_exit();
				exit(1);
			}
			break;

		case 'o': /* output-dotconfig */
			cfzy_args->output_dotconfig = strdup(optarg);
			if (cfzy_args->output_dotconfig == NULL) {
				LOG(ERR, "Cannot alloc output dotconfig str\n");
				free_args(cfzy_args);
				cfzy_log_exit();
				exit(1);
			}
			break;

		case 'O': /* output-c-hdr */
			cfzy_args->output_c_hdr = strdup(optarg);
			if (cfzy_args->output_c_hdr == NULL) {
				LOG(ERR, "Cannot alloc output c_hdr str\n");
				free_args(cfzy_args);
				cfzy_log_exit();
				exit(1);
			}
			break;

		case 'm': /* output-multi-c-hdr */
			cfzy_args->output_multi_c_hdr = strdup(optarg);
			if (cfzy_args->output_multi_c_hdr == NULL) {
				LOG(ERR, "Cannot alloc output multi_c_hdr str\n");
				free_args(cfzy_args);
				cfzy_log_exit();
				exit(1);
			}
			break;

		default:
			LOG(ERR, "invalid option\n");
			free_args(cfzy_args);
			cfzy_log_exit();
			usage(prog, 1); /* will exit */
		}
	}

	/* conftree file is mandatory */
	if (cfzy_args->input_conftree == NULL) {
		LOG(ERR, "input conftree is mandatory\n");
		free_args(cfzy_args);
		cfzy_log_exit();
		usage(prog, 1); /* will exit */
	}

	/* check that at least one output argument is given */
	if (cfzy_args->output_dotconfig == NULL &&
	    cfzy_args->output_c_hdr == NULL &&
	    cfzy_args->output_multi_c_hdr == NULL) {
		LOG(ERR, "at least one output argument is needed\n");
		free_args(cfzy_args);
		cfzy_log_exit();
		usage(prog, 1); /* will exit */
	}

	argc -= optind;
	argv += optind;
	/* XXX parse mode */
}

/*
 * configuration tree
 *
 * -i input .config:
 *   default is none
 *
 * -o output .config
 *   defaut is none
 *
 * ?? output autoconf.h
 *
 * output config/xxx.h
 *
 * -v verbose
 *
 * oldconfig:
 *   prompt for options that have no user value
 *
 * silentoldconfig
 *   set all options that have no user value to default
 *
 * config
 *    ?
 *
 * defconfig
 *    ?
 *
 * randconfig
 *    set all options to a random value (only for booleans, other
 *    values are set to default)
 *
 * allyesconfig
 *    set all options to y (booleans only)
 *
 * allnoconfig
 *    set all options to n (booleans only)
 *
 * - ouvrir le fichier conftree
 * - si -i, ouvrir .config
 * - effectue les operations
 * - si -o et/ou ... output
 */

static int
do_configuration(struct cfzy_arguments *cfzy_args)
{
	struct cfzy_confnode *n;
	struct cfzy_conftree *conftree;
	char buf[BUFSIZ];
	char *s;
	int ret;

	/* parse configuration tree */
	conftree = cfzy_conftree_new(cfzy_args->input_conftree);
	if (conftree == NULL) {
		LOG(ERR, "Cannot parse configuration tree\n");
		return -1;
	}

	/* parse .config if provided */
	if (cfzy_args->input_dotconfig != NULL &&
	    cfzy_dotconfig_read(conftree, cfzy_args->input_dotconfig) < 0) {
		cfzy_conftree_free(conftree);
		LOG(ERR, "Cannot parse dotconfig <%s>\n",
		    cfzy_args->input_dotconfig);
		return -1;
	}

#if 0
	/* evaluate effective values of conftree */
	if (cfzy_conftree_evaluate(conftree) < 0) {
		cfzy_conftree_free(conftree);
		LOG(ERR, "Cannot evaluate configuration tree\n");
		return -1;
	}
#endif

	memset(buf, 0, sizeof(buf));

	/* browse all nodes following priority list and prompt user */
	TAILQ_FOREACH(n, &conftree->prio_list, prio_next) {

		if (n->flags & CFZY_F_INVISIBLE)
			continue;
		if (n->flags & CFZY_F_NO_NAME)
			continue;

		/* node has a user value or user value cannot be set,
		 * just display it */
		if ((n->user_value != NULL) || (n->flags & CFZY_F_NO_SET_VALUE)) {
			/* XXX we should have a difference between
			 * NO_VALUE (menu for instance) and
			 * NO_SET_VALUE */
			cfzy_confnode_evaluate(n);
			printf("%s=%s\n", cfzy_confnode_name(n),
			       n->effective_value);
			continue;
		}

		/* node cannot be enabled due to deps */
		ret = cfzy_confnode_check_deps(n);
		if (ret == 0) {
			cfzy_confnode_evaluate(n);
			printf("%s=%s\n", cfzy_confnode_name(n),
			       n->effective_value);
			continue;
		}
		if (ret < 0) {
			LOG(ERR, "Cannot check deps\n");
			return -1;
		}

		/* ok, now prompt user */

		cfzy_confnode_dump(n, stdout, CFZY_DUMP_MODE_STANDARD);

		while (1) {
			printf("Enter user value for %s > ",
			       cfzy_confnode_name(n));
			fflush(stdout);

			fgets(buf, sizeof(buf) - 1, stdin);
			s = buf;
			strsep(&s, "\r\n");

			/* empty buf means default value */
			if (strcmp(buf, "") == 0)
				break;

			/* invalid value, prompt again */
			if (cfzy_confnode_set_uservalue(n, buf) < 0) {
				printf("invalid value\n");
				continue;
			}

			break;
		}

		cfzy_confnode_evaluate(n);
		printf("%s=%s\n", cfzy_confnode_name(n),
		       n->effective_value);
	}

	/* output .config if provided */
	if (cfzy_args->output_dotconfig != NULL &&
	    cfzy_dotconfig_write(conftree, cfzy_args->output_dotconfig) < 0) {
		cfzy_conftree_free(conftree);
		LOG(ERR, "Cannot output dotconfig <%s>\n",
		    cfzy_args->output_dotconfig);
		return -1;
	}

	/* output C header if provided */
	if (cfzy_args->output_c_hdr != NULL &&
	    cfzy_c_hdr_write(conftree, cfzy_args->output_c_hdr) < 0) {
		cfzy_conftree_free(conftree);
		LOG(ERR, "Cannot output C header <%s>\n",
		    cfzy_args->output_c_hdr);
		return -1;
	}

	/* output multiple C headers if provided */
	if (cfzy_args->output_multi_c_hdr != NULL &&
	    cfzy_multi_c_hdr_write(conftree, cfzy_args->output_multi_c_hdr) < 0) {
		cfzy_conftree_free(conftree);
		LOG(ERR, "Cannot output multiple C headers <%s>\n",
		    cfzy_args->output_multi_c_hdr);
		return -1;
	}

	cfzy_conftree_free(conftree);
	return 0;
}

int main(int argc, char **argv)
{
	int ret = EXIT_SUCCESS;
	struct cfzy_arguments cfzy_args;

	/* XXX a cfzy_init() would be better */
	cfzy_log_init();

	memset(&cfzy_args, 0, sizeof(cfzy_args));

	/* argument parsing is always successful (will exit on error) */
	parse_args(&cfzy_args, argc, argv);

	/* do the configuration */
	if (do_configuration(&cfzy_args) < 0) {
		LOG(ERR, "Configuration failed");
		ret = EXIT_FAILURE;
	}

	free_args(&cfzy_args);
	cfzy_log_exit();

	return ret;
}
