/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <libgen.h>

#include <cfzy_log.h>
#include <cfzy_list.h>
#include <cfzy_htable.h>
#include <cfzy_confnode.h>
#include <cfzy_conftree.h>
#include <cfzy_dotconfig.h>
#include <cfzy_c_hdr.h>

#include "test_conftree.h"

#define LOG(level, fmt, args...)				\
	CFZY_LOG("test_conftree", level, fmt, ##args)

#define SUCCESS                      0
#define CANNOT_PARSE_CONFTREE       -1
#define CANNOT_DUMP_CONFTREE        -2
#define CANNOT_FIND_SYMBOL          -3
#define CANNOT_EVAL_CONFTREE        -4
#define CANNOT_GET_VALUE            -5
#define CANNOT_SET_VALUE            -6
#define INVALID_VALUE               -7
#define CANNOT_FILTER_CONFTREE      -8
#define CANNOT_SAVE_CONFTREE_VALUES -9

/* check that a node is set to the expected value */
static int test_one_confnode(const struct cfzy_conftree *conftree,
			     const char *name, int expected_value)
{
	int val;

	val = cfzy_confnode_get_boolvalue(name, conftree);
	if (val < 0) {
		LOG(DEBUG, "cannot get node value\n");
		return CANNOT_GET_VALUE;
	}
	if (val != expected_value) {
		LOG(DEBUG, "invalid val for %s: val=%d, expect=%d\n",
			name, val, expected_value);
		return INVALID_VALUE;
	}

	return SUCCESS;
}

static int test_one_conftree(const char *name)
{
	struct cfzy_list_head *node_list = NULL;
	struct cfzy_list_elt *e;
	struct cfzy_conftree *conftree;
	struct cfzy_confnode *n;
	int ret;

	conftree = cfzy_conftree_new(name);
	if (conftree == NULL)
		return CANNOT_PARSE_CONFTREE;

	if (cfzy_conftree_evaluate(conftree) < 0) {
		LOG(DEBUG, "cannot eval conftree\n");
		cfzy_conftree_free(conftree);
		return CANNOT_EVAL_CONFTREE;
	}

	if (cfzy_conftree_dump(conftree, "/tmp/conftree.dump") < 0) {
		cfzy_conftree_free(conftree);
		return CANNOT_DUMP_CONFTREE;
	}

	/* node value should be 1 (default) */
	ret = test_one_confnode(conftree, "MENU1_CONFIG1", 1);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	/* node value should be 0 (default) */
	ret = test_one_confnode(conftree, "MENU1_CONFIG2", 0);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	/* lookup for this node */
	n = cfzy_htable_lookup(conftree->nodes_htable, "MENU1_CONFIG1");
	if (n == NULL) {
		LOG(DEBUG, "cannot find node\n");
		cfzy_conftree_free(conftree);
		return CANNOT_FIND_SYMBOL;
	}

	/* try to assign an invalid value: should fail */
	if (cfzy_confnode_set_uservalue(n, "dummy") == 0) {
		LOG(DEBUG, "set_uservalue returned 0 but should not\n");
		cfzy_conftree_free(conftree);
		return CANNOT_SET_VALUE;
	}

	/* set thiis node to NO */
	if (cfzy_confnode_set_uservalue(n, "n") < 0) {
		LOG(DEBUG, "cannot set node value\n");
		cfzy_conftree_free(conftree);
		return CANNOT_SET_VALUE;
	}

	/* node value should be 1 because we did not evaluate the tree */
	ret = test_one_confnode(conftree, "MENU1_CONFIG1", 1);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	if (cfzy_conftree_evaluate(conftree) < 0) {
		LOG(DEBUG, "cannot eval conftree\n");
		cfzy_conftree_free(conftree);
		return CANNOT_EVAL_CONFTREE;
	}

	if (cfzy_conftree_dump(conftree, "/tmp/conftree2.dump") < 0) {
		cfzy_conftree_free(conftree);
		return CANNOT_DUMP_CONFTREE;
	}

	if (cfzy_conftree_save_values(conftree) < 0) {
		cfzy_conftree_free(conftree);
		return CANNOT_SAVE_CONFTREE_VALUES;
	}

	if (cfzy_conftree_dump(conftree, "/tmp/conftree3.dump") < 0) {
		cfzy_conftree_free(conftree);
		return CANNOT_DUMP_CONFTREE;
	}

	/* node value should now be 0 */
	ret = test_one_confnode(conftree, "MENU1_CONFIG1", 0);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	/* node value should now be 1 (depends on !MENU1_CONFIG1) */
	ret = test_one_confnode(conftree, "MENU1_CONFIG2", 1);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	/* test "choice" nodes */

	/* node value should be 0 (default) */
	ret = test_one_confnode(conftree, "MENU1_CHOICE1", 0);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	/* node value should be 0 (default) */
	ret = test_one_confnode(conftree, "MENU1_CHOICE2", 0);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	/* node value should be 1 (default) */
	ret = test_one_confnode(conftree, "MENU1_CHOICE3", 1);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	/* lookup for this node */
	n = cfzy_htable_lookup(conftree->nodes_htable, "MENU1_CHOICE1");
	if (n == NULL) {
		LOG(DEBUG, "cannot find node\n");
		cfzy_conftree_free(conftree);
		return CANNOT_FIND_SYMBOL;
	}

	/* set this node to Y */
	if (cfzy_confnode_set_uservalue(n, "y") < 0) {
		LOG(DEBUG, "cannot set node value\n");
		cfzy_conftree_free(conftree);
		return CANNOT_SET_VALUE;
	}

	if (cfzy_conftree_evaluate(conftree) < 0) {
		LOG(DEBUG, "cannot eval conftree\n");
		cfzy_conftree_free(conftree);
		return CANNOT_EVAL_CONFTREE;
	}

	if (cfzy_conftree_dump(conftree, "/tmp/conftree.dump") < 0) {
		cfzy_conftree_free(conftree);
		return CANNOT_DUMP_CONFTREE;
	}

	/* node value should be 1 */
	ret = test_one_confnode(conftree, "MENU1_CHOICE1", 1);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	/* node value should be 0 */
	ret = test_one_confnode(conftree, "MENU1_CHOICE2", 0);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	/* node value should be 0 */
	ret = test_one_confnode(conftree, "MENU1_CHOICE3", 0);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	/* check the "menuconfig" node value, should be 1 (default) */
	ret = test_one_confnode(conftree, "MENUCONFIG1", 1);
	if (ret < 0) {
		cfzy_conftree_free(conftree);
		return ret;
	}

	/* check which node were changed by the user */
	node_list = cfzy_conftree_filter(conftree, CFZY_FILTER_F_USR_CHANGED);
	if (node_list == NULL) {
		cfzy_conftree_free(conftree);
		return CANNOT_FILTER_CONFTREE;
	}
	TAILQ_FOREACH(e, node_list, next) {
		n = e->ptr;
		printf("node %s\n", cfzy_confnode_name(n));
	}
	cfzy_list_free(node_list, NULL);

	/* check which effective value changed */
	node_list = cfzy_conftree_filter(conftree, CFZY_FILTER_F_EFF_CHANGED);
	if (node_list == NULL) {
		cfzy_conftree_free(conftree);
		return CANNOT_FILTER_CONFTREE;
	}
	TAILQ_FOREACH(e, node_list, next) {
		n = e->ptr;
		printf("node2 %s\n", cfzy_confnode_name(n));
	}
	cfzy_list_free(node_list, NULL);

	cfzy_conftree_free(conftree);
	return 0;
}

static int test_one_invalid_conftree(const char *name)
{
	struct cfzy_conftree *conftree;

	conftree = cfzy_conftree_new(name);
	if (conftree == NULL)
		return CANNOT_PARSE_CONFTREE;

	cfzy_conftree_free(conftree);
	return 0;
}

static void print_error(const char *in, int err)
{
	printf("Unexpected return value when parsing: <%s>\n", in);
	switch (err) {
		case SUCCESS:
			printf("Test returned success, but should not\n");
			break;
		case CANNOT_PARSE_CONFTREE:
			printf("Cannot parse configuration tree\n");
			break;
		case CANNOT_DUMP_CONFTREE:
			printf("Cannot dump configuration tree\n");
			break;
		case CANNOT_EVAL_CONFTREE:
			printf("Cannot evalulate configuration tree\n");
			break;
		case CANNOT_FIND_SYMBOL:
			printf("Cannot find symbol\n");
			break;
		case CANNOT_GET_VALUE:
			printf("Cannot get node value\n");
			break;
		case CANNOT_SET_VALUE:
			printf("Cannot set node value\n");
			break;
		case INVALID_VALUE:
			printf("invalid value\n");
			break;
		case CANNOT_FILTER_CONFTREE:
			printf("Cannot filter configuration tree\n");
			break;
		case CANNOT_SAVE_CONFTREE_VALUES:
			printf("Cannot save configuration tree values\n");
			break;
		default:
			printf("Invalid error %d\n", err);
			break;
	}
}

int test_conftree(const char *progname)
{
	char filename[PATH_MAX];
	char *progdir;
	int err;

	cfzy_log_set_default_level(CFZY_LOG_DEBUG);

	/* progdir is the directory where confizery-test is located */
	progdir = strdup(progname);
	dirname(progdir);

	/* unexistant file */
	snprintf(filename, sizeof(filename), "foobar");
	err = test_one_invalid_conftree(filename);
	if (err != CANNOT_PARSE_CONFTREE) {
		print_error(filename, err);
		goto fail;
	}

	/* invalid command */
	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/invalid-configs/"
		 "invalid-command.cfzy", progdir);
	err = test_one_invalid_conftree(filename);
	if (err != CANNOT_PARSE_CONFTREE) {
		print_error(filename, err);
		goto fail;
	}

	/* invalid attribute */
	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/invalid-configs/"
		 "invalid-attribute.cfzy", progdir);
	err = test_one_invalid_conftree(filename);
	if (err != CANNOT_PARSE_CONFTREE) {
		print_error(filename, err);
		goto fail;
	}

	/* circular dependency */
	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/invalid-configs/"
		 "circular-dep.cfzy", progdir);
	err = test_one_invalid_conftree(filename);
	if (err != CANNOT_PARSE_CONFTREE) {
		print_error(filename, err);
		goto fail;
	}

	/* circular dependency */
	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/invalid-configs/"
		 "circular-dep2.cfzy", progdir);
	err = test_one_invalid_conftree(filename);
	if (err != CANNOT_PARSE_CONFTREE) {
		print_error(filename, err);
		goto fail;
	}

	/* node not closed */
	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/invalid-configs/"
		 "node-not-closed.cfzy", progdir);
	err = test_one_invalid_conftree(filename);
	if (err != CANNOT_PARSE_CONFTREE) {
		print_error(filename, err);
		goto fail;
	}

	/* duplicated config name */
	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/invalid-configs/"
		 "dup-name.cfzy", progdir);
	err = test_one_invalid_conftree(filename);
	if (err != CANNOT_PARSE_CONFTREE) {
		print_error(filename, err);
		goto fail;
	}

	/* invalid default value */
	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/invalid-configs/"
		 "invalid-value.cfzy", progdir);
	err = test_one_invalid_conftree(filename);
	if (err != CANNOT_PARSE_CONFTREE) {
		print_error(filename, err);
		goto fail;
	}

	/* invalid expression */
	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/invalid-configs/"
		 "invalid-expr.cfzy", progdir);
	err = test_one_invalid_conftree(filename);
	if (err != CANNOT_PARSE_CONFTREE) {
		print_error(filename, err);
		goto fail;
	}

	/* valid config */

	/* an example config */
	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/test-configs/"
		 "conftree.cfzy", progdir);
	err = test_one_conftree(filename);
	if (err != SUCCESS) {
		print_error(filename, err);
		goto fail;
	}

	free(progdir);
	return 0;

 fail:
	free(progdir);
	return -1;
}
