/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <libgen.h>

#include <cfzy_log.h>
#include <cfzy_list.h>
#include <cfzy_htable.h>
#include <cfzy_confnode.h>
#include <cfzy_conftree.h>
#include <cfzy_dotconfig.h>
#include <cfzy_c_hdr.h>

#include "test_dotconfig.h"

#define LOG(level, fmt, args...)				\
	CFZY_LOG("test_dotconfig", level, fmt, ##args)

#define SUCCESS                      0
#define CANNOT_PARSE_CONFTREE       -1
#define CANNOT_EVAL_CONFTREE        -2
#define CANNOT_GENERATE_DOTCONFIG   -3
#define CANNOT_GENERATE_C_HDR       -4
#define CANNOT_READ_DOTCONFIG       -5

static void print_error(const char *in, int err)
{
	printf("Unexpected return value when parsing: <%s>\n", in);
	switch (err) {
		case SUCCESS:
			printf("Test returned success, but should not\n");
			break;
		case CANNOT_PARSE_CONFTREE:
			printf("Cannot parse configuration tree\n");
			break;
		case CANNOT_EVAL_CONFTREE:
			printf("Cannot evalulate configuration tree\n");
			break;
		case CANNOT_GENERATE_DOTCONFIG:
			printf("Cannot generate dotconfig file\n");
			break;
		case CANNOT_GENERATE_C_HDR:
			printf("Cannot generate C header file\n");
			break;
		case CANNOT_READ_DOTCONFIG:
			printf("Cannot read dotconfig file\n");
			break;
		default:
			printf("Invalid error %d\n", err);
			break;
	}
}

static int test_one_dotconfig(struct cfzy_conftree *conftree,
			      const char *filename, int expected)
{
	int ret, err;

	printf("-- Test dotconfig <%s>\n", filename);

	/* XXX we should reset config */

	ret = cfzy_dotconfig_read(conftree, filename);
	if (ret < 0)
		err = CANNOT_READ_DOTCONFIG;
	else
		err = SUCCESS;

	/* this is the expected result, return success */
	if (err == expected)
		return 0;

	print_error(filename, err);
	return -1;
}

int test_dotconfig(const char *progname)
{
	struct cfzy_conftree *conftree = NULL;
	int ret;
	char filename[PATH_MAX];
	char *progdir;

	cfzy_log_set_default_level(CFZY_LOG_DEBUG);

	/* progdir is the directory where confizery-test is located */
	progdir = strdup(progname);
	dirname(progdir);

	/* an example config */
	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/test-configs/"
		 "conftree.cfzy", progdir);

	conftree = cfzy_conftree_new(filename);
	if (conftree == NULL) {
		LOG(DEBUG, "cannot parse configuration tree\n");
		goto fail;
	}

	if (cfzy_conftree_evaluate(conftree) < 0) {
		LOG(DEBUG, "cannot eval conftree\n");
		goto fail;
	}

	/* write the dotconfig file */
	ret = cfzy_dotconfig_write(conftree, "/tmp/dotconfig");
	if (ret < 0) {
		LOG(DEBUG, "cannot write dotconfig file\n");
		goto fail;
	}

	/* write the autoconf.h file */
	ret = cfzy_c_hdr_write(conftree, "/tmp/autoconf.h");
	if (ret < 0) {
		LOG(DEBUG, "cannot write autoconf.h file\n");
		goto fail;
	}

	/* write the config/foo/bar.h files */
	ret = cfzy_multi_c_hdr_write(conftree, "/tmp/config");
	if (ret < 0) {
		LOG(DEBUG, "cannot write multiple headers file\n");
		goto fail;
	}

	/* test a valid config */
	if (test_one_dotconfig(conftree, "/tmp/dotconfig", SUCCESS) < 0)
		goto fail;

	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/test-configs/"
		 "dotconfig", progdir);
	if (test_one_dotconfig(conftree, filename, SUCCESS) < 0)
		goto fail;

	/* invalid configs */

	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/invalid-configs/"
		 "dotconfig-bad-val", progdir);
	if (test_one_dotconfig(conftree, filename, CANNOT_READ_DOTCONFIG) < 0)
		goto fail;

	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/invalid-configs/"
		 "dotconfig-bad-val2", progdir);
	if (test_one_dotconfig(conftree, filename, CANNOT_READ_DOTCONFIG) < 0)
		goto fail;

	snprintf(filename, sizeof(filename),
		 "%s/../../src/confizery-test/invalid-configs/"
		 "dotconfig-bad-val3", progdir);
	if (test_one_dotconfig(conftree, filename, CANNOT_READ_DOTCONFIG) < 0)
		goto fail;

	cfzy_conftree_free(conftree);
	free(progdir);
	return 0;

 fail:
	if (conftree != NULL)
		cfzy_conftree_free(conftree);
	free(progdir);
	return -1;
}
