/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <cfzy_expr.h>

#include "test_expr.h"

#define SUCCESS 0
#define CANNOT_PARSE -1
#define CANNOT_CONVERT_TO_STR -2
#define IN_AND_OUT_DIFFERS -3
#define CANNOT_EVAL -4
#define CANNOT_GET_VARS -5

/* all variables starting with "A" are evaluated to 1, the other are
 * evaluated to 0. The variable names "unknown" returns an error
 * (variable not found). */
static int getvalue(const char *varname, void *arg)
{
	(void)arg;

	if (!strcmp(varname, "unknown"))
		return -1;

	if (varname[0] == 'A')
		return 1;
	return 0;
}

static int test_one_expr(const char *in)
{
	char out[256];
	struct cfzy_expr *exp;
	int ret;

	/* valid expression */
	exp = cfzy_expr_parse(in);
	if (exp == NULL)
		return CANNOT_PARSE;

	cfzy_expr_graph_dump("test.graph", exp);
	ret = cfzy_expr_to_str(exp, out, sizeof(out));
	if (ret < 0) {
		cfzy_expr_free(exp);
		return CANNOT_CONVERT_TO_STR;
	}

	if (strncmp(in, out, sizeof(out))) {
		cfzy_expr_free(exp);
		return IN_AND_OUT_DIFFERS;
	}

	ret = cfzy_expr_eval(exp, getvalue, NULL);
	if (ret < 0) {
		cfzy_expr_free(exp);
		return CANNOT_EVAL;
	}

	cfzy_expr_free(exp);
	return ret;
}

static void print_error(const char *in, int err)
{
	printf("Unexpected return value when parsing: <%s>\n", in);
	switch(err) {
		case 0:
		case 1:
			printf("Test returned %d, but should not\n", err);
			break;
		case CANNOT_PARSE:
			printf("Cannot parse expression\n");
			break;
		case CANNOT_CONVERT_TO_STR:
			printf("Cannot convert expression to string\n");
			break;
		case IN_AND_OUT_DIFFERS:
			printf("Input and output expressions differ\n");
			break;
		case CANNOT_EVAL:
			printf("Cannot evaluate expression\n");
			break;
		case CANNOT_GET_VARS:
			printf("Cannot get variable list\n");
			break;
		default:
			printf("Invalid error %d\n", err);
			break;
	}
}

/* must be called with an expression containing only A and B as vars */
static int test_vars(const char *in)
{
	struct cfzy_expr *exp;
	struct cfzy_list_elt *e;
	struct cfzy_list_head *list;
	int ret = 0, count = 0;

	/* valid expression */
	exp = cfzy_expr_parse(in);
	if (exp == NULL)
		return CANNOT_PARSE;

	list = cfzy_expr_get_vars(exp);
	if (list == NULL) {
		cfzy_expr_free(exp);
		return CANNOT_GET_VARS;
	}

	/* list must contain 2 elements: A and B */
	TAILQ_FOREACH(e, list, next) {
		count++;
		if (!strcmp(e->ptr, "A"))
			continue;
		if (!strcmp(e->ptr, "B"))
			continue;
		ret = CANNOT_GET_VARS;
	}
	if (count != 2)
		ret = CANNOT_GET_VARS;

	cfzy_list_free(list, free);
	cfzy_expr_free(exp);
	return ret;
}

int test_expr(void)
{
	int ret = 0, err;
	const char *in;

	in = "true && true";
	err = test_one_expr(in);
	if (err != 1) {
		print_error(in, err);
		ret = -1;
	}

	in = "(true || false) && false";
	err = test_one_expr(in);
	if (err != 0) {
		print_error(in, err);
		ret = -1;
	}

	in = "true || (false && false)";
	err = test_one_expr(in);
	if (err != 1) {
		print_error(in, err);
		ret = -1;
	}

	/* && has priority */
	in = "true || false && false";
	err = test_one_expr(in);
	if (err != 1) {
		print_error(in, err);
		ret = -1;
	}

	/* && has priority */
	in = "false || true && false && true || true && false";
	err = test_one_expr(in);
	if (err != 0) {
		print_error(in, err);
		ret = -1;
	}

	/* '==' has priority */
	in = "false == true && false";
	err = test_one_expr(in);
	if (err != 0) {
		print_error(in, err);
		ret = -1;
	}

	in = "A";
	err = test_one_expr(in);
	if (err != 1) {
		print_error(in, err);
		ret = -1;
	}

	in = "B";
	err = test_one_expr(in);
	if (err != 0) {
		print_error(in, err);
		ret = -1;
	}

	in = "true && (true)";
	err = test_one_expr(in);
	if (err != 1) {
		print_error(in, err);
		ret = -1;
	}

	in = "(!(A1 && !(B1 || A2 && B2 || A_aa)) || "
		"(x && !(D && !A) && !(D || E))) == true";
	err = test_one_expr(in);
	if (err != 1) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = "  ";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = "(";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = ")";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = "( )";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = "A &&";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}
	/* invalid expression */
	in = "A && && B";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = "A B";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = "A && B || C D";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = "A (B)";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = "&& A B";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = "! A B";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = "A && !|| D";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression */
	in = "A && !";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression, variable must start with a letter */
	in = "A && _B";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* invalid expression, variable must start with a letter and
	 * numbers are not allowed (at least today) */
	in = "A == 4";
	err = test_one_expr(in);
	if (err != CANNOT_PARSE) {
		print_error(in, err);
		ret = -1;
	}

	/* valid expression, but unknown variable */
	in = "unknown";
	err = test_one_expr(in);
	if (err != CANNOT_EVAL) {
		print_error(in, err);
		ret = -1;
	}

	/* valid expression, but won't display like this because
	 * spaces are stripped */
	in = " true||(false    && \n false)";
	err = test_one_expr(in);
	if (err != IN_AND_OUT_DIFFERS) {
		print_error(in, err);
		ret = -1;
	}

	err = test_vars("A && B || true && A == !B");
	if (err != 0) {
		print_error(in, err);
		ret = -1;
	}

	return ret;
}
