/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <libgen.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "cfzy_log.h"
#include "cfzy_file.h"
#include "cfzy_conftree.h"
#include "cfzy_confnode.h"

#define LOG(level, fmt, args...)				\
	CFZY_LOG("c_hdr", level, fmt, ##args)

/* does not clean directories on error, if path exist, assume it's a
 * directory and ignore error */
int cfzy_recursive_mkdir(const char *path)
{
	char tmp[PATH_MAX];
	char *p = NULL;
	int n, ret;

	n = snprintf(tmp, sizeof(tmp),"%s", path);
	if (n <= 0 || n >= PATH_MAX)
		return -1;

	/* ignore all trailing '/' */
	while (n > 0 && tmp[--n] == '/')
		tmp[n] = '\0';

	/* create all intermediate directories */
	for (p = tmp + 1; *p; p++) {
		if (*p != '/')
			continue;

		*p = '\0';
		if (mkdir(tmp, S_IRWXU) < 0 && errno != EEXIST)
			return -1;

		*p = '/';
	}

	/* and last one */
	ret = mkdir(tmp, S_IRWXU);
	if (ret < 0 && errno == EEXIST)
		return 0;

	return ret;
}

/*
 * Get the file name from a config name. Replace first underscore
 * only:
 *   CONFIG_A_B -> config/a/b.h
 *   CONFIG_A__B___C -> config/a/_b/__c.h
 */
static char *configname_to_filename(const char *name)
{
	char *filename, *s;
	int len;

	/* allocate room for dst buffer */
	len = strlen(name);
	filename = malloc(len + 3);
	if (filename == NULL)
		return NULL;
	strncpy(filename, name, len + 1);

	for (s = filename; *s != '\0'; s++) {
		/* '/' is not allowed in variable name */
		if (*s == '/') {
			free(filename);
			return NULL;
		}

		/* convert the first '_' in '/' */
		if (*s == '_') {
			*s = '/';
			while (s[1] == '/')
				s++;
			continue;
		}

		if (isalpha(*s))
			*s = tolower(*s);
	}
	*s++ = '.';
	*s++ = 'h';
	*s = '\0';

	return filename;
}

static int __cfzy_multi_c_hdr_write(const struct cfzy_confnode *n)
{
	const struct cfzy_confnode *c;
	FILE *f = NULL;
	char *filename = NULL;
	char *dname_buf = NULL, *dname_ptr = NULL;
	char oldval[512];
	char newval[512];
	int ret, val;

	/* only dump the node if it has a name and a value */
	if (n->name != NULL && n->effective_value != NULL) {
		/* get the name of the file (full path) */
		filename = configname_to_filename(n->name);
		if (filename == NULL)
			goto fail;

		/* get directory name from file name */
		dname_buf = strdup(filename);
		if (dname_buf == NULL)
			goto fail;
		dname_ptr = dirname(dname_buf);

		/* create dirs */
		if (cfzy_recursive_mkdir(dname_ptr) < 0)
			goto fail;

		/* get old value */
		memset(oldval, 0, sizeof(oldval));
		f = fopen(filename, "r");
		if (f != NULL) {
			ret = fread(oldval, 1, sizeof(oldval) - 1, f);
			if (ret < 0) {
				LOG(ERR, "cannot read <%s>\n", filename);
				goto fail;
			}
			fclose(f);
			f = NULL;
		}

		/* if config node is a bool, we have NULL and n will
		 * output the same val (is not st) */
		if (n->flags & CFZY_F_VAL_IS_BOOL) {
			val = cfzy_confnode_str2bool(n, n->effective_value);
			if (val < 0)
				return -1;

			/* bool is false -> not set */
			if (val == 0) {
				if (snprintf(newval, sizeof(newval),
					     "/* %s is not set */",
					     cfzy_confnode_name(n)) < 0)
					goto fail;
			}
			else {
				/* bool is true */
				if (snprintf(newval, sizeof(newval),
					     "/* %s = %s */", cfzy_confnode_name(n),
					     n->effective_value) < 0)
					goto fail;
			}
		}
		else {
			if (snprintf(newval, sizeof(newval),
				     "/* %s = %s */", cfzy_confnode_name(n),
				     n->effective_value) < 0)
				goto fail;
		}

		/* if the value is different, write the new value */
		if (strncmp(newval, oldval, sizeof(newval)) != 0) {
			f = fopen(filename, "w");
			if (f == NULL) {
				LOG(ERR, "cannot open file <%s>\n", filename);
				goto fail;
			}

			if (fprintf(f, "%s", newval) < 0) {
				LOG(ERR, "cannot write in <%s>\n",
				    filename);
				goto fail;
			}
			fclose(f);
			f = NULL;
		}

		free(dname_buf);
		free(filename);
	}

	/* dump all the children of root node */
	TAILQ_FOREACH(c, &n->children, child_next) {
		if (__cfzy_multi_c_hdr_write(c) < 0)
			break;
	}

	return 0;

 fail:
	if (f != NULL)
		fclose(f);
	if (dname_buf != NULL)
		free(dname_buf);
	if (filename != NULL)
		free(filename);

	return -1;
}

/* write config/foo/bar.h */
int cfzy_multi_c_hdr_write(const struct cfzy_conftree *conftree,
			   const char *basedir)
{
	char old_cwd_buf[PATH_MAX];
	char *old_cwd = NULL;
	const struct cfzy_confnode *c;
	int ret;

	LOG(INFO, "multi C headers write <%s>\n", basedir);

	if (cfzy_recursive_mkdir(basedir) < 0)
		return -1;

	old_cwd = getcwd(old_cwd_buf, sizeof(old_cwd_buf));
	if (old_cwd == NULL) {
		LOG(ERR, "getcwd failed: %s\n", strerror(errno));
		return -1;
	}

	if (chdir(basedir) < 0) {
		LOG(ERR, "chdir failed: %s\n", strerror(errno));
		return -1;
	}

	/* dump all the children of root node */
	TAILQ_FOREACH(c, &conftree->root->children, child_next) {
		ret = __cfzy_multi_c_hdr_write(c);
		if (ret < 0)
			break;
	}

	if (chdir(old_cwd) < 0)
		LOG(ERR, "chdir back to <%s> failed: %s\n",
		    old_cwd, strerror(errno));

	return ret;
}

/* write config/autoconf.h */
int cfzy_c_hdr_write(const struct cfzy_conftree *conftree,
		     const char *filename)
{
	FILE *tmp_f, *f;
	const struct cfzy_confnode *c;
	int ret;

	LOG(INFO, "open old C header <%s>\n", filename);

	/* open old file in read mode */

	f = fopen(filename, "r");
	if (f == NULL && errno != ENOENT) {
		printf("cannot open file <%s>: %s\n",
		       filename, strerror(errno));
		return -1;
	}

	/* old file exists */
	if (f != NULL) {

		/* write C header in temp file */
		tmp_f = tmpfile();
		if (tmp_f == NULL) {
			printf("cannot open temp file\n");
			return -1;
		}

		/* dump all the children of root node */
		TAILQ_FOREACH(c, &conftree->root->children, child_next) {
			ret = cfzy_confnode_c_hdr_write(c, tmp_f);
			if (ret < 0)
				break;
		}

		/* files are the same, nothing to do */
		if (cfzy_file_compare(f, tmp_f) == 0) {
			LOG(INFO, "already up to date: <%s>\n", filename);
			fclose(tmp_f);
			fclose(f);
			return 0;
		}

		fclose(tmp_f);
		fclose(f);
	}

	/* reopen the file in write mode */

	LOG(INFO, "write new C header <%s>\n", filename);

	f = fopen(filename, "w");
	if (f == NULL) {
		printf("cannot open file <%s>\n", filename);
		return -1;
	}

	/* dump all the children of root node */
	TAILQ_FOREACH(c, &conftree->root->children, child_next) {
		ret = cfzy_confnode_c_hdr_write(c, f);
		if (ret < 0)
			break;
	}

	fclose(f);

	return ret;
}
