/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/queue.h>

#include "cfzy_log.h"
#include "cfzy_list.h"
#include "cfzy_htable.h"
#include "cfzy_string.h"
#include "cfzy_expr.h"
#include "cfzy_confnode.h"
#include "cfzy_conftree.h"

#define LOG(level, fmt, args...)				\
	CFZY_LOG("confnode", level, fmt, ##args)

/* alloc a new confnode */
struct cfzy_confnode *cfzy_confnode_alloc(struct cfzy_conftree *conftree)
{
	struct cfzy_confnode *n;

	n = malloc(sizeof(struct cfzy_confnode));
	if (n == NULL)
		return NULL;

	memset(n, 0, sizeof(*n));
	TAILQ_INIT(&n->children);

	n->conftree = conftree;

	n->expr_deps = cfzy_list_alloc();
	if (n->expr_deps == NULL) {
		free(n);
		return NULL;
	}

	n->default_val_list = cfzy_list_alloc();
	if (n->default_val_list == NULL) {
		cfzy_list_free(n->expr_deps, NULL);
		free(n);
		return NULL;
	}

	return n;
}

/* Free the node given as argument, and all its associated data */
void cfzy_confnode_free(struct cfzy_confnode *n)
{
	struct cfzy_confnode *c;
	struct cfzy_list_elt *e;
	struct cfzy_confnode_defvalue *defval;

	/* free the children */
	while ((c = TAILQ_FIRST(&n->children)) != NULL) {
		TAILQ_REMOVE(&n->children, c, child_next);
		cfzy_confnode_free(c);
	}

	/* do the specific free first */
	if (n->ops && n->ops->free)
		n->ops->free(n);

	/* free the deps expression list */
	cfzy_list_free(n->expr_deps, (void *)cfzy_expr_free);

	/* free the deps expression list (manually free each element
	 * instead of providing a function) */
	while ((e = TAILQ_FIRST(n->default_val_list)) != NULL) {
		TAILQ_REMOVE(n->default_val_list, e, next);
		defval = e->ptr;
		cfzy_expr_free(defval->expr);
		free(defval->val);
		free(defval);
		free(e);
	}
	cfzy_list_free(n->default_val_list, NULL);

	if (n->node_deps != NULL)
		cfzy_list_free(n->node_deps, NULL);

	/* free the name, help, prompt, ... */
	if (n->name)
		free(n->name);
	if (n->prompt)
		free(n->prompt);
	if (n->help)
		free(n->help);
	if (n->default_value)
		free(n->default_value);
	if (n->user_value)
		free(n->user_value);
	if (n->effective_value)
		free(n->effective_value);
	if (n->old_user_value)
		free(n->old_user_value);
	if (n->old_effective_value)
		free(n->old_effective_value);
	if (n->filename)
		free(n->filename);

	free(n);
}

/* return the list of nodes required to enable this one */
struct cfzy_list_head *
cfzy_confnode_get_deplist(const struct cfzy_confnode *n)
{
	struct cfzy_list_head *node_list = NULL, *var_list= NULL;
	struct cfzy_list_elt *e, *e2;
	struct cfzy_expr *expr;
	struct cfzy_confnode *c;
	const char *varname;
	const struct cfzy_confnode_defvalue *defval;
	const struct cfzy_conftree *conftree;

	node_list = cfzy_list_alloc();
	if (node_list == NULL) {
		LOG(ERR, "cannot allocate node list\n");
		return NULL;
	}

	/* get associated configuration tree */
	conftree = n->conftree;

	/* list of dependencies (expression list) */
	TAILQ_FOREACH(e, n->expr_deps, next) {

		expr = e->ptr;

		/* get list of variables for this expression */
		var_list = cfzy_expr_get_vars(expr);
		if (var_list == NULL) {
			LOG(ERR, "cannot allocate get list\n");
			goto fail;
		}

		/* list of variables in this expression */
		TAILQ_FOREACH(e2, var_list, next) {

			varname = e2->ptr;

			c = cfzy_htable_lookup(conftree->nodes_htable, varname);
			if (c == NULL) {
				LOG(ERR, "cannot find node <%s>\n", varname);
				goto fail;
			}

			if (cfzy_list_elt_is_in_list(node_list, c))
				continue;

			if (cfzy_list_add_tail(node_list, c) < 0) {
				LOG(ERR, "cannot add node in list\n");
				goto fail;
			}
		}

		cfzy_list_free(var_list, free);
		var_list = NULL;
	}

	/* list of conditional default values depending on an expression */
	TAILQ_FOREACH(e, n->default_val_list, next) {

		defval = e->ptr;
		expr = defval->expr;

		/* get list of variables for this expression */
		var_list = cfzy_expr_get_vars(expr);
		if (var_list == NULL) {
			LOG(ERR, "cannot allocate get list\n");
			goto fail;
		}

		/* list of variables in this expression */
		TAILQ_FOREACH(e2, var_list, next) {

			varname = e2->ptr;

			c = cfzy_htable_lookup(conftree->nodes_htable, varname);
			if (c == NULL) {
				LOG(ERR, "cannot find node <%s>\n", varname);
				goto fail;
			}

			if (cfzy_list_elt_is_in_list(node_list, c))
				continue;

			if (cfzy_list_add_tail(node_list, c) < 0) {
				LOG(ERR, "cannot add node in list\n");
				goto fail;
			}
		}

		cfzy_list_free(var_list, free);
		var_list = NULL;
	}

	/* parent */
	if (n->parent != NULL && !cfzy_list_elt_is_in_list(node_list, n->parent)) {

		if (cfzy_list_add_tail(node_list, n->parent) < 0) {
			LOG(ERR, "cannot add node in list\n");
			goto fail;
		}
	}

	return node_list;

 fail:
	if (node_list != NULL)
		cfzy_list_free(node_list, NULL);
	if (var_list != NULL)
		cfzy_list_free(var_list, free);

	return NULL;
}

/* Add a conditional default value to a node */
int cfzy_confnode_add_defval(struct cfzy_confnode *n, const char *val,
			     const char *expr_buf)
{
	struct cfzy_confnode_defvalue *defval;

	defval = malloc(sizeof(*defval));
	if (defval == NULL)
		return -1;

	memset(defval, 0, sizeof(*defval));
	defval->expr = cfzy_expr_parse(expr_buf);
	if (defval->expr == NULL) {
		free(defval);
		return -1;
	}

	defval->val = strdup(val);
	if (defval->val == NULL) {
		cfzy_expr_free(defval->expr);
		free(defval);
		return -1;
	}

	if (cfzy_list_add_tail(n->default_val_list, defval) < 0) {
		free(defval->val);
		cfzy_expr_free(defval->expr);
		free(defval);
		return -1;
	}

	return 0;
}

/* Parse a line structure (and its associated line buffer) to match an
 * attribute of a confnode. Return 0 if the line matches, else return
 * -1 if we cannot match a valid attribute. */
enum cfzy_parse_return
cfzy_confnode_add_attr(struct cfzy_confnode *n,
		       const struct cfzy_token_list *tklist)
{
	enum cfzy_parse_return ret;
	struct cfzy_token *first_tok, *tok;
	const char *linebuf = tklist->linebuf;

	first_tok = TAILQ_FIRST(&tklist->list);

	/* specific attribute */
	if (n->ops->add_attr != NULL) {
		ret = n->ops->add_attr(n, tklist, linebuf);
		if (ret != NO_MATCH)
			return ret;
	}

	/* syntax is: "prompt content" */
	if (tklist->n_token == 2 &&
	    !strcmp(first_tok->str, "prompt")) {

		if (n->flags & CFZY_F_NO_SET_PROMPT) {
			LOG(ERR, "node does not support 'prompt' attr\n");
			return ERROR;
		}

		tok = TAILQ_NEXT(first_tok, next);
		if (n->prompt != NULL) {
			free(n->prompt);
			n->prompt = NULL;
		}
		n->prompt = strdup(tok->str);
		if (n->prompt == NULL) {
			LOG(ERR, "cannot allocate prompt\n");
			return ERROR;
		}
		return SUCCESS;
	}

	/* syntax is: "requires EXPRESSION" */
	if (tklist->n_token > 1 &&
	    !strcmp(first_tok->str, "requires")) {
		struct cfzy_expr *exp;

		if (n->flags & CFZY_F_NO_SET_DEPS) {
			LOG(ERR, "node does not support 'requires' attr\n");
			return ERROR;
		}

		tok = TAILQ_NEXT(first_tok, next);
		exp = cfzy_expr_parse(linebuf + tok->offset);
		if (exp == NULL) {
			LOG(ERR, "cannot parse expression\n");
			return ERROR;
		}

		cfzy_list_add_tail(n->expr_deps, exp);
		return SUCCESS;
	}

	/* syntax is: "default VALUE" */
	if (tklist->n_token == 2 &&
	    !strcmp(first_tok->str, "default")) {

		if (n->flags & CFZY_F_NO_SET_DEFAULT) {
			LOG(ERR, "node does not support 'default' attr\n");
			return ERROR;
		}

		tok = TAILQ_NEXT(first_tok, next);
		if (cfzy_confnode_str2bool(n, tok->str) < 0) {
			LOG(ERR, "invalid value for this node\n");
			return ERROR;
		}

		if (n->default_value != NULL)
			free(n->default_value);
		n->default_value = strdup(tok->str);
		if (n->default_value == NULL) {
			LOG(ERR, "cannot add default value\n");
			return ERROR;
		}

		return SUCCESS;
	}

	/* syntax is: "default VALUE if EXPR" */
	if (tklist->n_token >= 4 &&
	    !strcmp(first_tok->str, "default")) {
		struct cfzy_token *tok2;

		tok = TAILQ_NEXT(first_tok, next); /* points to value */
		tok2 = TAILQ_NEXT(tok, next); /* points to "if" */

		if (strcmp(tok2->str, "if"))
			return NO_MATCH;

		if (n->flags & CFZY_F_NO_SET_DEFAULT) {
			LOG(ERR, "node does not support 'default' attr\n");
			return ERROR;
		}

		tok2 = TAILQ_NEXT(tok2, next); /* points to expression */

		if (cfzy_confnode_str2bool(n, tok->str) < 0) {
			LOG(ERR, "invalid value for this node\n");
			return ERROR;
		}

		if (cfzy_confnode_add_defval(n, tok->str,
					     linebuf + tok2->offset) < 0) {
			LOG(ERR, "cannot add default value\n");
			return ERROR;
		}

		return SUCCESS;
	}

	return NO_MATCH;
}

/* write config value in file f in a dotconfig-like manner. Return 0
 * on success. */
int cfzy_confnode_dotconfig_write(const struct cfzy_confnode *n, FILE *f)
{
	int val;
	const struct cfzy_confnode *c;

	if (n->ops->dotconfig_write)
		return n->ops->dotconfig_write(n, f);

	if (n->flags & CFZY_F_DOTCONF_SHOW_TITLE) {
		if (fprintf(f,
			    "#\n"
			    "# -- %s\n"
			    "#\n", n->prompt) < 0)
			return -1;
	}

	/* only dump the children if effective value is not defined
	 * (some nodes like "comment", "if", "menu", ... can have a
	 * NULL effective value) */
	if (n->effective_value == NULL)
		goto dump_children;

	/* when a boolean is unset, use the "is not set" syntax */
	if (n->flags & CFZY_F_VAL_IS_BOOL) {
		val = cfzy_confnode_str2bool(n, n->effective_value);
		if (val < 0)
			return -1;

		if (val == 0) {
			if (fprintf(f, "# CONFIG_%s is not set\n",
				    cfzy_confnode_name(n)) < 0)
				return -1;
			return 0;
		}
	}

	/* common case: dump the value, with or without quotes */
	if (n->flags & CFZY_F_QUOTE_VAL) {
		char *quoted_value = cfzy_string_quote(n->effective_value);

		if (quoted_value == NULL)
			return -1;

		if (fprintf(f, "CONFIG_%s=%s\n", cfzy_confnode_name(n),
			    quoted_value) < 0) {
			free(quoted_value);
			return -1;
		}
		free(quoted_value);
	}
	else {
		if (fprintf(f, "CONFIG_%s=%s\n", cfzy_confnode_name(n),
			    n->effective_value) < 0)
			return -1;
	}

 dump_children:
	TAILQ_FOREACH(c, &n->children, child_next) {
		if (cfzy_confnode_dotconfig_write(c, f) < 0)
			return -1;
	}
	return 0;
}

/* write config value in file f in a dotconfig-like manner. Return 0
 * on success. */
int cfzy_confnode_c_hdr_write(const struct cfzy_confnode *n, FILE *f)
{
	int val;
	const struct cfzy_confnode *c;

	if (n->ops->c_hdr_write)
		return n->ops->c_hdr_write(n, f);

	/* if no value, don't dump the value of the node, just the children */
	if (n->effective_value == NULL)
		goto dump_children;

	if (n->flags & CFZY_F_VAL_IS_BOOL) {
		val = cfzy_confnode_str2bool(n, n->effective_value);
		if (val < 0)
			return -1;

		/* bool is false -> undef */
		if (val == 0) {
			if (fprintf(f, "#undef CONFIG_%s\n",
				    cfzy_confnode_name(n)) < 0)
				return -1;
			return 0;
		}

		/* bool is true -> #define to 1 */
		if (fprintf(f, "#define CONFIG_%s 1\n",
			    cfzy_confnode_name(n)) < 0)
			return -1;
	}
	/* common case: dump the value, with or without quotes */
	else if (n->flags & CFZY_F_QUOTE_VAL) {
		char *quoted_value = cfzy_string_quote(n->effective_value);

		if (quoted_value == NULL)
			return -1;

		if (fprintf(f, "#define CONFIG_%s %s\n",
			    cfzy_confnode_name(n), quoted_value) < 0) {
			free(quoted_value);
			return -1;
		}
		free(quoted_value);
	}
	else {
		if (fprintf(f, "#define CONFIG_%s %s\n",
			    cfzy_confnode_name(n), n->effective_value) < 0)
			return -1;
	}

 dump_children:
	TAILQ_FOREACH(c, &n->children, child_next) {
		if (cfzy_confnode_c_hdr_write(c, f) < 0)
			return -1;
	}
	return 0;
}

/* Close the node beeing parsed */
int cfzy_confnode_close(struct cfzy_confnode *n)
{
	if (n->ops->close)
		return n->ops->close(n);

	return 0;
}

/* Return the boolean value of the node given the string value */
int cfzy_confnode_str2bool(const struct cfzy_confnode *n,
				const char *value)
{
	if (n->ops->str2bool)
		return n->ops->str2bool(n, value);

	/* if there is no specific function, any non-NULL value
	 * returns 1 */
	if (value == NULL)
		return 0;

	return 1;
}

/* Return a string identifying the node type ("config", "menuconfig",
 * "choice", ...) */
const char *cfzy_confnode_get_type_str(const struct cfzy_confnode *n)
{
	if (n->ops->get_type_str)
		return n->ops->get_type_str(n);

	return "unknown";
}

static int __get_path(const struct cfzy_confnode *n, char *buf,
		      int len, int first)
{
	int ret = 0, off;

	if (n->parent == NULL)
		return snprintf(buf, len, "/");

	ret = __get_path(n->parent, buf, len, 0);
	if (ret < 0)
		return ret;
	off = ret;
	if (off >= len)
		return -1;

	if (n->flags & CFZY_F_INVISIBLE)
		return ret;

	/* a visible node must have a name */
	if (n->name == NULL)
		return -1;

	ret = snprintf(buf + off, len - off, "%s%s", n->name,
		       first ? "" : "/");
	if (ret >= len - off)
		return -1;

	return ret + off;
}

/* Dump path in buffer. Invisible nodes like 'if' are ignored. */
int cfzy_confnode_path(const struct cfzy_confnode *n, char *buf, int len)
{
	int ret;

	ret = __get_path(n, buf, len, 1);
	if (ret < 0)
		return ret;
	return 0;
}

/* dump a config file in a file */
int cfzy_confnode_dump(const struct cfzy_confnode *n, FILE *f, int mode)
{
	const struct cfzy_confnode *c;
	struct cfzy_confnode_defvalue *defval;
	struct cfzy_list_elt *e;
	struct cfzy_expr *expr;
	char buf[512];

	if (fprintf(f, "%s\n", cfzy_confnode_get_type_str(n)) < 0)
		return -1;
	if (fprintf(f, "  name: %s\n", cfzy_confnode_name(n)) < 0)
		return -1;
	if (fprintf(f, "  prompt: %s\n", n->prompt) < 0)
		return -1;
	if (cfzy_confnode_path(n, buf, sizeof(buf)) == 0) {
		if (fprintf(f, "  path: %s\n", buf) < 0)
			return -1;
	}

	if (mode == CFZY_DUMP_MODE_MINIMAL)
		return 0;

	/* explicit dependencies */
	TAILQ_FOREACH(e, n->expr_deps, next) {
		expr = e->ptr;
		if (cfzy_expr_to_str(expr, buf, sizeof(buf)) < 0)
			return -1;
		if (fprintf(f, "  requires: %s\n", buf) < 0)
			return -1;
	}

	/* XXX display valid values */

	/* default values */
	TAILQ_FOREACH(e, n->default_val_list, next) {
		defval = e->ptr;

		if (cfzy_expr_to_str(defval->expr, buf, sizeof(buf)) < 0)
			return -1;

		if (fprintf(f, "  default: %s if %s\n", defval->val, buf) < 0)
				return -1;
	}
	if (fprintf(f, "  node_default %s\n", n->default_value) < 0)
		return -1;
	if (fprintf(f, "  user_value %s\n", n->user_value) < 0)
		return -1;
	if (fprintf(f, "  effective_value %s\n", n->effective_value) < 0)
		return -1;

	if (mode == CFZY_DUMP_MODE_STANDARD)
		return 0;

	if (fprintf(f, "  old_user_value %s\n", n->old_user_value) < 0)
		return -1;
	if (fprintf(f, "  old_effective_value %s\n", n->old_effective_value) < 0)
		return -1;
	if (fprintf(f, "  parsed at %s:%d\n", n->filename, n->linenum) < 0)
		return -1;

	if (fprintf(f, "\n") < 0)
		return -1;

	if (mode == CFZY_DUMP_MODE_FULL)
		return 0;

	/* dump children */
	TAILQ_FOREACH(c, &n->children, child_next) {
		if (cfzy_confnode_dump(c, f, CFZY_DUMP_MODE_FULL_RECURSIVE) < 0)
			return -1;
	}

	/* close node if it's a directory */
	if (n->flags & CFZY_F_IS_DIR) {
		if (fprintf(f, "end%s\n\n", cfzy_confnode_get_type_str(n)) < 0)
			return -1;
	}

	return 0;
}

/* set the uservalue of a node */
int cfzy_confnode_set_uservalue(struct cfzy_confnode *n, const char *value)
{
	char *newvalue = NULL;

	if (n->ops->set_uservalue)
		return n->ops->set_uservalue(n, value);

	/* check if value is valid */
	if (cfzy_confnode_str2bool(n, value) < 0)
		return -1;

	if (value != NULL) {
		newvalue = strdup(value);
		if (newvalue == NULL) {
			LOG(ERR, "cannot allocate new value\n");
			return -1;
		}
	}

	if (n->user_value != NULL)
		free(n->user_value);

	n->user_value = newvalue;
	return 0;
}

/* Get the boolean value of a node, given its name, used as a callback
 * by cfzy_expr_parse() when parsing an expression for
 * cfzy_confnode_evaluate() */
int cfzy_confnode_get_boolvalue(const char *name,
				const struct cfzy_conftree *conftree)
{
	const struct cfzy_confnode *n;

	n = cfzy_htable_lookup(conftree->nodes_htable, name);
	if (n == NULL)
		return -1;

	return cfzy_confnode_str2bool(n, n->effective_value);
}

/* return 0 if a parent node is disabled or if a prerequisite
 * expression is false, else return 1. On error, return -1. */
int cfzy_confnode_check_deps(struct cfzy_confnode *n)
{
	struct cfzy_list_elt *e;
	struct cfzy_expr *expr;
	struct cfzy_confnode *parent;
	int (*get_boolvalue)(const char *, const struct cfzy_conftree *);
	int (*get_boolvalue_casted)(const char *, void *);
	int ret;

	/* we use an intermediate variable to have a compiler warning
	 * if the prototype of cfzy_confnode_get_boolvalue changes. We
	 * know that that these types are compatible here. */
	get_boolvalue = cfzy_confnode_get_boolvalue;
	get_boolvalue_casted = (void *)get_boolvalue;

	/* check the dep list first */
	TAILQ_FOREACH(e, n->expr_deps, next) {
		expr = e->ptr;

		ret = cfzy_expr_eval(expr, get_boolvalue_casted, n->conftree);
		if (ret < 0)
			return -1;

		/* the expression is evaluated to 0, we cannot enable
		 * this node */
		if (ret == 0) {
			n->effective_value = NULL;
			LOG(DEBUG, "disable node <%s> (expr dep)\n",
			    cfzy_confnode_name(n));
			return 0;
		}
	}

	/* check the parent nodes */
	for (parent = n->parent; parent != NULL; parent = parent->parent) {
		ret = cfzy_confnode_str2bool(parent, parent->effective_value);
		if (ret < 0)
			return -1;

		/* a parent node is evaluated to 0, we cannot enable
		 * this node */
		if (ret == 0) {
			n->effective_value = NULL;
			LOG(DEBUG, "disable node <%s> (parent dep)\n",
			    cfzy_confnode_name(n));
			return 0;
		}
	}

	return 1;
}

/* Evaluate the effective value of a node, assuming all prerequisites
 * are already evaluated. */
int cfzy_confnode_evaluate(struct cfzy_confnode *n)
{
	struct cfzy_list_elt *e;
	struct cfzy_confnode_defvalue *defval;
	int (*get_boolvalue)(const char *, const struct cfzy_conftree *);
	int (*get_boolvalue_casted)(const char *, void *);
	int ret;

	LOG(DEBUG, "evaluating node <%s>\n", cfzy_confnode_name(n));

	/* we use an intermediate variable to have a compiler warning
	 * if the prototype of cfzy_confnode_get_boolvalue changes. We
	 * know that that these types are compatible here. */
	get_boolvalue = cfzy_confnode_get_boolvalue;
	get_boolvalue_casted = (void *)get_boolvalue;

	/* remove previous effective value */
	if (n->effective_value != NULL) {
		free(n->effective_value);
		n->effective_value = NULL;
	}

	/* check prerequisites and parents */
	ret = cfzy_confnode_check_deps(n);
	if (ret == 0) {
		/* we cannot enable this node due to dependencies */
		n->effective_value = NULL;
		LOG(DEBUG, "disable node <%s> (expr dep)\n",
		    cfzy_confnode_name(n));
		return 0;
	}
	/* error */
	if (ret < 0)
		return ret;

	/* user value has the priority */
	if (n->user_value != NULL) {
		n->effective_value = strdup(n->user_value);
		if (n->effective_value == NULL)
			return -1;

		LOG(DEBUG, "set node <%s> to user value <%s>\n",
		    cfzy_confnode_name(n), n->effective_value);
		return 0;
	}

	/* assign a default value */
	TAILQ_FOREACH(e, n->default_val_list, next) {
		defval = e->ptr;

		ret = cfzy_expr_eval(defval->expr, get_boolvalue_casted,
				     n->conftree);
		if (ret < 0)
			return -1;

		/* the expression is evaluated to 1, use this default
		 * value for this node */
		if (ret == 1) {
			n->effective_value = strdup(defval->val);
			if (n->effective_value == NULL)
				return -1;
			LOG(DEBUG, "set node <%s> to default_list value <%s>\n",
			    cfzy_confnode_name(n), n->effective_value);
			return 0;
		}
	}

	if (n->default_value != NULL)
		n->effective_value = strdup(n->default_value);

	LOG(DEBUG, "set node <%s> to default value <%s>\n",
	    cfzy_confnode_name(n), n->effective_value);

	return 0;
}
