/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file
 * Confizery configuration node
 *
 * This module provides functions related to configuration nodes. A
 * configuration node is a structure that stores a part of the
 * information of the configuration tree. For instance, a "config" or
 * "menuconfig" option is stored in a configuration node structure.
 *
 * It contains among other fields the name of the node (CONFIG_XYZ),
 * the prompt and the help. Each node references its parent and a list
 * of children as some nodes can contains some other nodes (menu, if,
 * choice, ...). It also contains the default values, the user value,
 * and the effective value of the node. Some nodes have no values (if,
 * comment, ...).
 */

#ifndef _CFZY_CONFNODE_H_
#define _CFZY_CONFNODE_H_

#include <stdio.h>

#include "cfzy_confnode_ops.h"
#include "cfzy_conftree_parser.h"

/* confnode flags */
#define CFZY_F_IS_DIR             0x0001 /**< node can contain other nodes */
#define CFZY_F_NO_SET_DEPS        0x0002 /**< does not support the "requires" attribute */
#define CFZY_F_NO_SET_PROMPT      0x0004 /**< does not support the "prompt" attribute */
#define CFZY_F_NO_SET_DEFAULT     0x0008 /**< does not support the "default" attribute */
#define CFZY_F_IS_ROOT            0x0010 /**< the root node */
#define CFZY_F_VAL_IS_BOOL        0x0020 /**< value is a boolean */
#define CFZY_F_INVISIBLE          0x0040 /**< never seen from cmdline (like the IF node) */
#define CFZY_F_NO_NAME            0x0080 /**< node has no name (if, comments, ...) */
#define CFZY_F_NO_SET_VALUE       0x0100 /**< user cannot set value (if, comments, ...) */
#define CFZY_F_QUOTE_VAL          0x0200 /**< quote value */
#define CFZY_F_DOTCONF_SHOW_TITLE 0x0400 /**< show title as a comment in dotconfig file  */

struct cfzy_conftree;

/**
 * Structure storing a conditional default value
 */
struct cfzy_confnode_defvalue {
	struct cfzy_expr *expr; /**< expression (condition) */
	char *val;              /**< value */
};

/**
 * List of cfzy_node, used to chain children together
 */
TAILQ_HEAD(cfzy_confnode_list, cfzy_confnode);

/**
 * This structure defines a configuration node in the configration
 * tree, for instance a "menu" entry, or a "bool" entry.
 */
struct cfzy_confnode {
	/* pointer to other nodes in conf tree */
	struct cfzy_confnode *parent;          /**< pointer to parent */
	struct cfzy_confnode_list children;    /**< list of children */
	TAILQ_ENTRY(cfzy_confnode) child_next; /**< next in children list */
	struct cfzy_conftree *conftree;        /**< back pointer to conftree */

	/* node attributes */
	char *name;                        /**< node name (dynamic alloc) */
	char *prompt;                      /**< node prompt (dynamic alloc) */
	char *help;                        /**< node help (dynamic alloc) */
	unsigned flags;                    /**< node flags */
	struct cfzy_list_head *expr_deps;  /**< list of prereq expressions */

	/* where it was parsed */
	char *filename;                    /**< name of conftree file */
	int linenum;                       /**< line where node was declared */

	/* specific to a node type */
	const struct cfzy_confnode_ops *ops;   /**< specific ops on node */
	void *private;

	/* values */
	char *default_value; /**< default value for this node */
	struct cfzy_list_head *default_val_list; /* conditional default vals */
	char *user_value;             /**< value wanted by the user */
	char *effective_value;        /**< real value, taking care of deps */
	char *old_user_value;         /**< saved user value */
	char *old_effective_value;    /**< saved effective value */

	/* allow to chain this node in a priority ordered list */
	TAILQ_ENTRY(cfzy_confnode) prio_next;  /**< next in ordered list */
	int in_prio_list;             /**< true if present in prio list */

	struct cfzy_list_head *node_deps; /**< list of nodes that must be
					   * evaluated before this one */
};

/**
 * Get the name of a node
 *
 * @param n
 *   configuration node
 * @return
 *   pointer to the node name (hosted in the node structure)
 */
static inline const char *cfzy_confnode_name(const struct cfzy_confnode *n)
{
	if (n->name == NULL)
		return "no-name";
	return n->name;
}

/**
 * Allocate a new node
 *
 * Allocate a new empty node, with no specific ops.
 *
 * @param conftree
 *   configuration tree
 * @return
 *   the allocated configuration node, or NULL on error
 */
struct cfzy_confnode *cfzy_confnode_alloc(struct cfzy_conftree *conftree);

/**
 * Free the node given as argument, and all its associated data
 *
 * This function will free the children of the nodes, then the node
 * itself, first by calling the specific function n->ops->free() if
 * defined, then by doing the generic operations.
 *
 * @param n
 *   configuration node
 */
void cfzy_confnode_free(struct cfzy_confnode *n);

/**
 * Get the prerequisites for this node
 *
 * Return the list of nodes required to evaluate the value of the
 * given node. Indeed, each node has a 'expr_deps' field storing a
 * list of expressions and a list of default value associated to
 * expressions. This function returns the list of all nodes referenced
 * in these expressions and all the ancestor nodes.
 *
 * The returned list must be freed by the user with
 * cfzy_list_free(list, NULL).
 *
 * @param n
 *   configuration node
 * @return
 *   list of nodes, or NULL on error
 */
struct cfzy_list_head *
cfzy_confnode_get_deplist(const struct cfzy_confnode *n);

/**
 * Add a conditional default value to a node
 *
 * @param n
 *   configuration node
 * @param val
 *   value to be used as default value
 * @param expr_buf
 *   condition string to enable the default value
 * @return
 *   0 on success, negative on error
 */
int cfzy_confnode_add_defval(struct cfzy_confnode *n, const char *val,
			     const char *expr_buf);

/**
 * Parse a line of configuration tree file, expecting an attribute
 *
 * This function first tries to match specific node attributes, using
 * the specific operation n->ops->add_attr() if registered. Then it
 * tries to match generic attributes (prompt, requires, default).
 *
 * This function is internal to the confizery library (called by the
 * configuration tree parser).
 *
 * @param n
 *   configuration node
 * @param tklist
 *   list of tokens for this line
 * @return
 *   SUCCESS if it matches an attribute of this node, NO_MATCH if the line
 *   is not an attribute for this node, ERROR on error
 */
enum cfzy_parse_return
cfzy_confnode_add_attr(struct cfzy_confnode *n,
		       const struct cfzy_token_list *tklist);

/**
 * Write .config-like configuration in a file
 *
 * Write in a file the configuration values of this node and its
 * descendants. The format of this file is like .config: it can be
 * sourced by a shell or a Makefile.
 *
 * The function calls the specific node operation
 * n->ops->dotconfig_write() if defined. Else, it uses the defaut
 * behavior, which varies depending on node flags.
 *
 * This function is internal and is called by the public function
 * cfzy_dotconfig_write().
 *
 * @param n
 *   configuration node
 * @param f
 *   open file with write permissions where config will be written
 * @return
 *   0 on success, negative on error
 */
int cfzy_confnode_dotconfig_write(const struct cfzy_confnode *n, FILE *f);

/**
 * Write the configuration in a C header file
 *
 * Write in a file the configuration values of this node and its
 * descendants. The format of this file is a C header, so it can be
 * included by a C application.
 *
 * The function calls the specific node operation
 * n->ops->c_hdr_write() if defined. Else, it uses the defaut
 * behavior, which varies depending on node flags.
 *
 * This function is internal and is called by the public function
 * cfzy_c_hdr_write().
 *
 * @param n
 *   configuration node
 * @param f
 *   open file with write permissions where config will be written
 * @return
 *   0 on success, negative on error
 */
int cfzy_confnode_c_hdr_write(const struct cfzy_confnode *n, FILE *f);

/**
 * Return a string identifying the node type
 *
 * The function calls the specific node operation
 * n->ops->get_type_str().
 *
 * The returned string points to static memory and contains the type
 * of the node (ex: "config", "menuconfig", "choice", ...)
 *
 * @param n
 *   configuration node
 * @return
 *   pointer to a node type string
 */
const char *cfzy_confnode_get_type_str(const struct cfzy_confnode *n);

/**
 * Return the boolean value of the node given the string value
 *
 * Convert the string value given as parameter to a boolean value. It
 * calls the specific node operation n->ops->str2bool() if
 * defined. Else, it uses the defaut behavior: return 0 for value=NULL
 * or 1 for any other value.
 *
 * The specific function (if defined) can also return a negative value
 * to indicate that the given value is not valid for this node.
 *
 * @param n
 *   configuration node
 * @param value
 *   string containing the value
 * @return
 *   the boolean value (0 or 1) if the string value is valid,
 *   else -1
 */
int cfzy_confnode_str2bool(const struct cfzy_confnode *n, const char *value);

/**
 * Get the path of the node
 *
 * @param n
 *   configuration node
 * @param buf
 *   buffer where the path should be written
 * @param len
 *   length of the buffer
 * @return
 *   0 on success, negative on error
 */
int cfzy_confnode_path(const struct cfzy_confnode *n, char *buf, int len);

/* dump modes */
#define CFZY_DUMP_MODE_MINIMAL        0x0
#define CFZY_DUMP_MODE_STANDARD       0x1
#define CFZY_DUMP_MODE_FULL           0x2
#define CFZY_DUMP_MODE_FULL_RECURSIVE 0x3

/**
 * Dump a human-readable configuration in a file
 *
 * Write in a file the structure fields of this node and its
 * descendants.
 *
 * @param n
 *   configuration node
 * @param f
 *   open file with write permissions where configuration will be written
 * @param mode
 *   behavior of dump (CFZY_DUMP_MODE_MINIMAL, CFZY_DUMP_MODE_STANDARD, ...)
 * @return
 *   0 on success, negative on error
 */
int cfzy_confnode_dump(const struct cfzy_confnode *n, FILE *f, int mode);

/**
 * Check prerequisites for a node
 *
 * @param n
 *   configuration node to be checked
 * @return
 *  0 if a parent node is disabled or if a prerequisite expression is
 *  false, 1 if node can be enable. On error, return -1.
 */
int cfzy_confnode_check_deps(struct cfzy_confnode *n);

/**
 * Evaluate the effective value of a node
 *
 * Set the new value of n->effective_value, assuming all prerequisites
 * are already evaluated. This function is used as a callback by
 * cfzy_expr_parse() when parsing an expression for
 * cfzy_confnode_evaluate().
 *
 * @return n
 *   configuration node
 * @return
 *   0 on success, negative on error
 */
int cfzy_confnode_evaluate(struct cfzy_confnode *n);

/**
 * Get the boolean value of a node, given its name
 *
 * Return the effective value of a node (assuming it has been
 * evaluated previously).
 *
 * @param name
 *   configuration node name
 * @param conftree
 *   configuration tree
 * @return
 *   the boolean value (0 or 1) if the string value is valid,
 *   else -1
 */
int cfzy_confnode_get_boolvalue(const char *name,
				const struct cfzy_conftree *conftree);

/**
 * Set user value of a configuration node
 *
 * To set the user value of a node, the user must not access to the
 * field directly but use this function. Indeed, a node may want to do
 * a specific operation (like a check, or modifying children nodes)
 * stored in n->ops->set_uservalue().
 *
 * @return n
 *   configuration node
 * @param value
 *   value to be set as user value
 * @return
 *   0 on success, negative on error
 */
int cfzy_confnode_set_uservalue(struct cfzy_confnode *n, const char *value);

/**
 * Create a new root node
 *
 * @param conftree
 *   configuration tree
 * @return
 *   a pointer to the configuration node, or NULL on error
 */
struct cfzy_confnode *cfzy_confnode_root_new(struct cfzy_conftree *conftree);

/**
 * Parse a line of configuration tree file, expecting new choice node
 *
 * @param n
 *   configuration node
 * @param tklist
 *   list of tokens for this line
 * @return
 *   SUCCESS if it matches an attribute of this node, NO_MATCH if the line
 *   is not an attribute for this node, ERROR on error
 */
enum cfzy_parse_return
cfzy_confnode_choice_new(struct cfzy_confnode *n,
			 const struct cfzy_token_list *tklist);
/**
 * Parse a line of configuration tree file, expecting new choiceconfig node
 *
 * @param n
 *   configuration node
 * @param tklist
 *   list of tokens for this line
 * @return
 *   SUCCESS if it matches an attribute of this node, NO_MATCH if the line
 *   is not an attribute for this node, ERROR on error
 */
enum cfzy_parse_return
cfzy_confnode_choiceconfig_new(struct cfzy_confnode *n,
			       const struct cfzy_token_list *tklist);

/**
 * Parse a line of configuration tree file, expecting new comment node
 *
 * @param n
 *   configuration node
 * @param tklist
 *   list of tokens for this line
 * @return
 *   SUCCESS if it matches an attribute of this node, NO_MATCH if the line
 *   is not an attribute for this node, ERROR on error
 */
enum cfzy_parse_return
cfzy_confnode_comment_new(struct cfzy_confnode *n,
			  const struct cfzy_token_list *tklist);

/**
 * Parse a line of configuration tree file, expecting new config node
 *
 * @param n
 *   configuration node
 * @param tklist
 *   list of tokens for this line
 * @return
 *   SUCCESS if it matches an attribute of this node, NO_MATCH if the line
 *   is not an attribute for this node, ERROR on error
 */
enum cfzy_parse_return
cfzy_confnode_config_new(struct cfzy_confnode *n,
			 const struct cfzy_token_list *tklist);

/**
 * Parse a line of configuration tree file, expecting new if node
 *
 * @param n
 *   configuration node
 * @param tklist
 *   list of tokens for this line
 * @return
 *   SUCCESS if it matches an attribute of this node, NO_MATCH if the line
 *   is not an attribute for this node, ERROR on error
 */
enum cfzy_parse_return
cfzy_confnode_if_new(struct cfzy_confnode *n,
		     const struct cfzy_token_list *tklist);

/**
 * Parse a line of configuration tree file, expecting new intconfig node
 *
 * @param n
 *   configuration node
 * @param tklist
 *   list of tokens for this line
 * @return
 *   SUCCESS if it matches an attribute of this node, NO_MATCH if the line
 *   is not an attribute for this node, ERROR on error
 */
enum cfzy_parse_return
cfzy_confnode_intconfig_new(struct cfzy_confnode *n,
			    const struct cfzy_token_list *tklist);

/**
 * Parse a line of configuration tree file, expecting new menu node
 *
 * @param n
 *   configuration node
 * @param tklist
 *   list of tokens for this line
 * @return
 *   SUCCESS if it matches an attribute of this node, NO_MATCH if the line
 *   is not an attribute for this node, ERROR on error
 */
enum cfzy_parse_return
cfzy_confnode_menu_new(struct cfzy_confnode *n,
		       const struct cfzy_token_list *tklist);

/**
 * Parse a line of configuration tree file, expecting new menuconfig node
 *
 * @param n
 *   configuration node
 * @param tklist
 *   list of tokens for this line
 * @return
 *   SUCCESS if it matches an attribute of this node, NO_MATCH if the line
 *   is not an attribute for this node, ERROR on error
 */
enum cfzy_parse_return
cfzy_confnode_menuconfig_new(struct cfzy_confnode *n,
			     const struct cfzy_token_list *tklist);

/**
 * Parse a line of configuration tree file, expecting new strconfig node
 *
 * @param n
 *   configuration node
 * @param tklist
 *   list of tokens for this line
 * @return
 *   SUCCESS if it matches an attribute of this node, NO_MATCH if the line
 *   is not an attribute for this node, ERROR on error
 */
enum cfzy_parse_return
cfzy_confnode_strconfig_new(struct cfzy_confnode *n,
			    const struct cfzy_token_list *tklist);

/**
 * Close the node beeing parsed
 *
 * Some nodes can contains other nodes (menu, menuconfig, ...). When
 * the parser has finished to parse the descendants of the node, it
 * calls this function to close the node. This can trigger some
 * specific operations (like checks) if the node type registers a
 * n->ops->close(). If no specific function is registered, it does
 * nothing.
 *
 * This function is internal to the confizery library (called by the
 * configuration tree parser).
 *
 * @param n
 *   configuration node
 * @return
 *   0 on success, negative on error
 */
int cfzy_confnode_close(struct cfzy_confnode *n);


#endif /* _CFZY_CONFNODE_H_ */
