/*
 * Copyright (c) 2010, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/queue.h>

#include "cfzy_log.h"
#include "cfzy_list.h"
#include "cfzy_expr.h"
#include "cfzy_confnode.h"

#define LOG(level, fmt, args...)				\
	CFZY_LOG("confnode_choice", level, fmt, ##args)

/* Return a string identifying the node type ("config", "menuconfig",
 * "choice", ...) */
static const char *choice_get_type_str(const struct cfzy_confnode *n)
{
	(void)n;
	return "choice";
}

static int choice_str2bool(const struct cfzy_confnode *n, const char *value)
{
	struct cfzy_confnode *c;

	/* NULL means "disabled" */
	if (value == NULL)
		return 0;

	/* if we have no children, it's because we are creating the
	 * node, so accept all values, we will check them later in
	 * choice_close() */
	if (TAILQ_EMPTY(&n->children))
		return 1;

	/* else, the value must be the name of a children */
	TAILQ_FOREACH(c, &n->children, child_next) {
		if (!strcmp(value, c->name))
			return 1;
	}

	/* invalid value */
	return -1;
}

static int choice_close(struct cfzy_confnode *n)
{
	struct cfzy_confnode_defvalue *defval;
	struct cfzy_list_elt *e;
	struct cfzy_confnode *c;
	char exprbuf[512];
	int ret;

	/* browse all default values and call choice_str2bool() on
	 * them to check that they are correct */
	TAILQ_FOREACH(e, n->default_val_list, next) {
		defval = e->ptr;

		if (choice_str2bool(n, defval->val) < 0) {
			LOG(ERR, "invalid default value <%s> for node <%s>\n",
			    defval->val, n->name);
			return -1;
		}
	}
	if (choice_str2bool(n, n->default_value) < 0) {
		LOG(ERR, "invalid default value <%s> for node <%s>\n",
		    defval->val, n->name);
		return -1;
	}

	/* for each child node */
	TAILQ_FOREACH(c, &n->children, child_next) {

		/* set default value */
		if (!strcmp(n->default_value, c->name))
			c->default_value = strdup("y");
		else
			c->default_value = strdup("n");

		if (c->default_value == NULL) {
			LOG(ERR, "cannot allocate default value>\n");
			return -1;
		}

		/* set conditional default values for child nodes with
		 * the same condition */
		TAILQ_FOREACH(e, n->default_val_list, next) {
			defval = e->ptr;

			if (cfzy_expr_to_str(defval->expr, exprbuf,
					     sizeof(exprbuf)) < 0) {
				LOG(ERR, "cannot convert expression to str\n");
				return -1;
			}

			if (!strcmp(defval->val, c->name))
				ret = cfzy_confnode_add_defval(c, "y", exprbuf);
			else
				ret = cfzy_confnode_add_defval(c, "n", exprbuf);

			if (ret < 0) {
				LOG(ERR, "cannot add conditional default\n");
				return -1;
			}
		}

	}

	return 0;
}

static int choice_set_uservalue(struct cfzy_confnode *n, const char *value)
{
	struct cfzy_confnode *c;
	char *newvalue;

	/* if value is NULL, unset this node and all its children,
	 * falling back to default value */
	if (value == NULL) {
		/* this node */
		if (n->user_value != NULL)
			free(n->user_value);
		n->user_value = NULL;

		/* children */
		TAILQ_FOREACH(c, &n->children, child_next) {
			if (c->user_value != NULL)
				free(c->user_value);
			c->user_value = NULL;
		}
		return 0;
	}

	/* if not NULL, the value must be the name of a children */
	TAILQ_FOREACH(c, &n->children, child_next) {
		if (!strcmp(value, c->name))
			break;
	}
	if (c == NULL)
		return -1;


	/* this node */
	if (n->user_value != NULL)
		free(n->user_value);
	n->user_value = strdup(c->name);
	if (n->user_value == NULL) {
		LOG(ERR, "cannot allocate value\n");
		return -1;
	}

	/* children */
	TAILQ_FOREACH(c, &n->children, child_next) {
		if (!strcmp(value, c->name))
			newvalue = strdup("y");
		else
			newvalue = strdup("n");

		if (newvalue == NULL) {
			LOG(ERR, "cannot allocate value\n");
			return -1;
		}

		if (c->user_value != NULL)
			free(c->user_value);
		c->user_value = newvalue;
	}

	return 0;
}

static struct cfzy_confnode_ops choice_ops = {
	.free = NULL,
	.add_attr = NULL,
	.close = choice_close,
	.dotconfig_write = NULL,
	.c_hdr_write = NULL,
	.str2bool = choice_str2bool,
	.get_type_str = choice_get_type_str,
	.set_uservalue = choice_set_uservalue,
};

/* Create a new node */
enum cfzy_parse_return cfzy_confnode_choice_new(struct cfzy_confnode *n,
						const struct cfzy_token_list *tklist)
{
	struct cfzy_token *tok;
	tok = TAILQ_FIRST(&tklist->list);

	if (strcmp(tok->str, "choice") || tklist->n_token != 2)
		return NO_MATCH;

	tok = TAILQ_NEXT(tok, next);
	n->ops = &choice_ops;
	n->flags = CFZY_F_IS_DIR | CFZY_F_QUOTE_VAL;

	n->name = strdup(tok->str);
	if (n->name == NULL)
		return ERROR;

	return SUCCESS;
}
