/*
 * Copyright (c) 2010, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/queue.h>

#include "cfzy_log.h"
#include "cfzy_expr.h"
#include "cfzy_confnode.h"

#define LOG(level, fmt, args...)				\
	CFZY_LOG("confnode_choiceconfig", level, fmt, ##args)

/* Return a string identifying the node type ("config", "menuconfig",
 * "choice", ...) */
static const char *choiceconfig_get_type_str(const struct cfzy_confnode *n)
{
	(void)n;
	return "choiceconfig";
}

static int choiceconfig_str2bool(const struct cfzy_confnode *n,
				 const char *value)
{
	(void)n;

	/* NULL or "n" means disabled */
	if (value == NULL || !strcmp("n", value))
		return 0;

	if (!strcmp("y", value))
		return 1;

	return -1;
}

static int choiceconfig_set_uservalue(struct cfzy_confnode *n,
				      const char *value)
{
	struct cfzy_confnode *c;

	if (value == NULL)
		return cfzy_confnode_set_uservalue(n->parent, NULL);

	if (!strcmp(value, "y"))
		return cfzy_confnode_set_uservalue(n->parent, n->name);

	if (!strcmp(value, "n")) {
		/* node is already set to "n", do nothing so we won't
		 * change the selected choice */
		if (!strcmp(n->user_value, "n"))
			return 0;

		/* else just set to 'y' the first brother that we find */
		TAILQ_FOREACH(c, &n->children, child_next) {
			if (c != n)
				break;
		}

		/* no brother found */
		if (c == NULL)
			return -1;

		return cfzy_confnode_set_uservalue(c, "y");
	}

	return -1;
}

static struct cfzy_confnode_ops choiceconfig_ops = {
	.free = NULL,
	.add_attr = NULL,
	.close = NULL,
	.str2bool = choiceconfig_str2bool,
	.dotconfig_write = NULL,
	.c_hdr_write = NULL,
	.get_type_str = choiceconfig_get_type_str,
	.set_uservalue = choiceconfig_set_uservalue,
};

/* Create a new node */
enum cfzy_parse_return
cfzy_confnode_choiceconfig_new(struct cfzy_confnode *n,
			       const struct cfzy_token_list *tklist)
{
	struct cfzy_token *tok;

	tok = TAILQ_FIRST(&tklist->list);

	if (strcmp(tok->str, "choiceconfig") || tklist->n_token != 2)
		return NO_MATCH;

	tok = TAILQ_NEXT(tok, next);
	n->ops = &choiceconfig_ops;
	n->flags = CFZY_F_NO_SET_DEFAULT | CFZY_F_VAL_IS_BOOL;

	n->name = strdup(tok->str);
	if (n->name == NULL)
		return ERROR;

	return SUCCESS;
}
