/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file
 * Confizery configuration node specific operations
 *
 * This file describes specific node operations, called internally by
 * cfzy_confnode module.
 */

#ifndef _CFZY_CONFNODE_OPS_H_
#define _CFZY_CONFNODE_OPS_H_

#include <stdio.h>

#include "cfzy_conftree_parser.h"

struct cfzy_confnode;
struct cfzy_token_list;

/**
 * Node operation to free the specific node data. Called by
 * cfzy_confnode_free().
 */
typedef void (confnode_free_t)(struct cfzy_confnode *n);

/**
 * Specific node operation to parse a conftree line, expecting a new
 * attribute. Called by cfzy_confnode_add_attr().
 */
typedef enum cfzy_parse_return
(confnode_add_attr_t)(struct cfzy_confnode *n,
		      const struct cfzy_token_list *tklist,
		      const char *linebuf);

/**
 * Specific node operation to close a node beeing parsed. Called by
 * cfzy_confnode_close().
 */
typedef int (confnode_close_t)(struct cfzy_confnode *n);

/**
 * Specific node operation that writes the configuration in a
 * file. Called by cfzy_confnode_dotconfig_write().
 */
typedef int (confnode_dotconfig_write_t)(const struct cfzy_confnode *n, FILE *f);

/**
 * Specific node operation that writes the configuration in a C header
 * file. Called by cfzy_confnode_c_hdr_write().
 */
typedef int (confnode_c_hdr_write_t)(const struct cfzy_confnode *n, FILE *f);

/**
 * Specific node operation that returns the boolean value of the node
 * given the string value. Called by cfzy_confnode_str2bool().
 */
typedef int (confnode_str2bool_t)(const struct cfzy_confnode *n,
				  const char *strvalue);

/**
 * Specific node operation that returns a string identifying the node
 * type. Called by cfzy_confnode_get_type_str().
 */
typedef const char *(confnode_get_type_str_t)(const struct cfzy_confnode *n);

/**
 * Specific node operation to set user value of a configuration
 * node. Called by cfzy_confnode_set_uservalue().
 */
typedef int (confnode_set_uservalue_t)(struct cfzy_confnode *n,
				       const char *strvalue);

/* all node-specific operations */
struct cfzy_confnode_ops {
	confnode_free_t *free;                 /**< free specific node data */
	confnode_add_attr_t *add_attr;         /**< try to parse specific attr */
	confnode_close_t *close;               /**< close node beeing parsed */
	confnode_dotconfig_write_t *dotconfig_write; /**< write in dotconfig */
	confnode_c_hdr_write_t *c_hdr_write;   /**< write in C header file */
	confnode_str2bool_t *str2bool;         /**< check strvalue and
						  return boolvalue */
	confnode_get_type_str_t *get_type_str; /**< return type string */
	confnode_set_uservalue_t *set_uservalue; /**< check and set user val */
};

#endif /* _CFZY_CONFNODE_OPS_H_ */
