/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "cfzy_log.h"
#include "cfzy_list.h"
#include "cfzy_htable.h"
#include "cfzy_confnode.h"
#include "cfzy_conftree.h"
#include "cfzy_conftree_parser.h"

#define LOG(level, fmt, args...)				\
	CFZY_LOG("conftree", level, fmt, ##args)

/* Parse a configuration tree and add all the nodes in the
 * hashtable. Return the number of nodes or -1 on error. */
static int conftree_fill_htable(struct cfzy_conftree *conftree,
				struct cfzy_confnode *n)
{
	struct cfzy_confnode *c;
	int ret, count;

	if (n->name != NULL) {
		if (cfzy_htable_lookup(conftree->nodes_htable, n->name) != NULL) {
			LOG(ERR, "duplicate symbol <%s>\n", n->name);
			return -1;
		}

		if (cfzy_htable_add(conftree->nodes_htable, n->name, n) < 0)
			return -1;
	}
	count = 1; /* count the nodes recursively */

	TAILQ_FOREACH(c, &n->children, child_next) {
		ret = conftree_fill_htable(conftree, c);
		if (ret < 0)
			return -1;
		count += ret;
	}

	return count;
}

/* Parse a configuration tree and add all the nodes in the hashtable */
static int conftree_generate_deplist(struct cfzy_confnode *n)
{
	struct cfzy_confnode *c;

	n->node_deps = cfzy_confnode_get_deplist(n);
	if (n->node_deps == NULL)
		return -1;

	TAILQ_FOREACH(c, &n->children, child_next) {
		if (conftree_generate_deplist(c) < 0)
			return -1;
	}

	return 0;
}

/* Parse a configuration tree and add all the nodes in the
 * hashtable. Return the number of nodes in the prio_list. */
static int conftree_fill_prio_list(struct cfzy_conftree *conftree,
				   struct cfzy_confnode *n)
{
	struct cfzy_list_elt *e;
	struct cfzy_confnode *c;
	int ok = 1;
	int count = 0, ret;

	if (n->in_prio_list == 1) {
		count = 1;
	}
	else {
		/* if the node is not in prio_list, check if all
		 * prerequisites are in prio list, and add in the
		 * list */
		TAILQ_FOREACH(e, n->node_deps, next) {
			c = e->ptr;
			if (c->in_prio_list == 0) {
				ok = 0;
				break;
			}
		}
		if (ok == 1) {
			TAILQ_INSERT_TAIL(&conftree->prio_list, n, prio_next);
			count = 1;
			n->in_prio_list = 1;
		}
	}

	/* process children */
	TAILQ_FOREACH(c, &n->children, child_next) {
		ret = conftree_fill_prio_list(conftree, c);
		if (ret < 0)
			return -1;
		count += ret;
	}

	return count;
}

/* create a new configuration tree from filename */
struct cfzy_conftree *cfzy_conftree_new(const char *filename)
{
	struct cfzy_conftree *conftree;
	int node_count, prio_count, prev;

	LOG(INFO, "Open conftree <%s>\n", filename);

	conftree = malloc(sizeof(*conftree));
	if (conftree == NULL)
		return NULL;
	memset(conftree, 0, sizeof(*conftree));

	conftree->path = strdup(filename);
	if (conftree->path == NULL) {
		cfzy_conftree_free(conftree);
		return NULL;
	}

	conftree->root = cfzy_confnode_root_new(conftree);
	if (conftree->root == NULL) {
		cfzy_conftree_free(conftree);
		return NULL;
	}

	/* parse the configuration tree file */
	if (cfzy_conftree_parse(filename, conftree) < 0) {
		LOG(ERR, "cannot parse configuration tree\n");
		cfzy_conftree_free(conftree);
		return NULL;
	}

	conftree->nodes_htable = cfzy_htable_alloc();
	if (conftree->nodes_htable == NULL) {
		LOG(ERR, "cannot allocate conftree htable\n");
		cfzy_conftree_free(conftree);
		return NULL;
	}

	/* fill a htable indexed by node name */
	node_count = conftree_fill_htable(conftree, conftree->root);
	if (node_count < 0) {
		LOG(ERR, "cannot fill conftree htable\n");
		cfzy_conftree_free(conftree);
		return NULL;
	}
	conftree->count = node_count;

	/* for each node, generate the list of prerequisite nodes */
	if (conftree_generate_deplist(conftree->root) < 0) {
		LOG(ERR, "cannot generate dep list\n");
		cfzy_conftree_free(conftree);
		return NULL;
	}

	/* generate a list ordered by node priority: if Node1 depends
	 * on Node2, then Node1 is located after Node2 in the list. */
	TAILQ_INIT(&conftree->prio_list);
	prev = 0;
	while (1) {
		prio_count = conftree_fill_prio_list(conftree, conftree->root);
		if (prio_count < 0) {
			LOG(ERR, "cannot generate prio list\n");
			cfzy_conftree_free(conftree);
			return NULL;
		}
		if (prio_count == node_count)
			break;
		if (prio_count == prev) {
			LOG(ERR, "circular dependency in conf tree\n");
			cfzy_conftree_free(conftree);
			return NULL;
		}
		prev = prio_count;
	}

	return conftree;
}

/* free a configuration tree */
void cfzy_conftree_free(struct cfzy_conftree *conftree)
{
	if (conftree->path != NULL)
		free(conftree->path);
	if (conftree->nodes_htable != NULL)
		cfzy_htable_free(conftree->nodes_htable, NULL);
	if (conftree->root != NULL)
		cfzy_confnode_free(conftree->root);
	free(conftree);
}

int cfzy_conftree_dump(const struct cfzy_conftree *conftree,
			const char *filename)
{
	FILE *f;

	f = fopen(filename, "w");
	if (f == NULL) {
		LOG(ERR, "Cannot open <%s>: %s\n", filename, strerror(errno));
		return -1;
	}

	if (cfzy_confnode_dump(conftree->root, f,
			       CFZY_DUMP_MODE_FULL_RECURSIVE) < 0) {
		LOG(ERR, "Cannot dump confnode tree\n");
		fclose(f);
		return -1;
	}

	fclose(f);
	return 0;
}

int cfzy_conftree_evaluate(struct cfzy_conftree *conftree)
{
	struct cfzy_confnode *n;

	TAILQ_FOREACH(n, &conftree->prio_list, prio_next) {
		if (cfzy_confnode_evaluate(n) < 0) {
			LOG(ERR, "Cannot evaluate conftree\n");
			return -1;
		}
		LOG(DEBUG, "eval %s -> %s\n", n->name, n->effective_value);
	}

	return 0;
}

/* same than strcmp() but allow string to be NULL */
static int strcmp2(const char *s1, const char *s2)
{
	if (s1 == NULL && s2 == NULL)
		return 0;
	if (s1 == NULL)
		return -1;
	if (s2 == NULL)
		return 1;
	return strcmp(s1, s2);
}

static int __filter(struct cfzy_confnode *n, struct cfzy_list_head *node_list,
		    int flags)
{
	int add_it = 0;
	struct cfzy_confnode *c;

	if ((flags & CFZY_FILTER_F_USR_UNSET) && n->user_value == NULL)
		add_it = 1;
	else if ((flags & CFZY_FILTER_F_EFF_UNSET) && n->effective_value == NULL)
		add_it = 1;
	else if ((flags & CFZY_FILTER_F_USR_CHANGED) &&
		 strcmp2(n->user_value, n->old_user_value))
		add_it = 1;
	else if ((flags & CFZY_FILTER_F_EFF_CHANGED) &&
		 strcmp2(n->effective_value, n->old_effective_value))
		add_it = 1;

	if (add_it == 1) {
		if (cfzy_list_add_tail(node_list, n) < 0) {
			LOG(ERR, "cannot add node in list\n");
			return -1;
		}
	}

	/* children nodes */
	TAILQ_FOREACH(c, &n->children, child_next) {
		if (__filter(c, node_list, flags) < 0)
			return -1;
	}

	return 0;
}

/* Get a list of nodes matching conditions (for instance the nodes
 * that have no user value) */
struct cfzy_list_head *
cfzy_conftree_filter(struct cfzy_conftree *conftree, int flags)
{
	struct cfzy_list_head *node_list = NULL;

	node_list = cfzy_list_alloc();
	if (node_list == NULL) {
		LOG(ERR, "cannot allocate node list\n");
		return NULL;
	}

	if (__filter(conftree->root, node_list, flags) < 0) {
		LOG(ERR, "cannot filter nodes\n");
		cfzy_list_free(node_list, NULL);
		return NULL;
	}

	return node_list;
}

static int __save_values(struct cfzy_confnode *n)
{
	struct cfzy_confnode *c;

	/* duplicate user value */
	if (n->old_user_value != NULL) {
		free(n->old_user_value);
		n->old_user_value = NULL;
	}
	if (n->user_value != NULL) {
		n->old_user_value = strdup(n->user_value);
		if (n->old_user_value == NULL) {
			LOG(ERR, "cannot duplicate value\n");
			return -1;
		}
	}

	/* duplicate effective value */
	if (n->old_effective_value != NULL) {
		free(n->old_effective_value);
		n->old_effective_value = NULL;
	}
	if (n->effective_value != NULL) {
		n->old_effective_value = strdup(n->effective_value);
		if (n->old_effective_value == NULL) {
			LOG(ERR, "cannot duplicate value\n");
			return -1;
		}
	}

	/* do the same on children */
	TAILQ_FOREACH(c, &n->children, child_next) {
		if (__save_values(c) < 0)
			return -1;
	}

	return 0;
}

int cfzy_conftree_save_values(struct cfzy_conftree *conftree)
{
	return __save_values(conftree->root);
}
