/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file
 * Confizery configuration tree
 *
 * This module defines a structure cfzy_conftree that stores the full
 * configuration tree (composed of cfzy_confnode structures), and some
 * meta data like the number of nodes in the tree, a hash table used
 * to do a node lookup by its name, the path of the configuration tree
 * file, ...
 */

#ifndef _CFZY_CONFTREE_H_
#define _CFZY_CONFTREE_H_

#include "cfzy_confnode.h"

struct cfzy_htable;

/**
 * Structure storing the config tree
 */
struct cfzy_conftree {
	char *path;                       /**< path of conftree file */
	struct cfzy_confnode *root;       /**< pointer to root node */
	int count;                        /**< number of nodes */
	struct cfzy_htable *nodes_htable; /**< htable of nodes, indexed by name */
	struct cfzy_confnode_list prio_list; /**< ordered list of nodes */
};

/**
 * Create a new configuration tree
 *
 * Parse a configuration tree and allocate a new conftree
 * structure. The structure must be freed with cfzy_conftree_free().
 *
 * @param filename
 *   name of the configuration tree file
 * @return
 *   the allocated conftree structure, or NULL on error
 */
struct cfzy_conftree *cfzy_conftree_new(const char *filename);

/**
 * Free a configuration tree structure
 *
 * Free a configuration tree structure previously allocated with
 * cfzy_conftree_new().
 *
 * @param conftree
 *   the configuration tree structure to be freed
 */
void cfzy_conftree_free(struct cfzy_conftree *conftree);

/**
 * Dump a configuration tree structure in a file
 *
 * @param conftree
 *   the configuration tree structure to be freed
 * @param filename
 *   file where conf should be dumped
 * @return
 *   0 on success, negative on error
 */
int cfzy_conftree_dump(const struct cfzy_conftree *conftree,
		       const char *filename);

/**
 * Evaluate a configuration tree structure
 *
 * Parse all nodes following priority list and set the effective_value
 * field by calling cfzy_confnode_evaluate() which takes care of
 * default values, user values, and dependencies.
 *
 * @param conftree
 *   the configuration tree structure
 * @return
 *   0 on success, negative on error
 */
int cfzy_conftree_evaluate(struct cfzy_conftree *conftree);

/**
 * Save values of configuration tree
 *
 * For each node, copy the effective_value field in
 * old_effective_value and the user_value field in old_user_value.
 * These values can be used later with cfzy_conftree_filter() to
 * compare the old values with the new ones.
 *
 * @param conftree
 *   the configuration tree structure
 * @return
 *   0 on success, negative on error
 */
int cfzy_conftree_save_values(struct cfzy_conftree *conftree);

/* flags to be used with cfzy_conftree_filter() */
#define CFZY_FILTER_F_USR_UNSET     0x01
#define CFZY_FILTER_F_EFF_UNSET     0x02
#define CFZY_FILTER_F_USR_CHANGED   0x04
#define CFZY_FILTER_F_EFF_CHANGED   0x08

/**
 * Get a list of nodes matching conditions
 *
 * Return the list of nodes in configuration tree that match
 * conditions. Example: cfzy_conftree_filter(contree,
 * CFZY_FILTER_F_USR_UNSET) will return all nodes that have no user
 * value (set to NULL).
 *
 * @param conftree
 *   the configuration tree structure
 * @param flags
 *   list of conditions (CFZY_FILTER_F_*)
 * @return
 *   0 on success, negative on error
 */
struct cfzy_list_head *
cfzy_conftree_filter(struct cfzy_conftree *conftree, int flags);

#endif /* _CFZY_CONFTREE_H_ */
