/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <libgen.h>
#include <sys/param.h>
#include <sys/queue.h>

#include "cfzy_log.h"
#include "cfzy_string.h"
#include "cfzy_expr.h"
#include "cfzy_htable.h"
#include "cfzy_conftree.h"
#include "cfzy_confnode.h"
#include "cfzy_conftree_parser.h"

#define LOG(level, fmt, args...)				\
	CFZY_LOG("conftree_parser", level, fmt, ##args)

/*
 * Structure that stores the state of the configuration tree parser.
 */
struct conftree_parse_status {
	const char *filename;           /* name of file beeing parsed */
	FILE *f;                        /* pointer to file beeing parsed */
	int linenum;                    /* current line number in file */
	char *curline;                  /* pointer to current line, or
					   NULL if no current line */
	struct cfzy_token_list *tklist; /* pointer to current token
					   list of NULL if no current line */
	struct cfzy_confnode *parent;   /* current parent node */
};

static int __cfzy_conftree_parse(const char *filename,
				 struct cfzy_confnode **pparent);

/* return true if line is empty or contains only spaces/comments */
static int line_is_empty(const char *buf)
{
	while (*buf != '\0' && *buf != '#') {
		if (!isspace(*buf))
			return 0;
		buf++;
	}
	return 1;
}

/*
 * Add a token in the tklist structure given as argument.
 *
 * 'buf': pointer to the beginning of line buffer.
 * 'offset': offset in the line where start parsing
 *
 * Return the number of consumed bytes if a token is found
 * Return 0 if there is no more token
 * Return -1 on error
 */
static int
append_token(struct cfzy_token_list *tklist, const char *buf, int offset)
{
	struct cfzy_token *tok;
	const char *s;
	unsigned len, space_len = 0;
	char *retbuf;

	/* skip spaces */
	s = buf + offset;
	while (*s != '\0' && isspace(*s))
		s++;
	if (*s == '\0' || *s == '#')
		return 0;

	tok = malloc(sizeof(struct cfzy_token));
	if (tok == NULL) {
		LOG(ERR, "not enough memory\n");
		return -1;
	}

	space_len = s - buf - offset;
	tok->offset = s - buf;

	/* if it's a quote, unquote the string */
	if (*s == '"' || *s == '\'') {
		retbuf = cfzy_string_unquote(s, &len);
		if (retbuf == NULL)
			goto free;
	}
	/* else wait a space */
	else {
		retbuf = strdup(s);
		if (retbuf == NULL) {
			LOG(ERR, "not enough memory\n");
			goto free;
		}

		for (len = 0;
		     s[len] != '\0' && !isspace(s[len]) && s[len] != '#';
		     len ++)
			;

		retbuf[len] = '\0';
	}

	/* fill token structure and append it in tklist */
	tok->str = retbuf;
	TAILQ_INSERT_TAIL(&tklist->list, tok, next);
	tklist->n_token ++;

	return len + space_len;

 free:
	free(tok);
	return -1;
}

/* free a token list previously allocated with tokenize() */
static void free_token_list(struct cfzy_token_list *tklist)
{
	struct cfzy_token *tok;

	while ((tok = TAILQ_FIRST(&tklist->list)) != NULL) {
		TAILQ_REMOVE(&tklist->list, tok, next);
		free(tok->str);
		free(tok);
	}
	free(tklist);
}

/*
 * Parse the line contained in buffer 'linebuf'. This function fills a
 * tklist structure, composed of several tokens.
 * Return the tklist pointer on success or NULL on error.
 */
static struct cfzy_token_list *tokenize(const char *linebuf)
{
	struct cfzy_token_list *tklist;
	int ret, offset = 0;

	tklist = malloc(sizeof(struct cfzy_token_list));
	if (tklist == NULL) {
		LOG(ERR, "not enough memory\n");
		return NULL;
	}

	memset(tklist, 0, sizeof(struct cfzy_token_list));
	TAILQ_INIT(&tklist->list);

	/* read tokens from buf and append it to tklist structure */
	do {
		ret = append_token(tklist, linebuf, offset);
		if (ret < 0) {
			free_token_list(tklist);
			return NULL;
		}
		offset += ret;
	} while (ret != 0);

	tklist->linebuf = linebuf;
	return tklist;
}

/* duplicate buf, replacing variables like $(VAR) by their values */
char *replace_variables(const char *buf)
{
	char *s, *buf_dup, *var;
	const char *start;
	char *out = NULL;
	char c;

	buf_dup = strdup(buf);
	if (buf_dup == NULL) {
		LOG(ERR, "not enough memory\n");
		return NULL;
	}

	s = buf_dup;

	for ( ; s[0] != '\0' ; s++) {

		start = s;

		/* wait a "$(" */
		for ( ; s[0] != '\0' && (s[0] != '$' || s[1] != '('); s++);
		c = s[0];
		s[0] = '\0';
		if (cfzy_string_asprintf(&out, "%s", start) < 0)
			goto fail;

		if (c == '\0')
			break;

		s += 2;
		start = s;

		for ( ; s[0] != '\0' && s[0] != ')'; s++);
		c = s[0];
		if (c == '\0')
			goto fail;

		s[0] = '\0';
		var = getenv(start);

		if (var != NULL) {
			if (cfzy_string_asprintf(&out, var) < 0)
				goto fail;
		}

	}

	free(buf_dup);
	return out;

 fail:
	free(buf_dup);
	if (out != NULL)
		free(out);
	return NULL;
}

/*
 * Fill state structure with current line. If state->curline is
 * already set, the function does nothing, else it will read a new
 * line in the file and tokenize it. Return 0 on success, and -1 on
 * error. If state->curline is still NULL after a call to this
 * function, we reached the end of the file.
 */
static int get_curline(struct conftree_parse_status *state)
{
	char buf[BUFSIZ];
	int line_len;
	char *line;

	/* nothing to do */
	if (state->curline != NULL)
		return 0;

	state->curline = fgets(buf, sizeof(buf), state->f);

	/* end of file */
	if (state->curline == NULL) {
		LOG(DEBUG, "EOF\n");
		return 0;
	}

	state->linenum ++;
	line_len = strlen(state->curline);

	/* if line ends with '\', get another line */
	while (line_len >= 2 &&
	       state->curline[line_len - 1] == '\n' &&
	       state->curline[line_len - 2] == '\\') {
		if (line_len >= (int)sizeof(buf) - 1)
			break;

		line = fgets(buf + line_len - 2,
			     sizeof(buf) - line_len - 2, state->f);
		if (line == NULL) /* EOF */
			break;

		line_len = strlen(state->curline);
		state->linenum ++;
	}

	state->curline = replace_variables(buf);
	if (state->curline == NULL) {
		LOG(ERR, "Cannot eval variables\n");
		return -1;
	}

	state->tklist = tokenize(state->curline);
	if (state->tklist == NULL) {
		LOG(ERR, "Cannot parse line\n");
		free(state->curline);
		return -1;
	}

	return 0;
}

/*
 * This function is called when we processed the current line. It will
 * free allocated memory (token list) and set state->curline to NULL.
 */
static void eat_curline(struct conftree_parse_status *state)
{
	if (state->curline != NULL) {
		free(state->curline);
		state->curline = NULL;
	}
	if (state->tklist != NULL) {
		free_token_list(state->tklist);
		state->tklist = NULL;
	}
}

/* Try to parse node help. Return NO_MATCH if the current line does
 * not match the 'help' token, SUCCESS if help is successfully parsed,
 * or ERROR on error. */
static enum cfzy_parse_return
parse_help(struct cfzy_confnode *n, struct conftree_parse_status *state)
{
	struct cfzy_token_list *tklist;
	struct cfzy_token *tok;
	char *buf;

	if (get_curline(state) < 0)
		return ERROR;

	buf = state->curline;

	/* EOF */
	if (buf == NULL)
		return NO_MATCH;

	/* we want 1 token */
	tklist = state->tklist;
	if (tklist->n_token != 1)
		return NO_MATCH;

	/* we want a "help" token */
	tok = TAILQ_FIRST(&tklist->list);
	if (strcmp(tok->str, "help") && strcmp(tok->str, "---help---"))
		return NO_MATCH;

	LOG(INFO, "parse help\n");
	eat_curline(state);

	/* parse help content */
	while (1) {

		/* read next line */
		if (get_curline(state) < 0)
			return ERROR;

		buf = state->curline;

		/* EOF */
		if (buf == NULL)
			return SUCCESS;

		/* skip empty lines */
		if (line_is_empty(buf)) {
			eat_curline(state);
			continue;
		}

		/* if line does not start with space, this is this end
		   of help parsing */
		if (!isspace(*buf))
			return SUCCESS;

		/* append string in buffer, stripping starting spaces */
		tok = TAILQ_FIRST(&state->tklist->list);
		if (cfzy_string_asprintf(&n->help, "%s",
				  buf + tok->offset) < 0)
			return ERROR;

		eat_curline(state);
	}
}

/* Try to parse the "source" command. Return NO_MATCH if the current
 * line does not match the "source" token, SUCCESS if the file is
 * successfully sourced, or ERROR on error. */
static enum cfzy_parse_return source_file(struct conftree_parse_status *state)
{
	struct cfzy_token_list *tklist;
	struct cfzy_token *tok;

	if (get_curline(state) < 0)
		return ERROR;

	/* EOF */
	if (state->curline == NULL)
		return NO_MATCH;

	tklist = state->tklist;
	tok = TAILQ_FIRST(&tklist->list);

	/* syntax is: source FILE_NAME */
	if (tklist->n_token != 2 || strcmp(tok->str, "source"))
		return NO_MATCH;

	tok = TAILQ_NEXT(tok, next);
	LOG(INFO, "source <%s>\n", tok->str);

	/* parse new file */
	if (__cfzy_conftree_parse(tok->str, &state->parent) < 0) {
		LOG(ERR, "error in sourced file\n");
		return ERROR;
	}

	eat_curline(state);
	return SUCCESS;
}

/* Try to create a new node command ("config", "menuconfig", "choice",
 * "if", ...). Return NO_MATCH if the current line is not a new node
 * command, SUCCESS if a new node is created, or ERROR on error. This
 * function does not parse the attributes of the node but only the
 * first line.  */
static enum cfzy_parse_return
new_node(struct conftree_parse_status *state, struct cfzy_confnode **nodep)
{
	struct cfzy_token_list *tklist;
	enum cfzy_parse_return ret;

	if (get_curline(state) < 0)
		return ERROR;

	/* EOF */
	if (state->curline == NULL)
		return NO_MATCH;

	tklist = state->tklist;

	*nodep = cfzy_confnode_alloc(state->parent->conftree);
	if (*nodep == NULL) {
		LOG(ERR, "cannot allocate node\n");
		return ERROR;
	}
	(*nodep)->parent = state->parent;

	ret = cfzy_confnode_choice_new(*nodep, tklist);
	if (ret != NO_MATCH)
		return ret;
	ret = cfzy_confnode_choiceconfig_new(*nodep, tklist);
	if (ret != NO_MATCH)
		return ret;
	ret = cfzy_confnode_comment_new(*nodep, tklist);
	if (ret != NO_MATCH)
		return ret;
	ret = cfzy_confnode_config_new(*nodep, tklist);
	if (ret != NO_MATCH)
		return ret;
	ret = cfzy_confnode_if_new(*nodep, tklist);
	if (ret != NO_MATCH)
		return ret;
	ret = cfzy_confnode_intconfig_new(*nodep, tklist);
	if (ret != NO_MATCH)
		return ret;
	ret = cfzy_confnode_menu_new(*nodep, tklist);
	if (ret != NO_MATCH)
		return ret;
	ret = cfzy_confnode_menuconfig_new(*nodep, tklist);
	if (ret != NO_MATCH)
		return ret;
	ret = cfzy_confnode_strconfig_new(*nodep, tklist);
	if (ret != NO_MATCH)
		return ret;

	cfzy_confnode_free(*nodep);
	return NO_MATCH;
}

/* Try to create a new node with all its attributes and help. Return
 * NO_MATCH if the current line is not a new node command, SUCCESS if
 * a new node is created, or ERROR on error. */
static enum cfzy_parse_return
parse_confnode(struct conftree_parse_status *state)
{
	char *buf;
	struct cfzy_confnode *n;
	enum cfzy_parse_return ret;
	struct cfzy_token_list *tklist;

	if (get_curline(state) < 0)
		return ERROR;

	buf = state->curline;

	/* EOF */
	if (buf == NULL)
		return NO_MATCH;

	/* line must not start with a space */
	if (isspace(*buf))
		return NO_MATCH;

	/* if it's a new node command */
	ret = new_node(state, &n);
	if (ret == ERROR || ret == NO_MATCH)
		return ret;

	/* save parsing infos */
	n->filename = strdup(state->filename);
	if (n->filename == NULL) {
		cfzy_confnode_free(n);
		return ERROR;
	}
	n->linenum = state->linenum;

	TAILQ_INSERT_TAIL(&state->parent->children, n, child_next);
	eat_curline(state);

	/* parse node attributes */
	while (1) {

		/* fill state structure with current or next line */
		if (get_curline(state) < 0)
			return ERROR;

		buf = state->curline;
		tklist = state->tklist;

		/* EOF */
		if (buf == NULL)
			return SUCCESS;

		/* skip empty lines */
		if (line_is_empty(buf)) {
			eat_curline(state);
			continue;
		}

		/* if line does not start with space, this node is
		 * parsed, break */
		if (!isspace(*buf))
			break;

		/* check if it's a node attribute */
		ret = cfzy_confnode_add_attr(n, tklist);
		if (ret == ERROR)
			return ERROR;
		if (ret == SUCCESS) {
			eat_curline(state);
			continue;
		}

		/* check if it's a help indication */
		ret = parse_help(n, state);
		if (ret == ERROR)
			return ERROR;
		if (ret == SUCCESS)
			break;

		/* the attribute is invalid */
		LOG(ERR, "invalid node attribute\n");
		return ERROR;
	}

	/* enter in a new node */
	if (n->flags & CFZY_F_IS_DIR)
		state->parent = n;

	return SUCCESS;
}

/* Try to match the command that closes the current parent node, if
 * it's a directory. Return NO_MATCH if the line does not match this
 * command, SUCCESS if the parent node is closed (in this case,
 * state->parent is updated), or ERROR on error. */
static enum cfzy_parse_return
close_dir(struct conftree_parse_status *state)
{
	struct cfzy_token_list *tklist;
	struct cfzy_token *tok;
	struct cfzy_confnode *n;

	n = state->parent;
	if ((n->flags & CFZY_F_IS_DIR) == 0)
		return NO_MATCH;

	/* fill state structure with current or next line */
	if (get_curline(state) < 0)
		return ERROR;

	/* EOF */
	if (state->curline == NULL)
		return NO_MATCH;

	tklist = state->tklist;
	if (tklist->n_token != 1)
		return NO_MATCH;

	/* token must be "endchoice", "endmenu", "endif", ... */
	tok = TAILQ_FIRST(&tklist->list);
	if (strncmp(tok->str, "end", 3))
		return NO_MATCH;
	if (strcmp(tok->str + 3, cfzy_confnode_get_type_str(n)))
		return NO_MATCH;
	if (cfzy_confnode_close(n) < 0) {
		LOG(ERR, "cannot close node\n");
		return ERROR;
	}

	state->parent = n->parent;
	return SUCCESS;
}

/*
 * Parse the configuration tree by reading the file stored in the
 * 'state' structure. Return SUCCESS or ERROR.
 */
static enum cfzy_parse_return parse_conftree(struct conftree_parse_status *state)
{
	enum cfzy_parse_return ret;
	char *buf;

	/* browse all lines of the file */
	while (1) {

		/* fill state structure with current or next line */
		if (get_curline(state) < 0)
			return ERROR;

		buf = state->curline;

		/* EOF */
		if (buf == NULL)
			break;

		LOG(DEBUG, "parent=%s: %s",
		    cfzy_confnode_name(state->parent), buf);

		/* skip empty / comments lines */
		if (line_is_empty(buf)) {
			eat_curline(state);
			continue;
		}

		/* if the line does not start with a space, syntax error */
		if (isspace(*buf)) {
			LOG(ERR, "line starts with space\n");
			return ERROR;
		}

		/* if it's a source command, parse the new file */
		ret = source_file(state);
		if (ret == ERROR)
			return ERROR;
		if (ret == SUCCESS)
			continue;

		/* if the command closes a menu, choice, ... */
		ret = close_dir(state);
		if (ret == ERROR)
			return ERROR;
		if (ret == SUCCESS) {
			eat_curline(state);
			continue;
		}

		/* parse new node */
		ret = parse_confnode(state);
		if (ret == ERROR)
			return ERROR;
		if (ret == SUCCESS)
			continue;

		/* invalid command */
		LOG(ERR, "invalid command\n");
		return ERROR;
	}

	return SUCCESS;
}

/*
 * Parse the configuration tree by reading "filename", knowing that
 * the current parent node is pointed by pparent (which can be
 * modified at the end of the function). This function will fill open
 * the file, and fill a state structure for the parser, then call
 * parse_conftree(). Return 0 on success and -1 on error.
 */
static int
__cfzy_conftree_parse(const char *filename, struct cfzy_confnode **pparent)
{
	struct conftree_parse_status state;
	FILE *f;
	char old_cwd_buf[PATH_MAX];
	char *cwd_buf = NULL, *cwd_ptr = NULL, *old_cwd = NULL;
	int ret;

	memset(&state, 0, sizeof(state));

	old_cwd = getcwd(old_cwd_buf, sizeof(old_cwd_buf));
	if (old_cwd == NULL) {
		LOG(ERR, "getcwd failed: %s\n", strerror(errno));
		goto fail;
	}

	f = fopen(filename, "r");
	if (f == NULL) {
		LOG(ERR, "cannot find file <%s>\n", filename);
		goto fail;
	}

	cwd_buf = strdup(filename);
	if (cwd_buf == NULL) {
		LOG(ERR, "not enough memory\n");
		goto fail;
	}
	cwd_ptr = dirname(cwd_buf);
	ret = chdir(cwd_ptr);
	if (ret < 0) {
		LOG(ERR, "chdir failed: %s\n", strerror(errno));
		goto fail;
	}

	state.filename = filename;
	state.f = f;
	state.linenum = 0;
	state.parent = *pparent;
	state.curline = NULL;

	/* browse all lines of the file */
	ret = parse_conftree(&state);
	if (ret != SUCCESS) {
		/* node will be freed by caller (cfzy_conftree_new) */
		LOG(ERR, "parse error at %s:%d\n", filename, state.linenum);
		LOG(ERR, ">>> %s", state.curline);
		goto fail;
	}

	eat_curline(&state); /* will free token_list */
	fclose(f);
	free(cwd_buf);
	if (chdir(old_cwd) < 0) {
		LOG(ERR, "chdir back to <%s> failed: %s\n",
		    old_cwd, strerror(errno));
		return -1;
	}

	*pparent = state.parent;
	return 0;

 fail:
	if (state.f != NULL)
		eat_curline(&state); /* will free token_list */
	if (f != NULL)
		fclose(f);
	if (cwd_buf != NULL)
		free(cwd_buf);
	if (old_cwd != NULL)
		if (chdir(old_cwd) < 0)
			LOG(ERR, "chdir back to <%s> failed: %s\n",
			    old_cwd, strerror(errno));

	return -1;
}

/* External api for parsing a new configuration tree file. Return 0 on
 * success and -1 on error. */
int cfzy_conftree_parse(const char *filename, struct cfzy_conftree *conftree)
{
	int ret;
	struct cfzy_confnode *parent = conftree->root;

	ret = __cfzy_conftree_parse(filename, &parent);
	if (ret < 0)
		return ret;

	if (parent != conftree->root) {
		LOG(ERR, "Node <%s> not closed properly\n",
		    cfzy_confnode_name(parent));
		return -1;
	}

	return 0;
}
