/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file
 * Confizery configuration tree parser
 *
 * This modules provides an API to parse the configuration tree file
 * (equivalent to Kconfig file in linux terminology) which describes
 * the structure of the tree, the option list, the dependencies, the
 * default values...
 */


#ifndef _CFZY_CONFTREE_PARSER_H_
#define _CFZY_CONFTREE_PARSER_H_

#include <sys/queue.h>

struct cfzy_conftree;

/**
 * Return value for some parsing functions.
 */
enum cfzy_parse_return {
	SUCCESS,
	NO_MATCH,
	ERROR,
};

/**
 * This is the structure defining a token. It is used by the
 * configuration tree parser. A token is either a word or a string
 * surrounded by double quotes.
 */
struct cfzy_token {
	TAILQ_ENTRY(cfzy_token) next; /**< pointer to next token */
	char *str;               /**< string */
	unsigned offset;         /**< offset in line */
};

/**
 * A list of token, and a pointer to the initial line buffer
 */
struct cfzy_token_list {
	TAILQ_HEAD(, cfzy_token) list; /**< list of tokens */
	unsigned n_token;        /**< number of tokens */
	const char *linebuf;     /**< the line buffer */
};

/**
 * Parse configuration tree file and fill the conftree
 *
 * This function is called by cfzy_conftree_new() to fill the newly
 * allocated tree.
 *
 * @param filename
 *   path to configuration tree file
 * @param conftree
 *   an empty configuration tree
 * @return
 *   0 on success, or negative on error
 */
int cfzy_conftree_parse(const char *filename, struct cfzy_conftree *conftree);

#endif /* _CFZY_CONFTREE_PARSER_H_ */
