/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>

#include "cfzy_log.h"
#include "cfzy_htable.h"
#include "cfzy_string.h"
#include "cfzy_file.h"
#include "cfzy_conftree.h"
#include "cfzy_confnode.h"

#define LOG(level, fmt, args...)				\
	CFZY_LOG("dotconfig", level, fmt, ##args)

/* Allocate and return string containing the value of the
 * option. Buffer must point to the option (after the '=' in the
 * .config), and it can be quoted. The string must be freed by the
 * user. */
static char *dotconfig_get_optval(const char *buf)
{
	const char *s;
	unsigned i, len;
	char *retbuf;

	/* skip spaces */
	s = buf;
	while (*s != '\0' && isspace(*s))
		s++;
	if (*s == '\0' || *s == '#')
		return 0;

	/* if it's a quote, unquote the string */
	if (*s == '"' || *s == '\'') {
		retbuf = cfzy_string_unquote(s, &len);
	}
	else {
		retbuf = strdup(s);
		if (retbuf == NULL)
			return NULL;

		for (i = 0; s[i] != '\0' && s[i] != '\n' && s[i] != '#'; i ++)
			;

		retbuf[i] = '\0';
	}
	return retbuf;
}

/* parse a line of .config: fill the uservalue of nodes in the
 * conftree structure */
static int dotconfig_parse_line(struct cfzy_conftree *conftree, const char *line)
{
	struct cfzy_confnode *n;
	char buf[BUFSIZ];
	char *nodename, *s;
	char *optval;

	/* copy line in a writable buffer */
	buf[sizeof(buf) - 1] = '\0';
	strncpy(buf, line, sizeof(buf) - 1);
	s = buf;

	/* skip spaces */
	while (*s != '\0' && isspace(*s))
		s++;

	/* nothing to do, empty line */
	if (*s == '\0')
		return 0;

	/* skip the first "#", expecting "is not set" */
	if (*s == '#') {
		s++;
		while (*s != '\0' && isspace(*s))
			s++;

		/* nothing to do, it's a comment */
		if (*s == '\0')
			return 0;

		/* get node name, ignore invalid names */
		if (strncmp("CONFIG_", s, strlen("CONFIG_")))
			return 0;
		nodename = s + strlen("CONFIG_");
		s = nodename;

		/* skip spaces */
		while (*s != '\0' && !isspace(*s))
			s++;

		/* nothing to do, it's a comment */
		if (*s == '\0')
			return 0;

		*s = '\0';
		s ++;

		/* should end with "is not set", else it's a comment */
		if (strcmp(s, "is not set\n"))
			return 0;

		optval = strdup("n");
		if (optval == NULL)
			return -1;
	}
	else {
		/* get node name, ignore invalid names */
		if (strncmp("CONFIG_", s, strlen("CONFIG_")))
			return -1;
		nodename = s + strlen("CONFIG_");

		while (*s != '\0' && *s != '=')
			s++;
		/* invalid line */
		if (*s != '=')
			return -1;

		*s = '\0';
		optval = dotconfig_get_optval(s + 1);
		if (optval == NULL)
			return -1;
	}

	/* find the node in the conf htable */
	n = cfzy_htable_lookup(conftree->nodes_htable, nodename);
	if (n == NULL) {
		LOG(NOTICE, "ignore unknown symbol <%s>\n", nodename);
		free(optval);
		return 0;
	}

	/* check that value is valid */
	if (cfzy_confnode_str2bool(n, optval) < 0) {
		LOG(ERR, "invalid option value %s=<%s>\n", nodename, optval);
		free(optval);
		return -1;
	}

	/* set value in node */
	LOG(DEBUG, "%s=<%s>\n", nodename, optval);
	if (cfzy_confnode_set_uservalue(n, optval) < 0) {
		LOG(ERR, "cannot set value %s=<%s>\n", nodename, optval);
		free(optval);
		return -1;
	}

	free(optval);
	return 0;
}

/* parse .config */
int cfzy_dotconfig_read(struct cfzy_conftree *conftree, const char *filename)
{
	FILE *f;
	char buf[BUFSIZ];
	int err = 0;

	LOG(INFO, "dotconfig read <%s>\n", filename);

	f = fopen(filename, "r");
	if (f == NULL) {
		LOG(ERR, "cannot find file <%s>\n", filename);
		return -1;
	}
	while (fgets(buf, sizeof(buf), f) != NULL) {

		if (dotconfig_parse_line(conftree, buf) < 0) {
			err = -1;
			break;
		}
	}
	fclose(f);

	return err;
}

/* write .config */
int cfzy_dotconfig_write(const struct cfzy_conftree *conftree,
			 const char *filename)
{
	FILE *f, *tmp_f;
	const struct cfzy_confnode *c;
	int ret;

	LOG(INFO, "open old dotconfig <%s>\n", filename);

	/* open old file in read mode */

	f = fopen(filename, "r");
	if (f == NULL && errno != ENOENT) {
		printf("cannot open file <%s>: %s\n",
		       filename, strerror(errno));
		return -1;
	}

	/* old file exists */
	if (f != NULL) {

		/* write C header in temp file */
		tmp_f = tmpfile();
		if (tmp_f == NULL) {
			printf("cannot open temp file\n");
			return -1;
		}

		/* dump all the children of root node */
		TAILQ_FOREACH(c, &conftree->root->children, child_next) {
			ret = cfzy_confnode_dotconfig_write(c, tmp_f);
			if (ret < 0)
				break;
		}

		/* files are the same, nothing to do */
		if (cfzy_file_compare(f, tmp_f) == 0) {
			LOG(INFO, "already up to date: <%s>\n", filename);
			fclose(tmp_f);
			fclose(f);
			return 0;
		}

		fclose(tmp_f);
		fclose(f);
	}

	/* reopen the file in write mode */

	LOG(INFO, "write new C header <%s>\n", filename);

	f = fopen(filename, "w");
	if (f == NULL) {
		LOG(ERR, "cannot open file <%s>\n", filename);
		return -1;
	}

	/* dump all the children of root node */
	TAILQ_FOREACH(c, &conftree->root->children, child_next) {
		ret = cfzy_confnode_dotconfig_write(c, f);
		if (ret < 0)
			break;
	}

	fclose(f);

	return ret;
}
