/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/queue.h>

#include "cfzy_list.h"
#include "cfzy_expr.h"

/* #define DEBUG */

#ifdef DEBUG
#define debug_printf(args...) fprintf(stderr, args)
#else
#define debug_printf(args...) do { } while(0)
#endif

/*
 * Return the operator (or the variable name) of the root node of the
 * given expression.
 */
static const char *op_print(const struct cfzy_expr *exp)
{
	switch (exp->optype) {
	case OP_OR:
		return "||";
	case OP_AND:
		return "&&";
	case OP_EQUAL:
		return "==";
	case OP_OBRACKET:
		return "(";
	case OP_NOT:
		return "!";
	case OP_CBRACKET:
		return ")";
	case OP_VAR:
		return exp->varname;
	default:
		return NULL;
	}
}

/* recursive function called by cfzy_expr_graph_dump() */
static void expr_dump_node(FILE *f, const struct cfzy_expr *exp, int level)
{
	if (exp == NULL)
		return;

	fprintf(f, "vertice %d %s:%p\n", level, op_print(exp), exp);

	if (exp->left) {
		fprintf(f, "edge %p %p\n", exp, exp->left);
		expr_dump_node(f, exp->left, level + 1);
	}

	if (exp->right) {
		fprintf(f, "edge %p %p\n", exp, exp->right);
		expr_dump_node(f, exp->right, level + 1);
	}
}

/* dump an expression graph in a file */
int cfzy_expr_graph_dump(const char *filename, const struct cfzy_expr *exp)
{
	FILE *f;

	f = fopen(filename, "w");
	if (f == NULL)
		return -1;
	expr_dump_node(f, exp, 0);
	fclose(f);
	return 0;
}

/* return true if it's a binary operator */
static int expr_is_bin_op(const struct cfzy_expr *exp)
{
	if (exp->optype == OP_AND ||
	    exp->optype == OP_OR ||
	    exp->optype == OP_EQUAL)
		return 1;
	return 0;
}

/*
 * Dump the expression 'exp' as a string into the buffer 'buf' of
 * length 'len'. The string is nul-terminated. Return the number of
 * written bytes on success (not including \0), else return -1.
 */
int cfzy_expr_to_str(const struct cfzy_expr *exp, char *buf, int len)
{
	int n, orig_len;

	orig_len = len;

	/* dump left expression if it's a binary operator (in case of
	 * unary operator, we only use the right child) */
	if (expr_is_bin_op(exp)) {
		n = cfzy_expr_to_str(exp->left, buf, len);
		if (n == -1 || n >= len || len <= 0)
			return -1;
		buf += n;
		len -= n;
	}

	/* dump the operator */
	if (expr_is_bin_op(exp))
		n = snprintf(buf, len, " %s ", op_print(exp));
	else
		n = snprintf(buf, len, "%s", op_print(exp));
	if (n == -1 || n >= len || len <= 0)
		return -1;
	buf += n;
	len -= n;

	/* dump the right expression except if it's a variable */
	if (exp->optype != OP_VAR) {
		n = cfzy_expr_to_str(exp->right, buf, len);
		if (n == -1 || n >= len || len <= 0)
			return -1;
		buf += n;
		len -= n;
	}

	/* close the bracket if needed */
	if (exp->optype == OP_OBRACKET) {
		n = snprintf(buf, len, ")");
		if (n == -1 || n >= len || len <= 0)
			return -1;
		buf += n;
		len -= n;
	}

	return orig_len - len;
}

/*
 * Evaluate an expression. Return -1 on error, else the value of the
 * expression (greater or equal to 0).
 */
int cfzy_expr_eval(const struct cfzy_expr *exp,
		   int (*getvalue)(const char *, void *), void *opaque_arg)
{
	switch (exp->optype) {
	case OP_OR:
		return cfzy_expr_eval(exp->left, getvalue, opaque_arg) ||
			cfzy_expr_eval(exp->right, getvalue, opaque_arg);
	case OP_AND:
		return cfzy_expr_eval(exp->left, getvalue, opaque_arg) &&
			cfzy_expr_eval(exp->right, getvalue, opaque_arg);
	case OP_EQUAL:
		return cfzy_expr_eval(exp->left, getvalue, opaque_arg) ==
			cfzy_expr_eval(exp->right, getvalue, opaque_arg);
	case OP_OBRACKET:
		return cfzy_expr_eval(exp->right, getvalue, opaque_arg);
	case OP_NOT:
		return !cfzy_expr_eval(exp->right, getvalue, opaque_arg);
	case OP_VAR:
		/* constants names (used for testing purposes) */
		if (!strcmp("true", exp->varname))
			return 1;
		if (!strcmp("false", exp->varname))
			return 0;
		if (getvalue == NULL)
			return -1;
		return getvalue(exp->varname, opaque_arg);
	default:
		debug_printf("%s(): bad operator\n", __FUNCTION__);
		return -1;
	}
}

/* The argument 'buf' is expression string that must start with an
 * opening bracket '('. This function returns the len of the
 * expression contained until we get the closing bracket corresponding
 * to buf[0]. Return -1 on error. */
static int get_brac_len(const char *buf)
{
	const char *s = buf;
	int i = 1;

	if (*s != '(')
		return -1;
	s++;

	while(*s != '\0' && i != 0) {
		if (*s == ')')
			i--;
		if (*s == '(')
			i++;
		s++;
	}
	if (i != 0)
		return -1;

	return s-buf;
}

/*
 * Allocate a new expression exp corresponding to the next characters
 * in "buf". On success, a new expression exp is returned and the
 * number of eaten characters is set in "eatlen". If the end of the
 * string is reached, the expression exp contains the OP_EOF
 * operator. On error, NULL is returned.
 */
 static struct cfzy_expr *get_next_op(const char *buf, unsigned *eatlen)
{
	struct cfzy_expr *exp;
	const char *s;
	unsigned len;

	s = buf;
	exp = malloc(sizeof(struct cfzy_expr));
	if (exp == NULL)
		return NULL;
	memset(exp, 0, sizeof(*exp));

	switch (s[0]) {
	case '#':
		exp->optype = OP_EOF;
		*eatlen = s - buf + 1;
		return exp;
	case '\0':
		exp->optype = OP_EOF;
		*eatlen = s - buf + 1;
		return exp;
	case '(':
		exp->optype = OP_OBRACKET;
		*eatlen = s - buf + 1;
		return exp;
	case ')':
		exp->optype = OP_CBRACKET;
		*eatlen = s - buf + 1;
		return exp;
	case '!':
		exp->optype = OP_NOT;
		*eatlen = s - buf + 1;
		return exp;
	case '&':
		exp->optype = OP_AND;
		if (s[1] != '&')
			goto fail;
		*eatlen = s - buf + 2;
		return exp;
	case '|':
		exp->optype = OP_OR;
		if (s[1] != '|')
			goto fail;
		*eatlen = s - buf + 2;
		return exp;
	case '=':
		if (s[1] != '=')
			goto fail;
		exp->optype = OP_EQUAL;
		*eatlen = s - buf + 2;
		return exp;
	default:
		/* a variable name must start with a letter */
		if (isalpha(s[0]))
			break;
		goto fail;
	}

	/* It's a variable, get name */
	while (isalnum(s[0]) || s[0] == '_')
		s++;

	exp->optype = OP_VAR;
	*eatlen = s - buf;

	/* alloc string for variable name */
	len = s - buf;
	exp->varname = malloc(len+1);
	memcpy(exp->varname, buf, len);
	exp->varname[len] = '\0';
	return exp;

 fail:
	free(exp);
	return NULL;
}

/*
 * Parse the expression contained in s (which is a modifiable
 * string). Return 0 on success, in this case the expression tree is
 * returned in exp_ptr. On error, return -1.
 */
static int expr_parse(char *s, struct cfzy_expr **exp_ptr)
{
	struct cfzy_expr *top = NULL, *exp, *tmp;
	int len;
	unsigned eatlen;

	debug_printf("parse <%s>\n", s);

	if (*s == '\0')
		return -1;

	while (1) {

		/* skip spaces */
		while (isspace(*s))
			s++;

		/* get next operator/operand */
		exp = get_next_op(s, &eatlen);
		if (exp == NULL) {
			debug_printf("Parse error\n");
			goto fail;
		}

		if (exp->optype == OP_EOF) {
			free(exp);
			exp = NULL;
			break;
		}

		debug_printf("New node: %s\n", op_print(exp));
		switch (exp->optype) {

		case OP_OR:
		case OP_AND:
		case OP_EQUAL:
			/* if it is the first node, error */
			if (top == NULL)
				goto fail;

			/* the rightest leaf must be a variable */
			tmp = top;
			while (tmp->right)
				tmp = tmp->right;
			if (tmp->optype != OP_VAR)
				goto fail;

			/* if new node priority is lower than root
			 * node, it becomes the new root */
			if (exp->optype < top->optype) {
				exp->left = top;
				top = exp;
				break;
			}

			/* exp is placed at the "rightest" leaf. Check
			 * that our parent is not a variable */
			tmp = top;
			while (tmp->right && exp->optype > tmp->right->optype)
				tmp = tmp->right;

			if (tmp->right == NULL)
				goto fail;

			exp->left = tmp->right;
			tmp->right = exp;
			break;

		case OP_OBRACKET:
			/* if operator is an opening bracket,
			 * recursively call ourself with the
			 * sub-expression */
			len = get_brac_len(s);
			if (len < 0) {
				debug_printf("Cannot find closing bracket\n");
				goto fail;
			}
			s[len-1] = '\0';
			if (expr_parse(s+1, &exp->right) < 0)
				goto fail;
			if (exp->right == NULL)
				goto fail;
			s[len-1] = ')';
			s += len - 1;

			/* if it is the first node */
			if (top == NULL) {
				top = exp;
				break;
			}

			/* exp is placed at the "rightest" leaf. Check
			 * that our parent is not a variable */
			tmp = top;
			while (tmp->right)
				tmp = tmp->right;
			if (tmp->optype == OP_VAR)
				goto fail;
			tmp->right = exp;
			break;

		case OP_NOT:
		case OP_VAR:
			/* if it is the first node */
			if (top == NULL) {
				top = exp;
				break;
			}

			/* exp is placed at the "rightest" leaf. Check
			 * that our parent is not a variable */
			tmp = top;
			while (tmp->right)
				tmp = tmp->right;
			if (tmp->optype == OP_VAR)
				goto fail;
			tmp->right = exp;
			break;

		default:
			/* should not happen */
			debug_printf("Parse error\n");
			goto fail;
		}

		s += eatlen;
	}

	if (top == NULL)
		goto fail;

	/* the rightest leaf must be a variable */
	tmp = top;
	while (tmp->right)
		tmp = tmp->right;
	if (tmp->optype != OP_VAR) {
		debug_printf("Missing operand\n");
		goto fail;
	}

	*exp_ptr = top;
	return 0;

 fail:
	if (top != NULL)
		cfzy_expr_free(top);
	if (exp != NULL)
		cfzy_expr_free(exp);

	return -1;
}

/*
 * Parse the expression contained in buf. Return an expression tree or
 * NULL on error.
 */
struct cfzy_expr *cfzy_expr_parse(const char *buf)
{
	struct cfzy_expr *top;
	char *s;
	int ret;

	s = strdup(buf);
	if (s == NULL)
		return NULL;

	ret = expr_parse(s, &top);
	free(s);
	if (ret < 0)
		return NULL;
	return top;
}

/* Free an expression tree */
void cfzy_expr_free(struct cfzy_expr *exp)
{
	if (exp == NULL)
		return;

	if (exp->left) {
		cfzy_expr_free(exp->left);
		exp->left = NULL;
	}

	if (exp->right) {
		cfzy_expr_free(exp->right);
		exp->right = NULL;
	}

	if (exp->varname != NULL)
		free(exp->varname);

	free(exp);
}

/* recursive function called by cfzy_expr_get_vars() that browse the
 * expr tree to fill the list */
static int
expr_get_vars(const struct cfzy_expr *exp, struct cfzy_list_head *list)
{
	struct cfzy_list_elt *e;
	char *name;

	if (exp->left) {
		if (expr_get_vars(exp->left, list) < 0)
			return -1;
	}

	if (exp->right) {
		if (expr_get_vars(exp->right, list) < 0)
			return -1;
	}

	if (exp->optype != OP_VAR)
		return 0;

	/* ignore true and false constants */
	if (!strcmp(exp->varname, "true"))
		return 0;
	if (!strcmp(exp->varname, "false"))
		return 0;

	/* nothing to do if var is already in list */
	TAILQ_FOREACH(e, list, next) {
		if (!strcmp(e->ptr, exp->varname))
			return 0;
	}

	name = strdup(exp->varname);
	if (name == NULL)
		return -1;

	if (cfzy_list_add_tail(list, name) < 0) {
		free(name);
		return -1;
	}

	return 0;
}

/* Return a list containing all variables on which expression depend */
struct cfzy_list_head *cfzy_expr_get_vars(const struct cfzy_expr *exp)
{
	struct cfzy_list_head *list;

	list = cfzy_list_alloc();
	if (list == NULL)
		return NULL;

	if (expr_get_vars(exp, list) < 0) {
		cfzy_list_free(list, free);
		return NULL;
	}

	return list;
}
