/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _CFZY_EXPR_H_
#define _CFZY_EXPR_H_

/**
 * @file
 * Confizery Expression
 *
 * This module provides functions to parse an expression string and
 * convert it in a logical expression tree, that can be evaluated,
 * according to the variables defined in the configuration tree.
 *
 * It is used in configuration tree file, for instance in the "if"
 * node or the "default" attribute.
 */

#include <sys/queue.h>

#include "cfzy_list.h"

/**
 * Type of operand in expression tree (from low to high priority)
 */
enum cfzy_expr_op {
	OP_OR,
	OP_AND,
	OP_EQUAL,
	OP_OBRACKET,
	OP_CBRACKET,
	OP_NOT,
	OP_VAR,
	OP_EOF,
};

/**
 * Structure describing an expression tree
 *
 * An expression tree is a binary tree. Each node contains an
 * operator, or a variable name. If it's a unary operator, only the
 * right child is used, else both left and right children are set.
 */
struct cfzy_expr {
	TAILQ_ENTRY(expr) next;   /**< next entry in list */
	enum cfzy_expr_op optype; /**< operator type */
	char *varname;            /**< var name, only valid if type is OP_VAR */
	struct cfzy_expr *left;   /**< left child */
	struct cfzy_expr *right;  /**< right child */
};

/**
 * Parse a string and convert it in an expression tree
 *
 * @param buf
 *   String containing an expression
 * @return
 *   Expression tree, or NULL on error
 */
struct cfzy_expr *cfzy_expr_parse(const char *buf);

/**
 * Free an expression tree previously allocated with cfzy_expr_parse()
 *
 * @param exp
 *   Expression to free
 */
void cfzy_expr_free(struct cfzy_expr *exp);

/**
 * Write an expression tree in a string buffer
 *
 * Dump the expression 'exp' as a string into the buffer 'buf' of
 * length 'len'. The output string is always nul-terminated.
 *
 * @param exp
 *   Expression tree to dump
 * @param buf
 *   Destination buffer
 * @param len
 *   Len of the destination buffer
 * @return
 *   Number of written bytes on success (not including '\0'), else
 *   return -1.
 */
int cfzy_expr_to_str(const struct cfzy_expr *exp, char *buf, int len);

/**
 * Evaluate an expression tree
 *
 * Evaluate the given expression tree according to variables defined
 * in conftree.
 *
 * @param exp
 *   Expression tree to evaluate
 * @param getvalue
 *   Function used to evaluate a variable. This function should
 *   return the value of the variable (positive), or a negative value
 *   on error (variable invalid or not found).
 *   This argument can be NULL, in this case only "true" and "false" are
 *   evaluated.
 * @param opaque_arg
 *   Second argument given as-is to getvalue() function.
 * @return
 *   The value of the expression (greater or equal to 0), or -1 on
 *   error
 */
int cfzy_expr_eval(const struct cfzy_expr *exp,
		   int (*getvalue)(const char *, void *), void *opaque_arg);

/**
 * Dump the expression graph in a file
 *
 * Useful for debug purpose only. The output can be parsed with
 * cfzy_expr_graph.py.
 *
 * @param filename
 *   Name of file
 * @param exp
 *   Expression tree to dump
 * @return
 *   0 on succes, -1 on error
 */
int cfzy_expr_graph_dump(const char *filename, const struct cfzy_expr *exp);

/**
 * Return a list containing all variables on which expression depend
 *
 * Allocate a list head, and for each variable of the expression,
 * allocate an element and add it in the list. Each element points to
 * a string which is allocated and should be freed by the user. Each
 * variable is present only once in the list.
 *
 * @param exp
 *   Expression tree
 * @return
 *   List containing the variables (it can be empty), or NULL on error.
 */
struct cfzy_list_head *cfzy_expr_get_vars(const struct cfzy_expr *exp);

#endif /* _CFZY_EXPR_H_ */
