/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/queue.h>

#include "cfzy_htable.h"

/* return the hash from name */
static uint32_t hash_name(const char *name)
{
	uint32_t h = 0;
	const unsigned char *c;

	/* XXX add murmurhash ? */
	for (c = (const unsigned char *)name; *c; c++)
		h += ((((uint32_t)*c) << 4) + (((uint32_t)*c) >> 4)) * 11;

	return h & CFZY_HTABLE_MASK;
}

/* return the hlist_elt associated to the given name. If not found,
 * return NULL. */
static struct cfzy_hlist_elt *
__cfzy_htable_lookup(const struct cfzy_htable *htable, const char *name)
{
	struct cfzy_hlist_elt *e;
	uint32_t h = hash_name(name);

	LIST_FOREACH(e, &htable->buckets[h], next) {
		if (!strcmp(name, e->name))
			break;
	}
	return e;
}

/* return the object associated to the given name. If not found,
 * return NULL. */
void *cfzy_htable_lookup(const struct cfzy_htable *htable, const char *name)
{
	struct cfzy_hlist_elt *e = __cfzy_htable_lookup(htable, name);
	if (e == NULL)
		return NULL;
	return e->ptr;
}

/* add an entry in the htable, allocating a new hlist_elt */
int cfzy_htable_add(struct cfzy_htable *htable, const char *name,
		    void *ptr)
{
	struct cfzy_hlist_elt *e;
	uint32_t h;

	/* The ptr should not be NULL, else we won't differentiate a
	 * lookup matching a NULL element and a non-matching one */
	if (ptr == NULL)
		return -1;

	e = malloc(sizeof(struct cfzy_hlist_elt));
	if (e == NULL)
		return -1;

	e->name = strdup(name);
	if (e->name == NULL) {
		free(e);
		return -1;
	}

	e->ptr = ptr;
	h = hash_name(name);
	LIST_INSERT_HEAD(&htable->buckets[h], e, next);

	return 0;
}

/* delete and free an hlist_elt from the htable */
static int __cfzy_htable_del(struct cfzy_hlist_elt *e)
{
	LIST_REMOVE(e, next);
	free(e->name);
	free(e);
	return 0;
}

/* delete an entry from the htable, freeing the hlist_elt */
int cfzy_htable_del(struct cfzy_htable *htable, const char *name)
{
	struct cfzy_hlist_elt *e = __cfzy_htable_lookup(htable, name);

	if (e == NULL)
		return -1;

	return __cfzy_htable_del(e);
}

/* allocate and initialize a new htable */
struct cfzy_htable *cfzy_htable_alloc(void)
{
	struct cfzy_htable *htable;
	unsigned i;

	htable = malloc(sizeof(struct cfzy_htable));
	if (htable == NULL)
		return NULL;

	for (i = 0; i < CFZY_HTABLE_SIZE; i++) {
		LIST_INIT(&htable->buckets[i]);
	}
	return htable;
}

/* empty the htable, freeing hlist_elt structures, and free the htable */
void cfzy_htable_free(struct cfzy_htable *htable, void (*free_fct)(void *))
{
	struct cfzy_hlist_elt *e;
	unsigned i;

	for (i = 0; i < CFZY_HTABLE_SIZE; i++) {
		while ((e = LIST_FIRST(&htable->buckets[i])) != NULL) {
			if (free_fct != NULL)
				free_fct(e->ptr);

			__cfzy_htable_del(e);
		}
	}
	free(htable);
}
