/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _CFZY_HTABLE_H_
#define _CFZY_HTABLE_H_

/**
 * @file
 * Confizery Hash Table
 *
 * This module is a simple hash table implementation, based on LIST
 * from sys/queue.h --see queue(3)-- and with a constant number of
 * buckets. It handles the allocation of chaining structure so any
 * kind of element can be added/deleted. Each element is referenced by
 * a string (key).
 */

#include <sys/queue.h>

/* we use a constant size of htable, it simplifies the code and it is
 * enough for our purpose */

#define CFZY_HTABLE_ORDER 10
#define CFZY_HTABLE_SIZE  (1 << CFZY_HTABLE_ORDER)
#define CFZY_HTABLE_MASK  (CFZY_HTABLE_SIZE - 1)

/**
 * A hash table bucket
 */
LIST_HEAD(cfzy_hlist_head, cfzy_hlist_elt);

/**
 * A hash table element, used to chain objects together and store the
 * name (key). This structure is allocated by cfzy_htable_add().
 */
struct cfzy_hlist_elt {
	LIST_ENTRY(cfzy_hlist_elt) next;
	char *name;
	void *ptr;
};

/**
 * A hash table structure
 */
struct cfzy_htable {
	struct cfzy_hlist_head buckets[CFZY_HTABLE_SIZE];
};

/**
 * Lookup for an element matching the given key
 *
 * @param htable
 *   Hash table pointer
 * @param name
 *   String identifying the element (key)
 * @return
 *   Pointer to the element, or NULL on error
 */
void *cfzy_htable_lookup(const struct cfzy_htable *htable, const char *name);


/**
 * Add an element in the hash table
 *
 * Allocate a new cfzy_hlist_elt structure, initialize it to point to
 * the given pointer "ptr", then queue the structure in the
 * appropriate hash table bucket.
 *
 * @param htable
 *   Hash table pointer
 * @param name
 *   String identifying the element (key)
 * @param ptr
 *   Pointer to the element (must not be NULL)
 * @return
 *   0 on success, -1 on error
 */
int cfzy_htable_add(struct cfzy_htable *htable, const char *name,
		    void *ptr);

/**
 * Remove an element from the hash table
 *
 * Unchain the structure from the hash table bucket, and free the
 * cfzy_hlist_elt structure.
 *
 * @param htable
 *   Hash table pointer
 * @param name
 *   String identifying the element (key)
 * @return
 *   0 on success, -1 on error (element not found)
 */
int cfzy_htable_del(struct cfzy_htable *htable, const char *name);

/**
 * Allocate and initialize a new empty htable
 *
 * @return
 *   Hash table pointer or NULL on error
 */
struct cfzy_htable *cfzy_htable_alloc(void);

/**
 * Free a htable
 *
 * Empty the htable, freeing all elements (cfzy_hlist_elt structures,
 * and its pointer, using the provided free_fct), then free the
 * cfzy_htable.
 *
 * @param htable
 *   Hash table pointer
 * @param free_fct
 *   Function to free an element
 */
void cfzy_htable_free(struct cfzy_htable *htable, void (*free_fct)(void *));

#endif /* _CFZY_HTABLE_H_ */
