/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _CFZY_LIST_H_
#define _CFZY_LIST_H_

/**
 * @file
 * Confizery List
 *
 * This module is a list implementation, based on TAILQ from
 * sys/queue.h -see queue(3)-. It handles the allocation of chaining
 * structure so any kind of element can be chained.
 *
 * To be really useful, this API must be used in conjonction with the
 * TAILQ_* macros. For instance, to browse the list:
 *
 *	struct cfzy_list_elt *e;
 *	TAILQ_FOREACH(e, list_head, next) {
 *		obj = e->ptr;
 *		...
 *	}
 */

#include <sys/queue.h>

/**
 * A list head
 */
TAILQ_HEAD(cfzy_list_head, cfzy_list_elt);

/**
 * A list element, used to chain objects together. This structure is
 * allocated by cfzy_list_add().
 */
struct cfzy_list_elt {
	TAILQ_ENTRY(cfzy_list_elt) next;
	void *ptr;
};

/**
 * Insert an element at the tail of the list
 *
 * Allocate a cfzy_list_elt structure, set its ptr field and insert it
 * at the tail of the list.
 *
 * @param list
 *   List head to insert the element
 * @param ptr
 *   Pointer to the object that will be referenced by list_elt->ptr
 * @return
 *   0 on succes, and negative on error
 */
int cfzy_list_add_tail(struct cfzy_list_head *list, void *ptr);

/**
 * Insert an element at the head of the list
 *
 * Allocate a cfzy_list_elt structure, set its ptr field and insert it
 * at the head of the list.
 *
 * @param list
 *   List head to insert the element
 * @param ptr
 *   Pointer to the object that will be referenced by list_elt->ptr
 * @return
 *   0 on succes, and negative on error
 */
int cfzy_list_add_head(struct cfzy_list_head *list, void *ptr);

/**
 * Remove an element from the list
 *
 * Unchain the cfzy_list_elt structure and free it.
 *
 * @param list
 *   List head to insert the element
 * @param elt
 *   List element
 */
void cfzy_list_del(struct cfzy_list_head *list, struct cfzy_list_elt *elt);

/**
 * Check if an element is in the list
 *
 * Browse the list and check if there is already an cfzy_list_elt
 * structure referencing ptr.
 *
 * @param list
 *   List head
 * @param ptr
 *   Pointer to the element
 * @return
 *   1 if the element is in list, else 0
 */
int cfzy_list_elt_is_in_list(const struct cfzy_list_head *list, const void *ptr);

/**
 * Allocate a new list
 *
 * @return
 *   Pointer to the new list head, or NULL on error
 */
struct cfzy_list_head *cfzy_list_alloc(void);

/**
 * Free a list
 *
 * Free all cfzy_list_elt of the list using the free_fct if provided,
 * then free the cfzy_list_head structure.
 *
 * @param list
 *   List head pointer
 * @param free_fct
 *   Function to free an element
 */
void cfzy_list_free(struct cfzy_list_head *list, void (*free_fct)(void *));

#endif /* _CFZY_LIST_H_ */
