/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this log of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this log of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include "cfzy_log.h"
#include "cfzy_htable.h"

static struct cfzy_htable *level_htable = NULL;
static int default_level = CFZY_LOG_NOTICE;

/* initialize log susbsystem */
int cfzy_log_init(void)
{
	level_htable = cfzy_htable_alloc();
	if (level_htable == NULL)
		return -1;
	return 0;
}

/* free all structures associated to log subsystem */
void cfzy_log_exit(void)
{
	cfzy_htable_free(level_htable, free);
}

/* set default log level (for log types that are not registered) */
int cfzy_log_set_default_level(int level)
{
	if (level != -1 && (level < CFZY_LOG_ERR || level > CFZY_LOG_DEBUG))
		return -1;
	default_level = level;
	return 0;
}

/* set log level for a specific log type: the string is added in the
 * hash table */
int cfzy_log_set_level(const char *logtype, int level)
{
	int *l;

	if (level != -1 && (level < CFZY_LOG_ERR || level > CFZY_LOG_DEBUG))
		return -1;
	l = cfzy_htable_lookup(level_htable, logtype);

	/* if log is already registered, it's easy */
	if (l != NULL) {
		*l = level;
		return 0;
	}

	/* else allocate a new (int *) and store it in the htable */
	l = malloc(sizeof(*l));
	if (l == NULL)
		return -1;

	*l = level;
	cfzy_htable_add(level_htable, logtype, l);
	return 0;
}

/* print a log if given level is <= to configured log level */
int cfzy_log_printf(const char *logtype, int level, const char *fmt, ...)
{
	int *l;
	va_list ap;
	int ret;

	l = cfzy_htable_lookup(level_htable, logtype);

	if (l == NULL && level > default_level)
		return 0;
	else if (l != NULL && level > *l)
		return 0;

	va_start(ap, fmt);
	ret = vfprintf(stderr, fmt, ap);
	va_end(ap);

	return ret;
}
