/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include "cfzy_log.h"

#define LOG(level, fmt, args...)				\
	CFZY_LOG("string", level, fmt, ##args)

/* Acts as a asprintf, except that it appends data after buf (if not
 * NULL), reallocating memory if necessary. */
int cfzy_string_asprintf(char **buf, const char *fmt, ...)
{
	va_list ap;
	char dummy;
	int offset = 0, buflen, ret;

	if (*buf != NULL)
		offset = strlen(*buf);

	va_start(ap, fmt);
	ret = vsnprintf(&dummy, 1, fmt, ap);
	va_end(ap);
	if (ret < 0) {
		LOG(ERR, "first vsnprintf failed\n");
		return ret;
	}

	buflen = offset + ret + 1;
	*buf = realloc(*buf, buflen);
	if (*buf == NULL) {
		LOG(ERR, "asprintf: not enough memory\n");
		return -1;
	}

	va_start(ap, fmt);
	ret = vsnprintf(*buf + offset, buflen - offset, fmt, ap);
	va_end(ap);

	if (ret >= buflen || ret < 0) {
		free(*buf);
		*buf = NULL;
		LOG(ERR, "second vsnprintf failed\n");
		return -1;
	}

	return ret;
}

/*
 * If the buffer is surrounded by quotes (simple or double), this
 * function will allocate a new string and remove the quotes properly,
 * also deleting the backslash in occurences of \\, \" or \'. If the
 * buffer does not start by a quote, the function just return a
 * duplicate of the input string. On sucess, the eatlen is updated to
 * the number of consumed bytes (including quotes if any, but not
 * including \0). On error, return NULL, in this case *eatlen is
 * undefined.
 */
char *cfzy_string_unquote(const char *buf, unsigned *eatlen)
{
	char delim;
	char *out, *s2;
	const char *s;

	s = buf;
	out = strdup(s);
	if (out == NULL) {
		LOG(ERR, "not enough memory\n");
		return NULL;
	}

	if (s[0] == '"')
		delim = '"';
	else if (s[0] == '\'')
		delim = '\'';
	else {
		*eatlen = strlen(s);
		return out;
	}
	s++;

	s2 = out;
	while (*s != delim) {
		if (*s == '\0') {
			free(out);
			return NULL;
		}
		if (*s == '\\' && *(s+1) == delim) {
			*s2 = delim;
			s += 2;
			s2 ++;
			continue;
		}
		if (*s == '\\' && *(s+1) == '\\') {
			*s2 = '\\';
			s += 2;
			s2 ++;
			continue;
		}
		*s2 = *s;
		s ++;
		s2 ++;
	}
	*s2 = '\0';

	*eatlen = s - buf + 1;
	return out;
}

/* quote a string and escape original quotes */
char *cfzy_string_quote(const char *src)
{
	int s, d;
	char *dst;

	/* get dst buf len */
	for (s = 0, d = 0; src[s] != '\0'; s++, d++) {
		if (src[s] == '"')
			d++;
		if (src[s] == '\\' && src[s+1] == '"')
			d++;
	}

	dst = malloc(d + 3); /* 3 for the 2 quotes and the \0 */
	if (dst == NULL)
		return NULL;

	dst[0] = '"';
	for (s = 0, d = 1; src[s] != '\0'; s++, d++) {
		if (src[s] == '"')
			dst[d++] = '\\';
		if (src[s] == '\\' && src[s+1] == '"')
			dst[d++] = '\\';

		dst[d] = src[s];
	}

	dst[d++] = '"';
	dst[d++] = '\0';

	return dst;
}

