/*
 * Copyright (c) 2013, Olivier MATZ <zer0@droids-corp.org>
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University of California, Berkeley nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file
 * String-related tools
 */

#ifndef CFZY_STRING_H
#define CFZY_STRING_H

/**
 * Append a formatted string in an allocated buffer
 *
 * This function acts as a asprintf, with some differences:
 * - *but can already contain a string, in this case, the new string
 *   is appended.
 * - *strp is set to NULL on error, like in BSD's asprintf
 *
 * @param strp
 *   Pointer to a (char *), must not be NULL. If *strp is NULL a new
 *   buffer is allocated to store the string. Else, if *strp is not
 *   NULL, it must point to a valid string in a buffer that was
 *   previously allocated using malloc.
 * @param fmt
 *   Format string, followed by other arguments, like in printf
 * @return
 *   The number of bytes printed. If memory allocation wasn't
 *   possible, or some other error occurs, these functions will return
 *   -1, and the contents of strp is set to NULL.
 **/
int cfzy_string_asprintf(char **buf, const char *fmt, ...);

/**
 * Remove quotes on a string
 *
 * If the buffer is surrounded by quotes (simple or double), this
 * function will allocate a new string and remove the quotes properly,
 * also deleting the backslash in occurences of \\, \" or \'. If the
 * buffer does not start by a quote, the function just return a
 * duplicate of the input string.
 *
 * On sucess, the value pointer by eatlen is updated to the len of
 * consumed bytes in the input buffer, including quotes if any, but
 * not including \0. On error *eatlen is undefined.
 *
 * @param buf
 *   string buffer
 * @param eatlen
 *   pointer to an int where the number of consumed bytes is written
 * @return
 *   the unquoted string (allocated), or NULL on error
 */
char *cfzy_string_unquote(const char *buf, unsigned *eatlen);

/**
 * Quote a string with double quotes
 *
 * Allocate a new string which is a copy of the first one, surrounded
 * by quotes and quotes of initial string are escaped with a
 * backslash.
 *
 * @param src
 *   input string
 * @return
 *   the quoted string (allocated), or NULL on error
 */
char *cfzy_string_quote(const char *src);

#endif /* CFZY_STRING */
